-- Keys
local IS_IN_SHIELD_MODE = 'isInShieldMode'
local IS_STARTING_SHIELD_MODE = 'isStartingShieldMode'
local STARTING_SHIELD_MODE_TIMER_KEY = 'startingShieldModeTimer'
local IS_ENDING_SHIELD_MODE = 'isEndingShieldMode'
local ENDING_SHIELD_MODE_TIMER_KEY = 'endingShieldModeTimer'
local ORIGINAL_ARMOR_AMOUNT = 'originalArmorKey'

-- Animation Values
local SHIELD_START_ANIMATION = 'B_Fortress_Mech_Shield_Start'
local NUMBER_OF_FRAMES_IN_SHIELD_START_ANIMATION = 18

local SHIELD_END_ANIMATION = 'B_FortressMech_Shield_End'
local NUMBER_OF_FRAMES_IN_SHIELD_END_ANIMATION = 18

local SHIELD_VFX = 'imp_B_ShieldLarge.lvl'

B_FortressMech_spawnTrigger = function(entity)
    local originalArmor = objc.get_entity_value_for_key(entity, ENTITY_VALUE_KEY_ARMOR)
    objc.set_local_value(entity, ORIGINAL_ARMOR_AMOUNT, originalArmor)
end

--[[
This actionSystemLoopStartTrigger looks to see if you're already in shield mode
for attacking. If you're not, it runs the animations to put you into shield
mode.
--]]
B_FortressMech_actionSystemLoopStartTrigger = function(sourceEntity, targetEntity)
    if objc.get_local_value(sourceEntity, IS_IN_SHIELD_MODE) == 0  then
        objc.set_local_value(sourceEntity, IS_IN_SHIELD_MODE, 1)

        local numberOfSubrounds = objc.convert_frames_to_subrounds(NUMBER_OF_FRAMES_IN_SHIELD_START_ANIMATION)
        objc.idle_entity(sourceEntity, numberOfSubrounds)
        objc.play_override_animation(sourceEntity, SHIELD_START_ANIMATION, objc.subrounds_to_milliseconds(numberOfSubrounds))

        objc.reset_tick_timer(sourceEntity, STARTING_SHIELD_MODE_TIMER_KEY)
        objc.set_local_value(sourceEntity, IS_STARTING_SHIELD_MODE, 1)
    end
end

--[[
This actionSystemLoopEndTrigger has the B_FortressMech run its animation to get
out of shield mode, and also turns off the B_FortressMech's armor boost.
--]]
B_FortressMech_actionSystemLoopEndTrigger = function(entity)
    local numberOfSubrounds = objc.convert_frames_to_subrounds(NUMBER_OF_FRAMES_IN_SHIELD_END_ANIMATION)
    objc.idle_entity(entity, numberOfSubrounds)
    objc.play_override_animation(entity, SHIELD_END_ANIMATION, objc.subrounds_to_milliseconds(numberOfSubrounds))

    objc.reset_tick_timer(entity, ENDING_SHIELD_MODE_TIMER_KEY)
    objc.set_local_value(entity, IS_ENDING_SHIELD_MODE, 1)


    local currentArmorAmount = objc.get_entity_value_for_key(entity, ENTITY_VALUE_KEY_ARMOR)
    local originalArmor = objc.get_local_value(entity, ORIGINAL_ARMOR_AMOUNT)
    objc.set_entity_value_for_key(entity, ENTITY_VALUE_KEY_ARMOR, originalArmor)
    objc.set_local_value(entity, IS_IN_SHIELD_MODE, 0)
end

--[[
This tickTrigger checks to see if the start and end shield animations are
complete, and updates state depending on it.

We stop overriding the start animation 1 round early early so that we don't run
into timing issues with the attack logic playing right afterwards. By stopping
the override, the attack animation will play whenever the unit is ready.
Previously, if we didn't stop overriding a round early, we ran into an issue
where the FortressMech would attack, but wouldn't play an attack animation on
its first attack.

When the start shield animation is done, the B_Fortress_Mech's armor boost
comes into efect.
--]]
B_FortressMech_tickTrigger = function(entity)
    if objc.get_local_value(entity, IS_STARTING_SHIELD_MODE) == 1 and objc.tick_timer(entity, STARTING_SHIELD_MODE_TIMER_KEY, objc.convert_frames_to_rounds(NUMBER_OF_FRAMES_IN_SHIELD_START_ANIMATION) - 1) == 0 then
        objc.set_local_value(entity, IS_STARTING_SHIELD_MODE, 0)
        objc.stop_override_animation(entity)

        local currentArmorAmount = objc.get_entity_value_for_key(entity, ENTITY_VALUE_KEY_ARMOR)
        local originalArmor = objc.get_local_value(entity, ORIGINAL_ARMOR_AMOUNT)
        objc.set_entity_value_for_key(entity, ENTITY_VALUE_KEY_ARMOR, originalArmor * 3)
    elseif objc.get_local_value(entity, IS_ENDING_SHIELD_MODE) == 1 and objc.tick_timer(entity, ENDING_SHIELD_MODE_TIMER_KEY, objc.convert_frames_to_rounds(NUMBER_OF_FRAMES_IN_SHIELD_END_ANIMATION)) == 0 then
        objc.set_local_value(entity, IS_IN_SHIELD_MODE, 0)
        objc.set_local_value(entity, IS_ENDING_SHIELD_MODE, 0)
    end
end 

B_FortressMech_onDamageTrigger = function(damage, targetEntity, sourceEntity)
    local armorAmount = objc.get_entity_value_for_key(targetEntity, ENTITY_VALUE_KEY_ARMOR)
    local flashCutoff = objc.divide_with_floor(armorAmount * 10, 4)
    local originalDamage = damage + armorAmount
    if originalDamage <= flashCutoff then
        objc.play_vfx(targetEntity, SHIELD_VFX, NODE_IMPACT)
    end
    return damage
end