-- Keys
local IS_USING_LEAP_ATTACK_ANIMATION_KEY = 'isUsingLeapAttackAnimation'
local NEXT_ATTACK_SHOULD_STUN_KEY = 'nextAttackShouldStun'
local OLD_PARAMS_KEY = 'oldParams'
local JUMP_COOLDOWN_KEY = 'jumpCooldown'
local TARGETED_ENEMY_KEY = 'targetedEnemy'
local HAS_FOUND_LEAP_TARGET = 'hasFoundLeapTarget'
local IS_ATTACKING_KEY = 'isAttackingKey'

-- Constant Values
local SEARCH_RADIUS = 3200
local FARTHEST_SEARCH_RADIUS = 5000
local ENEMY_RANGE_THRESHOLD = 2050
local MINIMUM_JUMP_RADIUS_SQUARED = 2000000
local PLAYER_1_BASE_CUTOFF = 6500
local PLAYER_2_BASE_CUTOFF = -6500
local JUMP_COOLDOWN = 30
local STUN_DURATION = 1500

local JUMP_DECELERATION = -220

-- Animation Values
local LEAP_MOVE_START_ANIMATON_STATE = ''
local LEAP_MOVE_ANIMATON_STATE = 'leapAnimationName'
local LEAP_ATTACK_GROUND_ANIMATON_STATE = 'leapAttackAnimationName'
local LEAP_START_ANIMATION_NAME = 'B_K9_Leap_Start_01'
local LEAP_ATTACK_IMPULSE_TIME = 1
local NUMBER_OF_FRAMES_IN_LEAP_START_ANIMATION = 6
local NUMBER_OF_FRAMES_IN_LEAP_ANIMATION = 15

local REGULAR_MOVE_START_ANIMATON_STATE = 'moveStartAnimationName'
local REGULAR_MOVE_ANIMATON_STATE = 'moveAnimationName'
local REGULAR_ATTACK_GROUND_ANIMATON_STATE = 'attackAnimationName'
local REGULAR_ATTACK_IMPULSE_TIME = 2

B_K9_spawnTrigger = function(entity)
   objc.set_entity_value_for_key(entity, ENTITY_VALUE_KEY_CAN_ATTACK_PROTECTED_ENTITIES, YES)
end

--[[
This tick trigger is designed to give the K9 the ability to leap to attack a
long range unit. It works as follows:
 - Waits for a cooldown between leap attacks
 - Looks for a long range unit past a specific minimum radius
    - If there is a long range unit within the minimum radius, the K9 will not leap
 - Once it finds the long range unit to attack, it will do a special leap
   animation, with a special initial leap attack that stuns the enemy
 - It then goes back to its regular attack/move loop
--]]
B_K9_tickTrigger = function(entity)
   local hasFoundLeapTarget = objc.get_local_value(entity, HAS_FOUND_LEAP_TARGET)
   if hasFoundLeapTarget == 0 then
      --[[
       We haven't found a leap target yet. Let's search for one.
      --]]
      local enemyCheckTimer = objc.tick_timer(entity, 'enemyCheckTimer', 5)
      if enemyCheckTimer <= 0 and objc.get_local_value(entity, JUMP_COOLDOWN_KEY) == 0 then
         local teamNumber = objc.get_team_number(entity)
         local enemyNumber = objc.get_enemy_team_number(entity)

         --Check to see if there's a nearby enemy that should get pounced on
         local nearbyEnemies = objc.list_entities_in_radius(entity, SEARCH_RADIUS, enemyNumber)
         local closestDistanceSquared = 99999999
         local closestEnemy = entity

         for _, enemy in pairs(nearbyEnemies) do
            if objc.is_normal_unit(enemy) or objc.is_commander(enemy) or objc.is_building_unit(enemy) or objc.is_barricade(enemy) or objc.is_base(enemy) then
               local enemyRange = objc.get_entity_value_for_key(enemy, ENTITY_VALUE_KEY_ATTACK_RANGE) 
               local enemyDistance = objc.quick_unit_distance_square(entity, enemy)

               if closestDistanceSquared >= enemyDistance and enemyRange >= ENEMY_RANGE_THRESHOLD and objc.is_ground_unit(enemy) then
                  closestDistanceSquared = enemyDistance
                  closestEnemy = enemy

                  if closestDistanceSquared <= MINIMUM_JUMP_RADIUS_SQUARED then
                     -- There's a long-range unit within the minimum jump radius. 
                     -- Don't jump
                     closestEnemy = entity
                     break
                  end
               end
            end
         end

         if closestEnemy ~= entity then
            -- We found a suitable long range enemy to attack. Save it here to
            -- attack after the initial leap start animations
            objc.set_local_value(entity, HAS_FOUND_LEAP_TARGET, 1)
            objc.set_local_value(entity, TARGETED_ENEMY_KEY, closestEnemy)

            local leapAnimations = {}
            leapAnimations[ANIMATION_STATE_MOVE_START] = LEAP_MOVE_START_ANIMATON_STATE
            leapAnimations[ANIMATION_STATE_MOVE] = LEAP_MOVE_ANIMATON_STATE
            leapAnimations[ANIMATION_STATE_ATTACK_GROUND] = LEAP_ATTACK_GROUND_ANIMATON_STATE
            objc.update_animation_state(entity, leapAnimations)

            objc.set_entity_value_for_key(entity, ENTITY_VALUE_KEY_IMPULSE_TIME, LEAP_ATTACK_IMPULSE_TIME)

            -- Play leap start animations
            local numberOfSubrounds = objc.convert_frames_to_subrounds(NUMBER_OF_FRAMES_IN_LEAP_START_ANIMATION)
            objc.idle_entity(entity, numberOfSubrounds)
            objc.play_override_animation(entity, LEAP_START_ANIMATION_NAME, objc.subrounds_to_milliseconds(numberOfSubrounds))
         end
      elseif objc.get_local_value(entity, JUMP_COOLDOWN_KEY) == 1 then
         jumpCooldownTimer = objc.tick_timer(entity, 'jumpCooldownTimer', JUMP_COOLDOWN)
         if jumpCooldownTimer == 0 then
            objc.set_local_value(entity, JUMP_COOLDOWN_KEY, 0)
         end
      end
   else
      --[[
       We've already found the leap target.
       Countdown until we're done with the leap start animation.
       Then pounce!
      --]]
      local leapStartTimer = objc.tick_timer(entity, 'leapStartTimer', objc.convert_frames_to_rounds(NUMBER_OF_FRAMES_IN_LEAP_START_ANIMATION))
      if leapStartTimer == 0 then
         -- Reset values!
         objc.set_local_value(entity, HAS_FOUND_LEAP_TARGET, 0)

         -- Figure out where the target entity may be at the end of the leap animation
         local numberOfTicks = objc.convert_frames_to_rounds(NUMBER_OF_FRAMES_IN_LEAP_ANIMATION)
         local targetEntity = objc.get_local_value(entity, TARGETED_ENEMY_KEY)
         local targetEntityXVelocity = objc.get_entity_value_for_key(entity, ENTITY_VALUE_KEY_VELOCITY_X)
         local targetEntityYVelocity = objc.get_entity_value_for_key(entity, ENTITY_VALUE_KEY_VELOCITY_Y)

         local targetEntityEstimatedXOffset = targetEntityXVelocity
         local targetEntityEstimatedYOffset = targetEntityYVelocity

         local targetPos = objc.get_entity_position(targetEntity)

         local estimatedTargetEntityDistance = objc.unit_distance_with_offset(entity, 0, 0, targetEntity, targetEntityEstimatedXOffset, targetEntityEstimatedYOffset)

         -- The K9 will stop leaping once it gets in range of its target.
         -- We want the leap animation to take the entire time of the leap,
         -- so modify accordingly.
         local attackRange = objc.get_entity_value_for_key(entity, ENTITY_VALUE_KEY_ATTACK_RANGE)
         estimatedTargetEntityDistance = estimatedTargetEntityDistance - attackRange

         -- We don't want the leap animation to play at 1.0 scale, not faster
         objc.override_move_animation_scale(entity, 10)

         -- Since we can't guarantee the exact timing of when the leap
         -- animation finishes and the attack animation starts. To make it
         -- look good, we're going to stretch the animation one round past
         -- how long it would normally take.
         local leapSpeed = objc.divide_with_ceiling(estimatedTargetEntityDistance, (objc.convert_frames_to_rounds(NUMBER_OF_FRAMES_IN_LEAP_ANIMATION) + 1))
         local timeoutRounds = 10

         local oldParams = objc.move_entity_to_entity_with_motion(entity, targetEntity, MOTION_STATE_LEAP, leapSpeed)
         objc.set_local_value(entity, OLD_PARAMS_KEY, oldParams)
      
         objc.set_local_value(entity, IS_USING_LEAP_ATTACK_ANIMATION_KEY, 1)
         
         objc.set_targeting(entity, targetEntity)

         --Start a jump cooldown
         objc.set_local_value(entity, JUMP_COOLDOWN_KEY, 1)
      end
   end
   return 0
end


--[[
This onOverrideTargetingCompletionTrigger switches the K9's movement animations
movement overrides to their previous values.
--]]
B_K9_onOverrideTargetingCompletionTrigger = function(sourceEntity, targetEntity, targetingFailed)
   --If we haven't stored any old params, we're being told to move by another entity.
   --So, don't do leap-related things.
   local oldParams = objc.get_local_value(sourceEntity, OLD_PARAMS_KEY)
   if oldParams ~= 0 then
      local regularAnimations = {}
      regularAnimations[ANIMATION_STATE_MOVE_START] = REGULAR_MOVE_START_ANIMATON_STATE
      regularAnimations[ANIMATION_STATE_MOVE] = REGULAR_MOVE_ANIMATON_STATE
      objc.update_animation_state(sourceEntity, regularAnimations)

      objc.stop_override_move_animation_scale(sourceEntity)

      objc.swap_out_entity_motion_state_with_dictionary(sourceEntity, oldParams)
      objc.set_targeting(sourceEntity, targetEntity)
   end

   return 0
end

--[[
This onAttackTrigger sets a flag to prevent the K9 from leaping to another 
target while currently attacking one. This helps prevent the K9 from leaping
behind the turret, a degenerate behavior that is unintentional.
]]
B_K9_onAttackTrigger = function(sourceEntity, targetEntity)
   objc.set_local_value(sourceEntity, IS_ATTACKING_KEY, 1)

   local isUsingLeapAttackAnimation = objc.get_local_value(sourceEntity, IS_USING_LEAP_ATTACK_ANIMATION_KEY)

   if isUsingLeapAttackAnimation == 1 then
      objc.set_local_value(sourceEntity, IS_USING_LEAP_ATTACK_ANIMATION_KEY, 0)
      local regularAnimations = {}
      regularAnimations[ANIMATION_STATE_ATTACK_GROUND] = REGULAR_ATTACK_GROUND_ANIMATON_STATE
      objc.update_animation_state(sourceEntity, regularAnimations)

      objc.set_entity_value_for_key(sourceEntity, ENTITY_VALUE_KEY_IMPULSE_TIME, REGULAR_ATTACK_IMPULSE_TIME)
   end
end

--[[
This movement trigger clears the flag preventing the K9 from jumping.
We do this in onMovement because onKill is unreliable.
]]
B_K9_onMovementTrigger = function(entity)
   objc.set_local_value(entity, IS_ATTACKING_KEY, 0)
end
