#define SUPPORT_GLSL300
#include "prefix.sh"
#include "texture_utility.sh"

// Based on work of Morgan McGuire (http://casual-effects.blogspot.com/2013/09/the-skylanders-swap-force-depth-of.html),
// Colin Barré-Brisebois and John White (https://colinbarrebrisebois.com/2017/04/18/hexagonal-bokeh-blur-revisited/)
                                                                           
uniform sampler2D s_cc3Texture2Ds[4];

uniform int CinematicDoFBlur_sampleStride;
uniform int CinematicDoFBlur_maxCoCRadiusPixels;
uniform int CinematicDoFBlur_nearBlurRadiusPixels;
uniform float CinematicDoFBlur_invNearBlurRadiusPixels;
uniform vec2 CinematicDoFBlur_blurDirectionsVertical;
uniform vec2 CinematicDoFBlur_blurDirectionsDiagonal;
uniform vec2 CinematicDoFBlur_blurDirectionsReflection;

VARYING_IN vec2 v_texCoord[1];

#if BLUR_PASS == 0
    // layout(location = 0) out vec4 midFarFieldOutput0;
    #define midFarFieldOutput0 destColor
    layout(location = 1) out vec4 nearFieldOutput0;
    layout(location = 2) out vec4 midFarFieldOutput1;
    layout(location = 3) out vec4 nearFieldOutput1;
#else
    // layout(location = 0) out vec4 midFarFieldOutput;
    #define midFarFieldOutput destColor
    layout(location = 1) out vec4 nearFieldOutput;
#endif

#define saturate(s) clamp(s, 0.0, 1.0)

bool inNearField(in float radiusPixels) 
{
    return radiusPixels > 0.25;
}

float unpackCoC(float packedCoC)
{
    return (packedCoC * 2.0 - 1.0) * float(CinematicDoFBlur_maxCoCRadiusPixels);
}

float calculateNearFieldCoverage(int delta, float sampleCoC)
{
    // Extract coverage from the near field radius. Curve the contribution based on the radius.
    float coverage = float(float(abs(delta)) <= sampleCoC) * saturate(sampleCoC * CinematicDoFBlur_invNearBlurRadiusPixels * 4.0);
    
    // Optionally increase edge fade contrast in the near field
    coverage *= coverage; coverage *= coverage; 
    
    return coverage;
}

void directionalBlur(sampler2D blurSource, sampler2D nearSource, vec2 direction, out vec4 nearOutput, out vec4 midFarOutput)
{
    nearOutput = vec4(0.0);
    midFarOutput.rgb = vec3(0.0);
    float nearWeightSum = 0.0;
    float midFarWeightSum = 0.0;
    
    vec4 centerTapValue = sampleLevelZero(blurSource, v_texCoord[0]);
    float centerTapPackedCoC = centerTapValue.a;
    float centerCoC = unpackCoC(centerTapPackedCoC);
    float centerAbsCoC = abs(centerCoC);
    float nearFieldness = saturate(centerCoC * 4.0); // Map centerCoC << 0 to 0, centerCoC >> 0 to 1
        
    for (int delta = 0; delta <= CinematicDoFBlur_maxCoCRadiusPixels; delta += CinematicDoFBlur_sampleStride) {
        vec2 sampleCoord = v_texCoord[0] + direction * float(delta);
        vec4 sampleTapValue = sampleLevelZero(blurSource, sampleCoord);
        float sampleCoC = unpackCoC(sampleTapValue.a);

        // Compute mid-far-field
        vec3 midFarColor = sampleTapValue.rgb;
        
        bool rejectSample = float(delta) > abs(sampleCoC) * 0.8 * 2.0;
        float midFarWeight = mix(/*float(!inNearField(sampleCoC)) * */float(!rejectSample), 1.0, nearFieldness);

        midFarWeightSum += midFarWeight;
        midFarOutput.rgb += midFarColor * midFarWeight;

        // Compute near-field
        vec4 nearColor;
        
        #if BLUR_PASS == 0
            nearColor.a = calculateNearFieldCoverage(delta, sampleCoC);
            nearColor.rgb = sampleTapValue.rgb;
            nearColor.rgb *= nearColor.a;
        #else
            // On the last pass, use the already-available alpha values
            nearColor = sampleLevelZero(nearSource, sampleCoord);
        #endif
        
        float nearWeight = float(delta < CinematicDoFBlur_nearBlurRadiusPixels);
        
        nearWeightSum += nearWeight;
        nearOutput += nearColor * nearWeight;
    }
    
    #if BLUR_PASS == 0
        // Retain the packed radius on the first pass. On the second pass it is not needed.
        midFarOutput.a = centerTapPackedCoC;
    #else
        midFarOutput.a = 1.0;
    #endif

    // Normalize the blur
    midFarOutput.rgb /= max(midFarWeightSum, 0.00001);
    nearOutput /= max(nearWeightSum, 0.00001);
}

void main()
{
    #if BLUR_PASS == 0
        vec4 near0, midFar0;
        directionalBlur(s_cc3Texture2Ds[0], s_cc3Texture2Ds[0] /* unused */, CinematicDoFBlur_blurDirectionsVertical, near0, midFar0);
    
        vec4 near1, midFar1;
        directionalBlur(s_cc3Texture2Ds[0], s_cc3Texture2Ds[0] /* unused */, CinematicDoFBlur_blurDirectionsDiagonal, near1, midFar1);
    
        nearFieldOutput0 = near0;
        midFarFieldOutput0 = midFar0;
        
        nearFieldOutput1 = (near0 + near1) * 0.5;
        midFarFieldOutput1 = (midFar0 + midFar1) * 0.5;
    #else
        vec4 near0, midFar0;
        directionalBlur(s_cc3Texture2Ds[0], s_cc3Texture2Ds[1], CinematicDoFBlur_blurDirectionsDiagonal, near0, midFar0);
    
        vec4 near1, midFar1;
        directionalBlur(s_cc3Texture2Ds[2], s_cc3Texture2Ds[3], CinematicDoFBlur_blurDirectionsReflection, near1, midFar1);
    
        nearFieldOutput = (near0 + near1 * 2.0) / 3.0;
        midFarFieldOutput = (midFar0 + midFar1 * 2.0) / 3.0;
    #endif
}
