#define SUPPORT_GLSL300
#include "prefix.sh"
#include "texture_utility.sh"

// Adopted from work of Morgan McGuire (http://casual-effects.blogspot.com/2013/09/the-skylanders-swap-force-depth-of.html)

uniform sampler2D s_cc3Texture2Ds[3];

VARYING_IN vec2 v_texCoord[1];

#define saturate(s) clamp(s, 0.0, 1.0)

const float kFarRadiusRescale = 1.0;
const vec2 kCocReadScaleBias = vec2(2.0, -1.0);

// Boost the coverage of the near field by this factor.  Should always be >= 1
//
// Make this larger if near-field objects seem too transparent
//
// Make this smaller if an obvious line is visible between the near-field blur and the mid-field sharp region
// when looking at a textured ground plane.
const float kCoverageBoost = 1.5;

void main()
{
    vec4 pack    = sampleLevelZero(s_cc3Texture2Ds[0], v_texCoord[0]);
    vec3 sharp   = pack.rgb;
    vec3 blurred = sampleLevelZero(s_cc3Texture2Ds[1], v_texCoord[0]).rgb;
    vec4 near    = sampleLevelZero(s_cc3Texture2Ds[2], v_texCoord[0]);

    // Signed, normalized radius of the circle of confusion.
    // |normRadius| == 1.0 corresponds to camera->maxCircleOfConfusionRadiusPixels()
    float normRadius = pack.a * kCocReadScaleBias.x + kCocReadScaleBias.y;

    // Fix the far field scaling factor so that it remains independent of the 
    // near field settings
    normRadius *= (normRadius < 0.0) ? kFarRadiusRescale : 1.0;

    // Boost the blur factor
    // normRadius = clamp(normRadius * 2.0, -1.0, 1.0);

    // Decrease sharp image's contribution rapidly in the near field
    // (which has positive normRadius)
    if (normRadius > 0.1) {
       normRadius = min(normRadius * 1.5, 1.0);
    }

    if (kCoverageBoost != 1.0) {
        float a = saturate(kCoverageBoost * near.a);
        near.rgb = near.rgb * (a / max(near.a, 0.001));
        near.a = a;
    }
    
    #if defined(DEBUG_VIEW)
        if (pack.a < 0.49) {
            // Far field: Dark blue
            gl_FragColor.rgb = vec3(0.0, 0.07, 0.4) * (dot(sharp, vec3(1.0 / 3.0)) * 0.7 + 0.3);
        } else if (pack.a <= 0.51) {
            // Mifield: Red
            gl_FragColor.rgb = vec3(1.0, 0.15, 0.15) * (dot(sharp, vec3(1.0 / 3.0)) * 0.7 + 0.3);
        } else {
            gl_FragColor.rgb = vec3(1.0, 1.0, 0.15) * (dot(sharp, vec3(1.0 / 3.0)) * 0.7 + 0.3);
        }
    #else
        // Two mixs, the second of which has a premultiplied alpha
        gl_FragColor.rgb = mix(sharp, blurred, abs(normRadius)) * (1.0 - near.a) + near.rgb;
        gl_FragColor.a = 1.0;
    #endif
}
