-- Keys
local IS_STARTING_IMMOBILIZATION_KEY = 'isStartingImmobilization'
local IS_STARTING_IMMOBILIZATION_TIMER_KEY = 'isStartingImmobilizationTimer'
local WAS_IMMOBILIZED_KEY = 'wasImmobilized'
local VFX_TIMER_KEY = 'vfxTimer'

-- Constant Values
local IMMOBILIZATION_SUBROUNDS = 5000 --Five seconds!
local REGULAR_MAX_VELOCITY = 60

-- Animation Values
local IMMOBILIZE_IDLE_STATE = 'damageIdleAnimationName'
local IMMOBILIZE_ATTACK_GROUND_STATE = 'damageAttackAnimationName'

local IMMOBILIZE_START_ANIMATION = 'R_Junker_Damage_Start_01'
local NUMBER_OF_FRAMES_IN_IMMOBILIZE_START_ANIMATION = 35

local IMMOBILIZE_END_ANIMATION = 'R_Junker_Damage_End_01'
local NUMBER_OF_FRAMES_IN_IMMOBILIZE_END_ANIMATION = 20

local REGULAR_IDLE_STATE = ''
local REGULAR_ATTACK_GROUND_STATE = 'attackAnimationName'

--[[
This OnDamageTrigger checks to see if the unit is already immobilized. If they
are, it will reset the immobilization timer. If not, it will play the
animations to get the unit into an immobilization state..
--]]
R_Junker_onDamageTrigger = function(damage,targetEntity,sourceEntity)
    if objc.is_entity_immobilized(targetEntity) then
        -- We're already immobilized. Just reset the immobilize timer.
        objc.immobilize_entity(targetEntity, IMMOBILIZATION_SUBROUNDS)
    elseif objc.get_local_value(targetEntity, IS_STARTING_IMMOBILIZATION_KEY) == 0 then
        -- We're not immobilized, and we haven't started the immobilization
        -- transition. Let's play our animations to go into immobilize.
        local numberOfSubrounds = objc.convert_frames_to_subrounds(NUMBER_OF_FRAMES_IN_IMMOBILIZE_START_ANIMATION)
        objc.idle_entity(targetEntity, numberOfSubrounds)
        objc.play_override_animation(targetEntity, IMMOBILIZE_START_ANIMATION, objc.subrounds_to_milliseconds(numberOfSubrounds))

        local immobilizedAnimations = {}
        immobilizedAnimations[ANIMATION_STATE_ATTACK_GROUND] = IMMOBILIZE_ATTACK_GROUND_STATE
        immobilizedAnimations[ANIMATION_STATE_IDLE] = IMMOBILIZE_IDLE_STATE
        objc.update_animation_state(targetEntity, immobilizedAnimations)

        objc.reset_tick_timer(targetEntity, IS_STARTING_IMMOBILIZATION_TIMER_KEY)
        objc.set_local_value(targetEntity, IS_STARTING_IMMOBILIZATION_KEY, 1)
    end
    return damage
end


--[[
This TickTrigger:
  - If the Junker was running the start immobilization animation and is finished, starts being immobilized
  - If the Junker is immobilized, plays the immobilization VFX
  - If the Junker was immobilized, and is finished being immobilized, starts playing the finished immobilization animation
  - If the Junker was playing the finished immobilization animation, goes back to normal gameplay
--]]
R_Junker_tickTrigger = function(entity)
    if objc.get_local_value(entity, IS_STARTING_IMMOBILIZATION_KEY) == 1 and objc.tick_timer(entity, IS_STARTING_IMMOBILIZATION_TIMER_KEY, objc.convert_frames_to_rounds(NUMBER_OF_FRAMES_IN_IMMOBILIZE_START_ANIMATION)) == 0 then
        -- We were running the starting immobilization animation
        -- And we've finished. Start being immobilized!
        objc.set_local_value(entity, IS_STARTING_IMMOBILIZATION_KEY, 0)
        objc.immobilize_entity(entity, IMMOBILIZATION_SUBROUNDS)
    elseif objc.is_entity_immobilized(entity) then
        -- We're immobilized. Play the VFX!
        objc.set_local_value(entity, WAS_IMMOBILIZED_KEY, 1)
    elseif not objc.is_entity_immobilized(entity) and objc.get_local_value(entity, WAS_IMMOBILIZED_KEY) == 1 then
        -- We're were immobilized and are not anymore. Let's play our
        -- animations to go into regular gameplay.
        objc.set_local_value(entity, WAS_IMMOBILIZED_KEY, 0)

        local numberOfSubrounds = objc.convert_frames_to_subrounds(NUMBER_OF_FRAMES_IN_IMMOBILIZE_END_ANIMATION)
        objc.idle_entity(entity, numberOfSubrounds)
        objc.play_override_animation(entity, IMMOBILIZE_END_ANIMATION, objc.subrounds_to_milliseconds(numberOfSubrounds))

        local regularAnimations = {}
        regularAnimations[ANIMATION_STATE_IDLE] = REGULAR_IDLE_STATE
        regularAnimations[ANIMATION_STATE_ATTACK_GROUND] = REGULAR_ATTACK_GROUND_STATE 
        objc.update_animation_state(entity, regularAnimations)
    end

	return 0
end
