//
//  Shaders.metal
//  MetalTest
//
//  Created by Thiago Goulart on 12/16/14.
//  Copyright (c) 2014 Thiago Goulart. All rights reserved.
//

#include <metal_stdlib>
#include <simd/simd.h>

using namespace metal;

typedef struct {
    float3 position;
    float2 uv[3];
    float3 normal;
    float3 tangent;
    float3 bitangent;
    float4 color;
    float4 boneWeight;
    uchar4 boneIndex;
} vertex_t;

typedef struct {
    float4 position;  // xyz = position
    float4 speed;   // xyz = linear speed, w = rotation speed
    float4 uv_life; // xy = UV, z = lifespan, w = delay
    float4 color1;  // rgba color
    float4 color2;  // rgba color
    float4 color3;  // rgba color
    float4 scale;   // x, y = intermediate scale (x, y); z, w = final scale (x, y)
    float4 offsetPosition; // xyz = offset position
    float4 rotation;   // x = initial rotation
} particle_vertex_t;

#define MAX_BONES 33

typedef struct {
    matrix_float4x4 model_matrix;
    matrix_float4x4 view_matrix;
    matrix_float4x4 projection_matrix;
    matrix_float4x4 bone_matrices[MAX_BONES];

    float4 diffuseColor;

    float time;

    // godray
    float nearDist;
    float nearScale;
    float farDist;
    float farScale;

    float minOpacity;
    float maxOpacity;
    float noisePeriod;

    float2 speed;

    // ocean
    float3 cameraPositionGlobal;

    float waveMinHeight;
    float waveMaxHeight;

    float waveLength;
    float waveFrequency;

    float2 waveSpeed;
    float2 noiseSpeed;
    float2 foamSpeed;

    float waveTexCoordFactor;
    float noiseTexCoordFactor;

    float noiseLevel;

    float3 normalClamp;

    float reflectionLevel;
    float reflectionPower;

    float3 deepOceanBlendFactor;
    float3 shallowOceanBlendFactor;

    float3 foamColor;
    float foamMultiplier;

    // padding
    float __padding__;
} uniforms_t;

typedef struct {
    float brightness_threshold;

    // padding
    float __padding__;
} brightnesspp_uniforms_t;

#define NUM_LIGHTS 5

typedef struct {
    float shouldUseLights[NUM_LIGHTS];
    float lightsX[NUM_LIGHTS];
    float lightsY[NUM_LIGHTS];
    float lightsIntensity[NUM_LIGHTS];
} godraypp_uniforms_t;

typedef struct {
    // vert
    matrix_float4x4 modelMatrix;
    matrix_float4x4 viewMatrix;
    matrix_float4x4 orientationMatrix;
    matrix_float4x4 projectionMatrix;
    float emitterDuration;
    float time;
    float particleSize;
    float xzAngularSpeed;
    float xzAngularAcceleration;
    int2 numberExtraRowsColumns;
    float3 radialSpeed;
    float3 radialAcceleration;
    float3 linearAcceleration;
    // frag
    float opacity;
    float useColor;
    float multiplyColorWithRGB;
    float4 diffuseColor;
} particle_uniforms_t;

typedef struct {
    float4 position [[position]];
    float3 position_view;
    float2 texCoord0;
    float2 texCoord0_scrolling;
    float2 texCoord1;
    float3 normal;
    float3 tangent;
    float3 bitangent;
    float4 color;
} ColorInOut;

typedef struct {
    float4 position [[position]];
    float3 viewpoint_world;
    float2 waveTexCoord;
    float2 noiseTexCoord;
    float3 normal_world;
    float2 foamTexCoord;
    float2 lightMapTexCoord;
} OceanInOut;

// definition of utility functions
#define M_PI 3.141592653589793f

#define CUBEMAP_DIR_CORRECTION float3(-1.0, -1.0, 1.0)

matrix_float4x4 mat4(matrix_float3x3 m);
matrix_float3x3 mat3(matrix_float4x4 m);

float fadeFactor(float cameraDist, constant uniforms_t &uniforms);
float blinkNoise(float time, float period, constant uniforms_t &uniforms);
matrix_float4x4 mat4FromQuaternion(float4 q);
float4 quaternionMultiply(float4 a, float4 b);
float4 quaternionFromAxisAngle(float3 axis, float angle);
matrix_float4x4 rotationInverse(matrix_float4x4 matrix);
matrix_float4x4 billboarded_orientation_matrix(matrix_float4x4 modelMatrix, matrix_float4x4 viewMatrix, float3 normalInModel, float3 billboard_axis);

float2 sanitize_time2(float originalTime, float2 multiplier);

// #############################
// ## Vertex shader functions ##
// #############################

vertex ColorInOut static_vertex(device vertex_t *vertex_array [[buffer(0)]],
                                constant uniforms_t &uniforms [[buffer(1)]],
                                unsigned short vid [[vertex_id]]) {
    ColorInOut out;

    out.position = uniforms.projection_matrix * uniforms.view_matrix * uniforms.model_matrix * float4(vertex_array[vid].position, 1.0);
    out.texCoord0 = vertex_array[vid].uv[0];
    out.texCoord0_scrolling = vertex_array[vid].uv[0] + uniforms.time * uniforms.speed;
    out.texCoord1 = vertex_array[vid].uv[1];
    out.normal = normalize((uniforms.model_matrix * float4(vertex_array[vid].normal, 0.0)).xyz);
    out.tangent = normalize((uniforms.model_matrix * float4(vertex_array[vid].tangent, 0.0)).xyz);
    out.bitangent = normalize((uniforms.model_matrix * float4(vertex_array[vid].bitangent, 0.0)).xyz);

    return out;
}

vertex ColorInOut skinning_vertex(device vertex_t *vertex_array [[buffer(0)]],
                                  constant uniforms_t &uniforms [[buffer(1)]],
                                  unsigned short vid [[vertex_id]]) {
    float4 skinned_position = float4(0);
    float4 skinned_normal = float4(0);
    float4 skinned_tangent = float4(0);
    float4 skinned_bitangent = float4(0);
    for (int i = 0; i < 4; ++i) {
        int index = vertex_array[vid].boneIndex[i];
        float weight = vertex_array[vid].boneWeight[i];
        skinned_position += uniforms.bone_matrices[index] * float4(vertex_array[vid].position, 1.0) * weight;
        skinned_normal += uniforms.bone_matrices[index] * float4(vertex_array[vid].normal, 0.0) * weight;
        skinned_tangent += uniforms.bone_matrices[index] * float4(vertex_array[vid].tangent, 0.0) * weight;
        skinned_bitangent += uniforms.bone_matrices[index] * float4(vertex_array[vid].bitangent, 0.0) * weight;
    }

    ColorInOut out;

    out.position = uniforms.projection_matrix * uniforms.view_matrix * uniforms.model_matrix * skinned_position;
    out.texCoord0 = vertex_array[vid].uv[0];
    out.texCoord1 = vertex_array[vid].uv[1];
    out.normal = normalize((uniforms.model_matrix * skinned_normal).xyz);
    out.tangent = normalize((uniforms.model_matrix * skinned_tangent).xyz);
    out.bitangent = normalize((uniforms.model_matrix * skinned_bitangent).xyz);

    return out;
}

vertex ColorInOut fullscreen_vertex(device vertex_t *vertex_array [[buffer(0)]],
                                    constant uniforms_t &uniforms [[buffer(1)]],
                                    unsigned short vid [[vertex_id]]) {
    ColorInOut out;

    out.position = float4(vertex_array[vid].position, 1.0);
    out.texCoord0 = vertex_array[vid].uv[0];
    out.texCoord1 = vertex_array[vid].uv[1];

    return out;
}

vertex ColorInOut godray_vertex(device vertex_t *vertex_array [[buffer(0)]],
                                constant uniforms_t &uniforms [[buffer(1)]],
                                unsigned short vid [[vertex_id]]) {
    device vertex_t &v = vertex_array[vid];

    ColorInOut out;

    matrix_float4x4 billboardMatrix = billboarded_orientation_matrix(uniforms.model_matrix, uniforms.view_matrix, v.normal, float3(1.0, 0.0, 0.0));

    float4 position_view = uniforms.view_matrix * uniforms.model_matrix * billboardMatrix * float4(v.position, 1.0);

    out.texCoord0 = v.uv[0];
    out.texCoord1 = v.uv[1];

    out.position_view = position_view.xyz;
    out.position = uniforms.projection_matrix * position_view;

    return out;
}

vertex ColorInOut particle_vertex(device particle_vertex_t *vertex_array [[buffer(0)]],
                                  constant particle_uniforms_t &uniforms [[buffer(1)]],
                                  unsigned short vid [[vertex_id]]) {
    device particle_vertex_t &v = vertex_array[vid];
    ColorInOut out;

    float delay = v.uv_life.w;
    float lifetime = v.uv_life.z;
    float realTime = uniforms.time - delay;
    float t = realTime - float(int(realTime / lifetime)) * lifetime;
    float t2 = 0.5 * t * t; // just to save some instructions

    float isFirstHalf = float(t < 0.5 * lifetime);
    float isSecondHalf = float(t >= 0.5 * lifetime);
    float halfTimeLerp = 2.0 * t / lifetime;

    float2 vectorScale = ((float2(mix(1.0, v.scale.x, halfTimeLerp), mix(1.0, v.scale.y, halfTimeLerp)) * isFirstHalf) +
                          (float2(mix(v.scale.x, v.scale.z, halfTimeLerp - 1.0), mix(v.scale.y, v.scale.w, halfTimeLerp - 1.0)) * isSecondHalf)) * uniforms.particleSize;

    float2 vertexOffset = (v.uv_life.xy * 2.0 - 1.0) * vectorScale;

    // rotation about origin
    float xzAngle = uniforms.xzAngularSpeed * t + t2 * uniforms.xzAngularAcceleration;
    matrix_float2x2 xzRotation = matrix_float2x2(float2( cos(xzAngle), -sin(xzAngle) ),
                                                 float2( sin(xzAngle),  cos(xzAngle) ));

    float3 translation = v.position.xyz;
    translation.xz = xzRotation * translation.xz;

    // vec multiplication is pointwise, so this scales about the origin (emitter's coordinates is handled by the cc3node.location)
    translation += normalize(translation) * (t * uniforms.radialSpeed + t2 * uniforms.radialAcceleration);

    // xyz added afterwards is equivalent to the above polar world being 'windswept'
    translation += v.speed.xyz * t + uniforms.linearAcceleration * t2;

    translation += v.offsetPosition.xyz;

    // Local rotation on the sprite itself
    float rotationSpeed = v.speed.w;
    float initialRotation = v.rotation.x;
    float angle = initialRotation + rotationSpeed * t;
    matrix_float2x2 rotationMatrix = matrix_float2x2(float2( cos(angle), sin(angle)),
                                                     float2(-sin(angle), cos(angle)));

    matrix_float4x4 modelMatrix = uniforms.modelMatrix;
    matrix_float4x4 orientationMatrix = uniforms.orientationMatrix;

    out.position = uniforms.projectionMatrix * uniforms.viewMatrix * modelMatrix * (float4(translation, 0) + orientationMatrix * float4(rotationMatrix * vertexOffset, 0, 1));

    int frameIndex = int(float((uniforms.numberExtraRowsColumns.x + 1) * (uniforms.numberExtraRowsColumns.y + 1)) * t / lifetime);
    int row = frameIndex / (uniforms.numberExtraRowsColumns.y + 1);
    int col = frameIndex - row * (uniforms.numberExtraRowsColumns.y + 1);
    out.texCoord0 = float2((float(col) + v.uv_life.x) / float(uniforms.numberExtraRowsColumns.y + 1),
                          (float(uniforms.numberExtraRowsColumns.x - row) + v.uv_life.y) / float(uniforms.numberExtraRowsColumns.x + 1));

    out.color = isFirstHalf * mix(v.color1, v.color2, halfTimeLerp) + isSecondHalf * mix(v.color2, v.color3, halfTimeLerp - 1.0);

    float shouldDiscard = float(t < 0.0) + float(uniforms.emitterDuration > 0.0) * (float(uniforms.emitterDuration < delay) + float(realTime > float(int((uniforms.emitterDuration - delay) / lifetime) + 1) * lifetime));
    out.color.a *= float(shouldDiscard < 0.5);

    return out;
}

vertex OceanInOut ocean_vertex(device vertex_t *vertex_array [[buffer(0)]],
                               constant uniforms_t &u [[buffer(1)]],
                               unsigned short vid [[vertex_id]]) {
#define SECOND_UV_INDEX 0
#define THIRD_UV_INDEX 0

    device vertex_t &v = vertex_array[vid];
    OceanInOut out;

    float4 position_world = u.model_matrix * float4(v.position, 1);
    
    out.waveTexCoord = u.waveTexCoordFactor * v.uv[0] + sanitize_time2(u.time, u.waveSpeed);

    out.noiseTexCoord = u.noiseTexCoordFactor * v.uv[0] + sanitize_time2(u.time, u.noiseSpeed);

    out.foamTexCoord = v.uv[SECOND_UV_INDEX] + sanitize_time2(u.time, u.foamSpeed);
    out.lightMapTexCoord = v.uv[THIRD_UV_INDEX];

    float waveHeight = mix(u.waveMinHeight, u.waveMaxHeight, v.color.r);

    float dist = distance(position_world.xz, float2(0.0));
    float phase = u.waveLength * dist + u.time * u.waveFrequency;

    position_world.y += waveHeight * (cos(phase) + 1.0) / 2.0;

    float jacobianXZ = 0.5 * u.waveLength * waveHeight * sin(phase) / dist;

    out.normal_world = mat3(u.model_matrix) * float3(-jacobianXZ * position_world.x, 1.0, -jacobianXZ * position_world.z);

    out.viewpoint_world = u.cameraPositionGlobal - position_world.xyz;

    out.position = u.projection_matrix * u.view_matrix * position_world;

    return out;
}

fragment float4 ocean_fragment(OceanInOut in [[stage_in]],
                               constant uniforms_t &u [[buffer(0)]],
                               texture2d<float> diffuseTexture [[texture(0)]],
                               texture2d<float> normalTexture [[texture(1)]],
                               texture2d<float> foamTexture [[texture(2)]],
                               texture2d<float> controlTexture [[texture(3)]],
                               texturecube<float> irradianceMap [[texture(6)]],
                               sampler defaultSampler [[sampler(0)]]) {
    float4 noiseTexel = normalTexture.sample(defaultSampler, in.noiseTexCoord);
    float2 delta = u.noiseLevel * (noiseTexel.xy * 2.0 - 1.0);
    float2 waveTexCoord = in.waveTexCoord + delta;

    float4 oceanTexel = diffuseTexture.sample(defaultSampler, waveTexCoord);
    float4 normalTexel = normalTexture.sample(defaultSampler, waveTexCoord);

//    float3 normal = normalTexel.xyz * u.normalClamp * 2.0 - u.normalClamp;
//    float3 V = in.viewpoint_world;

//    float3 N = float3(normal.x, normal.z, -normal.y);

//    float3 R = reflect(-V, N);

    float3 reflectionColor = float3(0, 0, 0);//reflectionMap.sample(defaultSampler, R).rgb;
    float glossiness = normalTexel.a;

    float4 foamTexel = foamTexture.sample(defaultSampler, in.foamTexCoord);
    float4 lightMapTexel = controlTexture.sample(defaultSampler, in.lightMapTexCoord);

    float foamMask = lightMapTexel.r;
    float shadowMask = lightMapTexel.g;
    float colorLerp = lightMapTexel.b;
    float opacityMod = lightMapTexel.a;

    float3 oceanColor = oceanTexel.rgb * mix(u.deepOceanBlendFactor, u.shallowOceanBlendFactor, colorLerp);

    float3 color = mix(oceanColor, reflectionColor, u.reflectionLevel * glossiness);

    float foamStrength = foamMask * foamTexel.r * u.foamMultiplier;
    color = mix(color, u.foamColor, min(0.92, foamStrength)) * float3(shadowMask);
    float opacity = mix(1.0, foamStrength, min(0.92, foamStrength)) * opacityMod;
    
    return float4(color, opacity);
}

// ###############################
// ## Fragment shader functions ##
// ###############################

fragment float4 texture_fragment(ColorInOut in [[stage_in]],
                                 constant uniforms_t &uniforms [[buffer(0)]],
                                 texture2d<float> diffuseTexture [[texture(0)]],
                                 sampler defaultSampler [[sampler(0)]]) {
    return diffuseTexture.sample(defaultSampler, in.texCoord0) * uniforms.diffuseColor;
}

fragment float4 diffuse_fragment(ColorInOut in [[stage_in]],
                                 constant uniforms_t &uniforms [[buffer(0)]],
                                 texture2d<float> diffuseTexture [[texture(0)]],
                                 texturecube<float> irradianceMap [[texture(6)]],
                                 sampler defaultSampler [[sampler(0)]]) {
    float3 DIR = in.normal.xyz * CUBEMAP_DIR_CORRECTION;
    float4 texel = diffuseTexture.sample(defaultSampler, in.texCoord0);
    float4 light = irradianceMap.sample(defaultSampler, DIR);
    return texel * light * uniforms.diffuseColor;
}

fragment float4 diffuse_bumped_fragment(ColorInOut in [[stage_in]],
                                        constant uniforms_t &uniforms [[buffer(0)]],
                                        texture2d<float> diffuseTexture [[texture(0)]],
                                        texture2d<float> normalTexture [[texture(1)]],
                                        texturecube<float> irradianceMap [[texture(6)]],
                                        sampler defaultSampler [[sampler(0)]]) {
    float3 normal = normalTexture.sample(defaultSampler, in.texCoord0).xyz;
    float3 normal_t = normal * 2.0 - 1.0;
    float3 N = normalize(matrix_float3x3(in.tangent.xyz, in.bitangent.xyz, in.normal.xyz) * normal_t);

    float3 DIR = N * CUBEMAP_DIR_CORRECTION;

    float4 texel = diffuseTexture.sample(defaultSampler, in.texCoord0);
    float4 light = irradianceMap.sample(defaultSampler, DIR);

    return texel * light * uniforms.diffuseColor;
}

fragment float4 lightmap_fragment(ColorInOut in [[stage_in]],
                                  constant uniforms_t &uniforms [[buffer(0)]],
                                  texture2d<float> diffuseTexture [[texture(0)]],
                                  texture2d<float> lightmapTexture [[texture(1)]],
                                  sampler defaultSampler [[sampler(0)]]) {
    float4 color = diffuseTexture.sample(defaultSampler, in.texCoord0_scrolling) * uniforms.diffuseColor;
    float4 lightmap = lightmapTexture.sample(defaultSampler, in.texCoord0);
    return color * lightmap;
}

fragment float4 godray_fragment(ColorInOut in [[stage_in]],
                                constant uniforms_t &uniforms [[buffer(0)]],
                                texture2d<float> diffuseTexture [[texture(0)]],
                                sampler defaultSampler [[sampler(0)]]) {
    float cameraDist = distance(in.position_view, float3(0.0, 0.0, 0.0));
    float opacity = fadeFactor(cameraDist, uniforms) * blinkNoise(uniforms.time, uniforms.noisePeriod, uniforms);
    return diffuseTexture.sample(defaultSampler, in.texCoord0) * float4(1, 1, 1, opacity);
}

fragment float4 brightpass_fragment(ColorInOut in [[stage_in]],
                                    constant brightnesspp_uniforms_t &uniforms [[buffer(0)]],
                                    texture2d<float> diffuseTexture [[texture(0)]],
                                    sampler defaultSampler [[sampler(0)]]) {
    float4 texel = diffuseTexture.sample(defaultSampler, in.texCoord0);
    float threshold = uniforms.brightness_threshold;
    return clamp((texel - float4(threshold)) / (1.0 - threshold), 0.0, 1.0);
}

fragment float4 glow_fragment(ColorInOut in [[stage_in]],
                              texture2d<float> diffuseTexture1 [[texture(0)]],
                              texture2d<float> diffuseTexture2 [[texture(1)]],
                              sampler defaultSampler [[sampler(0)]]) {
    float4 dst = diffuseTexture1.sample(defaultSampler, in.texCoord0);
    float4 src = diffuseTexture2.sample(defaultSampler, in.texCoord0);
    return min(dst + src, 1.0);
}

fragment float4 volumetric_fragment(ColorInOut in [[stage_in]],
                                    constant godraypp_uniforms_t &uniforms [[buffer(0)]],
                                    texture2d<float> diffuseTexture [[texture(0)]],
                                    sampler defaultSampler [[sampler(0)]]) {
    const float NUM_SAMPLES = 16.0;

    float4 color = diffuseTexture.sample(defaultSampler, in.texCoord0);

    for (int light = 0; light < NUM_LIGHTS; light++) {
        if (uniforms.shouldUseLights[light] > 0.5) {
            float2 lightPosition = float2(uniforms.lightsX[light], uniforms.lightsY[light]);
            float2 delta = in.texCoord0 - lightPosition;
            float2 texCoord = in.texCoord0;
            delta *= 1.0 / NUM_SAMPLES * 0.4;

            float decay = 1.0 * uniforms.lightsIntensity[light];
            float weight = 1.0 / NUM_SAMPLES;

            for (int i = 0; i < int(NUM_SAMPLES); i++) {
                texCoord -= delta;
                float4 sample = diffuseTexture.sample(defaultSampler, texCoord);
                sample *= decay * weight;
                color += sample;
//                decay *= exp((float(-i) + 2.0)/NUM_SAMPLES);
            }
        }
    }

    return clamp(color, 0.0, 1.0);
}

fragment float4 particle_fragment(ColorInOut in [[stage_in]],
                                  constant particle_uniforms_t &uniforms [[buffer(0)]],
                                  texture2d<float> diffuseTexture [[texture(0)]],
                                  sampler defaultSampler [[sampler(0)]]) {
    float4 texel = diffuseTexture.sample(defaultSampler, in.texCoord0);
    float3 colorMultiplier = mix(float3(1.0), texel.rgb, float(uniforms.multiplyColorWithRGB));
    return float4(mix(texel.rgb, in.color.rgb * colorMultiplier, float(uniforms.useColor)),
                  texel.a * in.color.a * uniforms.opacity) * uniforms.diffuseColor;
}

// #######################
// ## Utility functions ##
// #######################

matrix_float4x4 mat4(matrix_float3x3 m) {
    return matrix_float4x4(float4(m[0], 0),
                           float4(m[1], 0),
                           float4(m[2], 0),
                           float4(0, 0, 0, 1));
}

matrix_float3x3 mat3(matrix_float4x4 m) {
    return matrix_float3x3(m[0].xyz,
                           m[1].xyz,
                           m[2].xyz);
}

float fadeFactor(float cameraDist, constant uniforms_t &uniforms) {
    cameraDist = clamp(cameraDist, uniforms.nearDist, uniforms.farDist);
    return (cameraDist - uniforms.nearDist) / (uniforms.farDist - uniforms.nearDist);
}

float blinkNoise(float time, float period, constant uniforms_t &uniforms) {
    float noiseTime = time;//modf(time, period) * 2.0 * M_PI / period;
    float noise = sin(noiseTime) * (0.5 * cos(noiseTime * 0.6366 + 56.7272) + 0.5);
    return mix(uniforms.minOpacity, uniforms.maxOpacity, 0.5 * noise + 0.5);
}

matrix_float4x4 rotationInverse(matrix_float4x4 matrix) {
    matrix_float3x3 rotationScale = mat3(matrix);
    float scaleX = length(matrix[0]);
    float scaleY = length(matrix[1]);
    float scaleZ = length(matrix[2]);
    matrix_float3x3 scaleInverse = matrix_float3x3(float3(1.0/scaleX, 0, 0),
                                                   float3(0, 1.0/scaleY, 0),
                                                   float3(0, 0, 1.0/scaleZ));
    return mat4(transpose(rotationScale * scaleInverse));
}

float4 quaternionFromAxisAngle(float3 axis, float angle) {
    float halfAngle = angle / 2.0;
    float sinHalfAngle = sin(halfAngle);
    float cosHalfAngle = cos(halfAngle);
    return float4(axis * sinHalfAngle, cosHalfAngle);
}

float4 quaternionMultiply(float4 a, float4 b) {
    return float4(a.w * b.x + a.x * b.w + a.y * b.z - a.z * b.y,
                  a.w * b.y + a.y * b.w + a.z * b.x - a.x * b.z,
                  a.w * b.z + a.z * b.w + a.x * b.y - a.y * b.x,
                  a.w * b.w - a.x * b.x - a.y * b.y - a.z * b.z);
}

matrix_float4x4 mat4FromQuaternion(float4 q) {
    float4 col1 = float4(1.0 - 2.0 * q.y * q.y - 2.0 * q.z - 2.0 * q.z,
                         2.0 * q.x * q.y + 2.0 * q.w * q.z,
                         2.0 * q.x * q.z - 2.0 * q.w * q.y,
                         0.0);
    float4 col2 = float4(2.0 * q.x * q.y - 2.0 * q.w * q.z,
                         1.0 - 2.0 * q.x * q.x - 2.0 * q.z * q.z,
                         2.0 * q.y * q.z + 2.0 * q.w * q.x,
                         0.0);
    float4 col3 = float4(2.0 * q.x * q.z + 2.0 * q.w * q.y,
                         2.0 * q.y * q.z - 2.0 * q.w * q.x,
                         1.0 - 2.0 * q.x * q.x - 2.0 * q.y * q.y,
                         0.0);
    float4 col4 = float4(0.0, 0.0, 0.0, 1.0);
    return matrix_float4x4(col1, col2, col3, col4);
}

matrix_float4x4 billboarded_orientation_matrix(matrix_float4x4 modelMatrix, matrix_float4x4 viewMatrix, float3 normalInModel, float3 billboard_axis) {
    matrix_float4x4 rotationalInverse = rotationInverse(viewMatrix * modelMatrix);

    float3 cameraInCamera = float3(0.0, 0.0, -1.0);
    float3 cameraInModel = mat3(rotationalInverse) * cameraInCamera;

    float3 xInModel = cross(billboard_axis, normalInModel);
    float3 xNormInModel = normalize(xInModel);
    float thetaRadians = asin(min(1.0, length(xInModel) / (length(billboard_axis) * length(normalInModel))));

    float3 yInModel = cross(billboard_axis, cameraInModel);
    float3 aInModel = cross(yInModel, xInModel);
    float3 aNormInModel = normalize(aInModel);
    float dotA = dot(yInModel, xInModel);
    float omegaArcSinRadians = asin(min(1.0, length(aInModel) / (length(xInModel) * length(yInModel))));
    float omegaRadians = float(dotA < 0.0) * omegaArcSinRadians + float(dotA > 0.0) * (M_PI - omegaArcSinRadians);

    float4 thetaQuaternion = quaternionFromAxisAngle(xNormInModel, M_PI / 2.0 - thetaRadians);
    float4 omegaQuaternion = quaternionFromAxisAngle(aNormInModel, omegaRadians);
    float4 omegaThetaQuaternion = quaternionMultiply(omegaQuaternion, thetaQuaternion);
    matrix_float4x4 omegaThetaRotation = mat4FromQuaternion(omegaThetaQuaternion);
    
    return omegaThetaRotation;
}

float2 sanitize_time2(float originalTime, float2 multiplier) {
    float moddedTime = fmod(originalTime, 1.0 / (1e-5 + fmax(fabs(multiplier.x), fabs(multiplier.y))));
    return moddedTime * multiplier;
}
