--[[
The BaseDamage UnitAbility gives entites the ability to damage their own base.

@param entity        The entity whose base should take damage.
@param damageAmount  The amount to damage the base by.
--]]
function UnitAbility_baseDamage(entity, damageAmount)
   local teamNumber = objc.get_team_number(entity)
   local base = objc.get_base_entity(teamNumber)

   objc.inflict_direct_damage(entity, base, damageAmount)
end


--[[
The Charge UnitAbility gives entities the ability to speed up when they get close to a unit.

@param entity               The entity to give the charge ability to.
@param initialCooldown      The intial number of ticks before this entity can use Charge.
@param enemyRadius          The maximum distance an enemy unit can be before this entity uses Charge.
@param chargeMaxVelocity    The maximum velocity to use when using Charge
@param regularMaxVelocity   The regular velocity to use when not using Charge
--]]
function UnitAbility_charge(entity, initialCooldown, enemyRadius, chargeMaxVelocity, regularMaxVelocity)
   local targetKey = 'hasCharged'
   local tickTimer = objc.tick_timer(entity, 'chargeTickTime', initialCooldown)

   if tickTimer == 0 then
      if objc.is_targetable_enemy_in_radius(entity, enemyRadius) then
         --objc.set_combat_text(entity, 'Charging!')
         objc.set_entity_value(entity, 'velocityComponent.primaryMaxVelocity', chargeMaxVelocity)
         objc.set_local_value(entity, targetKey, 1)
      elseif objc.get_local_value(entity, targetKey) == 1 then
         --objc.set_combat_text(entity, '')
         objc.set_entity_value(entity, 'velocityComponent.primaryMaxVelocity', regularMaxVelocity)
      end
   end
end


--[[
The DamageCap UnitAbility caps damage given to an entity.

@param currentDamage The current damage inflicted.
@param damageCap     The damage cap.

@returns The amount of damage that should be inflicted on the entity.
--]]
function UnitAbility_damageCap(currentDamage, damageCap)
   if currentDamage > damageCap then
      currentDamage = damageCap
   end

   return currentDamage
end


--[[
The Mercenary UnitAbility makes it so that this unit only spawns once.

This is best used as part of a SpawnTrigger.

- @param factory        The factory that created the entity
- @param spawnedEntity  The entity that was just spawned
--]]
function UnitAbility_mercenary(factory, spawnedEntity)
   if objc.get_entity_value(factory, 'spawnQuantity') > 0 then
      objc.set_factory_spawn_number(spawnedEntity, objc.get_entity_value(factory, 'spawnQuantity') - 1)
   end
end


--[[
The Negate UnitAbility shields the entity for the first set of damage inflicted
upon it.

@param entity        The entity to have the Negate ability
@param currentDamage The current amount of damage being inflicted on the entity
@param negateLimit   The total number of attacks this entity should negate.

@returns The amount of damage that should be inflicted on the entity.
--]]
function UnitAbility_negate(entity, currentDamage, negateLimit)
   local key = 'negateCount'

   local negateCount = objc.get_local_value(entity, key)

   if negateCount < negateLimit then
      objc.set_local_value(entity, key, negateCount + 1)
      currentDamage = 0
      --objc.set_combat_text(entity, 'Blocked')
   else
      --objc.set_combat_text(entity, '')
   end

   return currentDamage
end


--[[
The Shield UnitAbility allows the ability to negate any damage below a given
threshold.

@param damage           The current damage given to the unit.
@param damageThreshold  The threshold below which damage turned to 0.

@returns The damage the should be applied to this unit.
--]]
function UnitAbility_shield(damage, damageThreshold)
      if damage < damageThreshold then
         damage = 0
      end

      return damage
end


--[[
This Slow UnitAbility slows down a given target entity by changing their max
velocity to a new speed if their current max velocity is above a given
threshold.

@param targetEntity           The entity to slow
@param maxVelocityThreshold   The max velocity threshold above which to slow the entity
@param slowMaxVelocity        The new max velocity if the entity is becoming slowed
--]]
function UnitAbility_slow(targetEntity, maxVelocityThreshold, slowMaxVelocity)
   if objc.is_normal_unit(targetEntity) then
      local moveSpeed = objc.get_entity_value_for_key(targetEntity, ENTITY_VALUE_KEY_MAX_VELOCITY)
      if moveSpeed > maxVelocityThreshold then
         moveSpeed = slowMaxVelocity
         objc.set_entity_value_for_key(targetEntity, ENTITY_VALUE_KEY_MAX_VELOCITY, moveSpeed)
         --objc.set_combat_text(targetEntity, 'Slowed')
      end
   end
end


--[[
The Stun UnitAbility stuns a given entity for a given number of subrounds.

Note: there are 1000 sub rounds in a second

@param entity                    The entity to stun.
@param numberOfSubRoundsToStun   The number of subrounds to stun the entity.
--]]
function UnitAbility_stun(entity, numberOfSubRoundsToStun)
   if objc.is_normal_unit(entity) and objc.get_local_value(entity, IS_DEBUFF_IMMUNE_KEY) == 0 then
      objc.stun_entity(entity, numberOfSubRoundsToStun)
      --objc.set_combat_text(entity, 'Stunned!')
   end
end


--[[
The Surge UnitAbility builds addition units the first time a unit spawns.

@param factory
@param spawnedEntityIdentifier                  The entity identifier of the entity to spawn
@param playerNumber                             The player's number
@param numberOfAdditionalUnitsToBuildPerSurge   The number of additional units to build the first time a unit spawns
--]]
function UnitAbility_surge(factory, spawnEntityIdentifier, playerNumber, numberOfAdditionalUnitsToBuildPerSurge)
   local currentSpawnedQuantity = objc.get_entity_value(factory, 'spawnQuantity')
   local key = 'tickTime' .. objc.get_entity_value(factory, 'description')

   local previousSpawnQuantity = objc.get_global_value(key)
   if currentSpawnedQuantity > previousSpawnQuantity then
      local surgeSpawnMultiplier =  currentSpawnedQuantity - previousSpawnQuantity
      objc.spawn(spawnEntityIdentifier, numberOfAdditionalUnitsToBuildPerSurge * surgeSpawnMultiplier, playerNumber)
   end

   objc.set_global_value(key, currentSpawnedQuantity)
end


--[[
The TrueDamage UnitAbility inflicts guaranteed damage on a unit.

@param targetEntity  The target entity to receive damage.
@param sourceEntity  The source entity dealing the damage.
--]]
function UnitAbility_trueDamage(targetEntity, sourceEntity)
   local parent = objc.get_entity_value(sourceEntity, 'identityComponent.sourceEntity')
   local guaranteedDamage = objc.get_entity_value(parent, 'aiComponent.damageComponent.attackDamage') + objc.get_entity_value_for_key(parent, ENTITY_VALUE_KEY_DAMAGE_BUFF)
   objc.inflict_direct_damage(sourceEntity, targetEntity, guaranteedDamage)
end


--[[
This ability is strictly for prototyping a "tombstone" type unit that activates on ability activation.
It is not intended for real use
]]
function UnitAbility_activateTombstones(playerNumber, teamNumber)
   local listOfEntities = objc.list_entities(teamNumber)
   local tombstoneTransformCount = 0
   local maxTransform = 50

   for _, entity in pairs(listOfEntities) do
      if objc.get_entity_value(entity, 'identityComponent.identifier') == 'W_Tombstone' then
         local tombstonePosition = objc.get_entity_position(entity)
         tombstoneTransformCount = tombstoneTransformCount + 1

         objc.kill_entity(entity)
         objc.spawn_at_point('W_Regenerator', 1, playerNumber, tombstonePosition[1], tombstonePosition[2], 0, 0)
      end

      if tombstoneTransformCount >= maxTransform then
         break
      end
   end
end


function UnitAbility_incomeLineMainLoop(entity)

   local teamNumber = objc.get_team_number(entity)
   local playerNumber = objc.get_player_number(entity)
   local enemyNumber = objc.get_enemy_team_number(entity)
   local enemyPlayerNumbers = objc.get_player_numbers_for_team_number(enemyTeamNumber)
   local entityPos = objc.get_entity_position(entity)

   --Increment the game time counter by 1 game round
   local heldTimeKey = 'heldTimeKey'
   local shouldCaptureKey = 'shouldBeCapturedKey'
   local lineEntityTypeKey = 'lineEntityTypeKey'
   local timeToCapture = 150
   local heldTime = objc.get_local_value(entity, heldTimeKey)
   local xPosForBuffEntity = 3000
   local xMod = 1

   if teamNumber == 1 then
      xMod = -1
   end

   if heldTime > timeToCapture then
      local lineEntityType = objc.get_local_value(entity, lineEntityTypeKey)
      objc.spawn_at_point(lineEntityType, 1, playerNumber, (xPosForBuffEntity * xMod), entityPos[2], 0, 0)
      objc.sacrifice_entity(entity)
   else
      objc.set_local_value(entity, heldTimeKey, heldTime + 1)
   end

   --Check to see if the crystal mine should change ownership
   local ownershipCheckInterval = 5
   local ownershipCheckTimer = objc.tick_timer(entity, 'ownershipCheckTimer', ownershipCheckInterval)
   if ownershipCheckTimer == 0 then
      local isOwnershipSame = UnitAbility_incomeLineSideCheck(entity, teamNumber, enemyNumber)

      if isOwnershipSame == false then
         local entityList = objc.list_entities(teamNumber)

         if teamNumber == 1 then
            objc.spawn_at_point('IncomeLineEntityRight', 1, enemyPlayerNumbers[1], 0, entityPos[2], 0, 0)
         elseif teamNumber == 2 then
            objc.spawn_at_point('IncomeLineEntityLeft', 1, enemyPlayerNumbers[1], 0, entityPos[2], 0, 0)
         end

         objc.kill_entity(entity)
      end
   end
end



function UnitAbility_incomeLine_spawnDoobers(entity)
   local currentCrystalMineRampLevelKey = 'currentCrystalMineRampLevelKey'
   local rampCrystalAmounts = {1,1,2,4,8}
   local currentCrystalMineRampLevel = objc.get_local_value(entity, currentCrystalMineRampLevelKey)

   if currentCrystalMineRampLevel >= listCount(rampCrystalAmounts) - 1 then
      currentCrystalMineRampLevel = listCount(rampCrystalAmounts) - 1
   end

   local numberOfCrystalDoobersToSpawn = rampCrystalAmounts[currentCrystalMineRampLevel + 1]
   local playerNumber = objc.get_player_number(entity)
   local xSpawnPoint = 0
   local ySpawnPoint = 700
   local sittingCrystalDooberList = objc.spawn_at_point('CrystalGainEffectEntity', numberOfCrystalDoobersToSpawn, playerNumber, xSpawnPoint, ySpawnPoint, 0, 0)
   objc.spawn_at_point('blueBar', 1, playerNumber, -520, 900, 1, 0)

   for i, doober in pairs(sittingCrystalDooberList) do
      if numberOfCrystalDoobersToSpawn > 1 then
         local spaceBetweenDoobers = 200
         local numberOfDooberBeingEvaluated = i
         local xStartOffset = 0
         local yStartOffset = ySpawnPoint
         local secondRowThreshold = 4

         if numberOfCrystalDoobersToSpawn == 2 then
            xStartOffset = xStartOffset - objc.divide_with_floor(spaceBetweenDoobers, 2)
         elseif numberOfCrystalDoobersToSpawn >= 4 then
            xStartOffset = xStartOffset - (objc.divide_with_floor(spaceBetweenDoobers, 2) + spaceBetweenDoobers)
         end

         if numberOfDooberBeingEvaluated > secondRowThreshold then
            yStartOffset = yStartOffset - spaceBetweenDoobers
            numberOfDooberBeingEvaluated = numberOfDooberBeingEvaluated - (secondRowThreshold)
         end

         objc.teleport_entity(doober, xStartOffset + spaceBetweenDoobers * (numberOfDooberBeingEvaluated - 1), yStartOffset)
      end
      objc.set_entity_value(doober, 'velocityComponent.primaryMaxVelocity', 0)
   end


   objc.set_local_value(entity, 'sittingCrystalDooberListKey', sittingCrystalDooberList)
   objc.set_local_value(entity, currentCrystalMineRampLevelKey, currentCrystalMineRampLevel + 1)
end


function UnitAbility_incomeLineSideCheck(entity, teamNumber, enemyNumber)
   local allies = objc.list_entities(teamNumber)
   local thisPos = objc.get_entity_position(entity)
   local sideModifier = 1
   local yModifier = 1

   if teamNumber == 1 then
      sideModifier = 1
   elseif teamNumber == 2 then
      sideModifier = -1
   else
      objc.scripting_error('invalid player number found for income line')
   end

   if thisPos[2] < 0 then
      yModifier = -1
   end

   for _, ally in pairs(allies) do
      if objc.is_normal_unit(ally) then
         local allyPosition = objc.get_entity_position(ally)
         if allyPosition[1] * sideModifier > 0 and (allyPosition[2] * yModifier) > 0 then
            return true
         end
      end
   end

   local enemies = objc.list_entities(enemyNumber)

   if teamNumber == 1 then
      sideModifier = -1
   elseif teamNumber == 2 then
      sideModifier = 1
   else
      objc.scripting_error('invalid player number found for income line')
   end

   for _, enemy in pairs(enemies) do
      if objc.is_normal_unit(enemy) then
         local enemyPosition = objc.get_entity_position(enemy)
         if enemyPosition[1] * sideModifier > 0 and (enemyPosition[2] * yModifier) > 0 then
            return false
         end
      end
   end

   return true
end



local HAS_SPAWNED_KEY = 'hasSpawnedMultiple'
local SHOULD_DUMMY_KILL_KEY = 'shouldMultipleDummyDie'
local SPAWN_CIRCLE_RADIUS = 50
local SEARCH_RADIUS = 5



function UnitAbility_spawnMultiple(entity, numToSpawn)
   if objc.get_local_value(entity, SHOULD_DUMMY_KILL_KEY) == 1 then
      objc.kill_entity(entity)
   end

   if objc.get_local_value(entity, HAS_SPAWNED_KEY) == 0 then
      local teamNumber = objc.get_team_number(entity)
      local playerNumber = objc.get_player_number(entity)
      local position = objc.get_entity_position(entity)
      local thisIdentifier = objc.get_entity_value(entity, 'identityComponent.identifier')
      local entities = objc.list_entities_in_radius(entity, SEARCH_RADIUS, teamNumber)

      --uncomment this to disable factories spawning multiple
      --[[for _, ally in pairs(entities) do
         local identifier = objc.get_entity_value(ally, 'identityComponent.identifier')
         if identifier == (thisIdentifier .. 'Spawner') then
            objc.set_local_value(entity, HAS_SPAWNED_KEY, 1)
            return
         end
      end]]

      local radBetweenUnits = objc.divide_with_floor((math.pi * 2), numToSpawn)
      local sideCoeff = 1
      if teamNumber == 2 then
         sideCoeff = -1
      end

      if numToSpawn == 2 then
         for i = 1, numToSpawn do
            local offsetFromSpawner = 175
            if i > 1 then
               offsetFromSpawner = offsetFromSpawner * (-1)
            end
            local newUnit = objc.spawn_at_point(thisIdentifier, 1, playerNumber, position[1], position[2] + offsetFromSpawner, 1 * sideCoeff, 0)
            objc.set_local_value(newUnit[1], HAS_SPAWNED_KEY, 1)
         end

      else
         for i = 1, numToSpawn do
            local thisRad = radBetweenUnits * i
            local x = position[1] + (math.cos(thisRad) * SPAWN_CIRCLE_RADIUS)
            local y = position[2] + (math.sin(thisRad) * SPAWN_CIRCLE_RADIUS)
            local newUnit = objc.spawn_at_point(thisIdentifier, 1, playerNumber, math.floor(position[1] + ((math.cos(thisRad) * SPAWN_CIRCLE_RADIUS) * sideCoeff)), math.floor(position[2] + (math.sin(thisRad) * SPAWN_CIRCLE_RADIUS * sideCoeff)), 1 * sideCoeff, 0)
            objc.set_local_value(newUnit[1], HAS_SPAWNED_KEY, 1)
         end
      end

      objc.set_local_value(entity, HAS_SPAWNED_KEY, 1)
      objc.teleport_entity(entity, -100000, 10000)
      objc.set_local_value(entity, SHOULD_DUMMY_KILL_KEY, 1)
      --objc.kill_entity(entity)
   end
end

function UnitAbility_barricadeDeathTrigger(entity)
   local tag = objc.get_local_value(entity, LUA_KEY_TAG)

   objc.set_global_value(tag .. "Destroyed", 1)
   objc.set_global_value("placementHasChanged", 1)

   local teamNumber = objc.get_team_number(entity)
   local teamEntities = objc.list_entities(teamNumber)
   for _,teamEntity in pairs(teamEntities) do
      if objc.is_miner(teamEntity) then
         objc.set_targetable(teamEntity, true)
      end
   end
end

function UnitAbility_noDamageTrigger(damage, targetEntity, sourceEntity)
    return 0
end

local TOTAL_HEALTH_TICK_TIME = 100
function UnitAbility_tickDownHealthTrigger(entity)
    local maxHealth = objc.get_entity_value_for_key(entity, ENTITY_VALUE_KEY_MAX_HEALTH)
    local currentHealth = objc.get_entity_value_for_key(entity, ENTITY_VALUE_KEY_CURRENT_HEALTH)
    local healthProp = objc.divide_with_floor(maxHealth, TOTAL_HEALTH_TICK_TIME)

    local healthToSet = currentHealth - healthProp
    objc.set_entity_value_for_key(entity, ENTITY_VALUE_KEY_CURRENT_HEALTH, healthToSet)
end
