local ATTACK_COUNTER_KEY = 'attackCounterKey'
local AOE_LINE_LENGTH = 4000
local AOE_LINE_WIDTH = 750
local BATTLEFIELD_LENGTH = 10000
local BATTLEFIELD_WIDTH = 4000
local AOE_DAMAGE_COEFFICIENT = 1.0
local SLOW_AMOUNT = 1 --as a percent, 1 = 100%
local SLOW_DURATION = 50
local SLOW_TIMER_KEY = 'slowTimerKey'
local HAS_INITIALIZED_SLOW_KEY = 'hasInitializedSlowKey'
local SLOW_SHADER_NAME = 'skinned_toon_freeze'
local ORIGINAL_SPEED_KEY = 'originalSpeedKey'
local DAMAGE_ENTITY_IDENTIFIER = 'W_DragonDamageScriptEntity'
local DAMAGE_ENTITY_DAMAGE_KEY = 'toDealDamageKey'
local DAMAGE_ENTITY_TARGET_KEY = 'damageEntityTargetKey'

local IS_VFX_ACTIVE_KEY = 'isVFXActive'
local VFX_SCENE_NAME = 'head_W_Dragon.lvl'
local VFX_NODE = 'node_head'
local VFX_BASE_OFFSET = 2000

local DRAGON_SUMMON_SPAWN_ANIMATION = 'W_Dragon_Air_Spawn_02'
local FRAMES_IN_SUMMON_SPAWN_ANIMATION = 30
local SUMMON_SPAWN_KEY = 'shouldSummonSpawn'

local DRAGON_SLOW_TRIGGER_NAME = 'W_Dragon_immobilizeTickTrigger'

--[[
This attack trigger sets a unique value per attack and assigns it to the target so we can track who the primary
target for each attack is and only apply the visual effect when that target is damaged (or else everything in the splash area
would spawn an effect)
]]--
W_Dragon_onAttackTrigger = function(sourceEntity, targetEntity)
	local primaryTargetKey = objc.get_entity_value(sourceEntity, 'identityComponent.description')
	local attackCounter = objc.get_local_value(sourceEntity, ATTACK_COUNTER_KEY)

	attackCounter = attackCounter + 1

	objc.set_local_value(targetEntity, primaryTargetKey, attackCounter)
	objc.set_local_value(sourceEntity, ATTACK_COUNTER_KEY, attackCounter)
end

W_Dragon_tickTrigger = function(entity)
	local shouldSummonSpawn = objc.get_local_value(entity, SUMMON_SPAWN_KEY)
	if shouldSummonSpawn == 1 then
		--If I've been cultist summoned, use the correct spawn animation
        local numberOfSubrounds = objc.convert_frames_to_subrounds(FRAMES_IN_SUMMON_SPAWN_ANIMATION)
        objc.idle_entity(entity, numberOfSubrounds)
        objc.play_override_animation(entity, DRAGON_SUMMON_SPAWN_ANIMATION, objc.subrounds_to_milliseconds(numberOfSubrounds))


		--No need to call this every time
		objc.set_local_value(entity, SUMMON_SPAWN_KEY, 0)
	end
end

--[[
This DamageTrigger gives W_Dragon the Slow UnitAbility.
--]]
W_Dragon_damageTrigger = function(damage, targetEntity, sourceEntity)
	local dragon = objc.get_entity_value(sourceEntity, 'identityComponent.sourceEntity')
	local primaryTargetKey = objc.get_entity_value(dragon, 'identityComponent.description')
	local attackCounter = objc.get_local_value(dragon, ATTACK_COUNTER_KEY)
	local isCurrentPrimaryTargetValueCheck = objc.get_local_value(targetEntity, primaryTargetKey)


	if attackCounter == isCurrentPrimaryTargetValueCheck then
		local playerNumber = objc.get_player_number(dragon)
		local enemyNumber = objc.get_team_number(targetEntity)
		local targetPosition = objc.get_entity_position(targetEntity)
		local dragonPosition = objc.get_entity_position(dragon)
		local enemyList = objc.list_entities(enemyNumber)
		local directionX = 0

		if dragonPosition[1] <= targetPosition[1] then
			directionX = 1
		else
			directionX = -1
		end

		for _, enemy in pairs(enemyList) do
			local enemyPosition = objc.get_entity_position(enemy)
			local distanceFromTargetX = (enemyPosition[1] + BATTLEFIELD_LENGTH) - (targetPosition[1] + BATTLEFIELD_LENGTH)
			distanceFromTargetX = distanceFromTargetX * directionX
			local distanceFromTargetY = math.abs((enemyPosition[2] + BATTLEFIELD_WIDTH) - (targetPosition[2] + BATTLEFIELD_WIDTH))

			if distanceFromTargetX > -100 
				and distanceFromTargetX < AOE_LINE_LENGTH 
				and distanceFromTargetY < objc.divide_with_floor(AOE_LINE_WIDTH, 2) then
				
				--if it's a ground unit, damage it	
				if objc.is_ground_unit(enemy) then
					local damageEntity = objc.spawn(DAMAGE_ENTITY_IDENTIFIER, 1, playerNumber)
					objc.set_local_value(damageEntity[1], DAMAGE_ENTITY_DAMAGE_KEY, math.ceil(damage * AOE_DAMAGE_COEFFICIENT))
					objc.set_local_value(damageEntity[1], DAMAGE_ENTITY_TARGET_KEY, enemy)

					--if it's a normal unit, slow that bad boy down
					if objc.is_normal_unit(enemy) or objc.is_commander(enemy) then
						local tickTriggers = objc.lua_functions_for_trigger(enemy, TRIGGER_NAME_TICK_TRIGGER)
						local hasImmobilizeTrigger = false
						for _, trigger in pairs(tickTriggers) do
							if string.match(trigger, 'W_Dragon_immobilizeTickTrigger') then
								hasImmobilizeTrigger = true
							end
						end

						objc.set_local_value(enemy, SLOW_TIMER_KEY, SLOW_DURATION)
						if hasImmobilizeTrigger == false then
							objc.add_trigger(enemy, TRIGGER_NAME_TICK_TRIGGER, 'W_Dragon.lua', DRAGON_SLOW_TRIGGER_NAME)
						end
					end
				end
				
			end
		end

		local enemyBaseEntity = objc.get_base_entity(enemyNumber)
		local baseOffset = 0
		if targetEntity == enemyBaseEntity then
			baseOffset = VFX_BASE_OFFSET * directionX
		end
		objc.spawn_at_point('W_DragonTrailEffect', 1, playerNumber, targetPosition[1] - baseOffset, targetPosition[2], directionX, 0)

	end
	return 0
end


W_Dragon_immobilizeTickTrigger = function(entity)
	local hasInitSlow = objc.get_local_value(entity, HAS_INITIALIZED_SLOW_KEY)
	if hasInitSlow == 0 then
		local originalMaxVel = objc.get_entity_value_for_key(entity, ENTITY_VALUE_KEY_MAX_VELOCITY)
		objc.apply_shader(entity, SLOW_SHADER_NAME)
		objc.set_local_value(entity, ORIGINAL_SPEED_KEY, originalMaxVel)
		objc.set_entity_value_for_key(entity, ENTITY_VALUE_KEY_MAX_VELOCITY, 3)
		objc.set_local_value(entity, HAS_INITIALIZED_SLOW_KEY, 1)
	end

	local slowTimerValue = objc.get_local_value(entity, SLOW_TIMER_KEY)
	if slowTimerValue <= 0 then
		local originalMaxVel = objc.get_local_value(entity, ORIGINAL_SPEED_KEY)
		objc.set_local_value(entity, HAS_INITIALIZED_SLOW_KEY, 0)
		objc.set_entity_value_for_key(entity, ENTITY_VALUE_KEY_MAX_VELOCITY, originalMaxVel)
		objc.remove_shader(entity, SLOW_SHADER_NAME)
		objc.remove_trigger(entity, TRIGGER_NAME_TICK_TRIGGER, 'W_Dragon_immobilizeTickTrigger')
		objc.idle_entity(entity, 1)
	else
		objc.set_local_value(entity, SLOW_TIMER_KEY, slowTimerValue - 1)
	end
end
