#include "texture_utility.sh"

#if defined(BORDER_ALPHA_CHANNEL_METHOD) || (__VERSION__ < 300) || defined(BORDER_DIRECT_ZBUFFER)
    #define BORDER_SAMPLER lowp sampler2D
#else
    #define BORDER_SAMPLER lowp isampler2D
#endif


//
// Border parameters
//      u_borderParams.x is the slope factor, which determines how great the slope in the depth
//          buffer needs to be in order to generate a border. There are no clear units for this
//          parameter, due to optimisation and approximations in the shader. The final result will
//          be dependent on the near and far clip planes for the camera.
//          Adjust to taste!
//
uniform vec4 BorderUniforms_borderParams;
uniform vec4 BorderUniforms_borderColor;
uniform vec2 BorderUniforms_reciprocalInputTextureDims;
uniform vec2 BorderUniforms_fragCoordYTransform;
uniform int BorderUniforms_thickBorderWeight;

vec2 reciprocalInputTextureDims() { return BorderUniforms_reciprocalInputTextureDims.xy; }
vec4 getBorderParams() { return BorderUniforms_borderParams; }
vec4 getBorderColor() { return BorderUniforms_borderColor; }
vec2 getFragCoordYTransform() { return BorderUniforms_fragCoordYTransform; }
int getThickBorderWeight() { return BorderUniforms_thickBorderWeight; }

vec2 inputPixelCoord() {
    // Due to frame buffer management, in some cases the border texture gets drawn upside-down.
    // Effectively, the viewport has a negative height value. We must reverse that transform here.
    // u_reciprocalInputTextureDims.w-gl_FragCoord.y
    return vec2(gl_FragCoord.x, getFragCoordYTransform().x + getFragCoordYTransform().y * gl_FragCoord.y);
}

////////////////////////////////////////////////////////////////////////////////////////////////////

float sampleForBorderFloat(highp sampler2D depthSampler, vec2 coord) {
#if defined(BORDER_DIRECT_ZBUFFER)
    return sampleLevelZero(depthSampler, coord).x;
#else
    return sqrt(float(sampleLevelZero(depthSampler, coord).x) / 65535.0);
#endif
}

float borderEstimateFloat(highp sampler2D depthSampler, vec2 baseCoord) {

    vec2 rViewport = reciprocalInputTextureDims();
    baseCoord = rViewport * baseCoord;      // (multiply out first to allow MADD operations below)

    //#define BORDER_USE_PLUS_PATTERN
    #if defined(BORDER_USE_PLUS_PATTERN)
        // This is Erik's original iteration, that uses a "+" pattern and five taps
        // It's designed to work with bilinear enabled
        vec2 offset1X = vec2(1.3846153846, 0.0);
        vec2 offset2X = vec2(3.2307692308, 0.0);
        
        vec2 offset1Y = vec2(0.0, 1.3846153846);
        vec2 offset2Y = vec2(0.0, 3.2307692308);
        
        float depth = sampleForBorderFloat(depthSampler, baseCoord + vec2(0.0, 0.0));
        
        float depth_lX = sampleForBorderFloat(depthSampler, baseCoord + offset1X * rViewport);
        float depth_dX = sampleForBorderFloat(depthSampler, baseCoord + offset2X * rViewport);
        
        float depth_lY = sampleForBorderFloat(depthSampler, baseCoord + offset1Y * rViewport);
        float depth_dY = sampleForBorderFloat(depthSampler, baseCoord + offset2Y * rViewport);
        
        float borderX = 1.0 - min(1.0, abs(depth_lX + depth_dX - depth * 2.0) * getBorderParams().x);
        float borderY = 1.0 - min(1.0, abs(depth_lY + depth_dY - depth * 2.0) * getBorderParams().x);
    #else
        // This is slightly tweaked version with an "X" pattern and 4 taps
        // It's designed to work with bilinear disabled
        vec2 offset1X = vec2((-1.0) * rViewport.x, (-1.0) * rViewport.y);
        vec2 offset2X = vec2((+1.0) * rViewport.x, (+1.0) * rViewport.y);
        
        vec2 offset1Y = vec2((-1.0) * rViewport.x, (+1.0) * rViewport.y);
        vec2 offset2Y = vec2((+1.0) * rViewport.x, (-1.0) * rViewport.y);
        
        float depth_lX = sampleForBorderFloat(depthSampler, baseCoord + offset1X);
        float depth_dX = sampleForBorderFloat(depthSampler, baseCoord + offset2X);
        
        float depth_lY = sampleForBorderFloat(depthSampler, baseCoord + offset1Y);
        float depth_dY = sampleForBorderFloat(depthSampler, baseCoord + offset2Y);
        
        float borderX = 1.0 - min(1.0, abs(depth_lX - depth_dX) * getBorderParams().x);
        float borderY = 1.0 - min(1.0, abs(depth_lY - depth_dY) * getBorderParams().x);
    #endif
    
    return 1.0 - (borderX * borderY);
}

////////////////////////////////////////////////////////////////////////////////////////////////////

#if __VERSION__ >= 300

int sampleForBorderInt(BORDER_SAMPLER depthSampler, vec2 coord) {
#if defined(BORDER_ALPHA_CHANNEL_METHOD)
    return int(float(sampleLevelZero(depthSampler, coord).a) * 255.0);
#else
    return sampleLevelZero(depthSampler, coord).x;
#endif
}

float borderEstimateInt(BORDER_SAMPLER depthSampler, vec2 baseCoord) {

    vec2 rViewport = reciprocalInputTextureDims();
    baseCoord = rViewport * baseCoord;      // (multiply out first to allow MADD operations below)

    // This is slightly tweaked version with an "X" pattern and 4 taps
    // It's designed to work with bilinear disabled
    vec2 offset1X = vec2((-1.0) * rViewport.x, (-1.0) * rViewport.y);
    vec2 offset2X = vec2((+1.0) * rViewport.x, (+1.0) * rViewport.y);
    
    vec2 offset1Y = vec2((-1.0) * rViewport.x, (+1.0) * rViewport.y);
    vec2 offset2Y = vec2((+1.0) * rViewport.x, (-1.0) * rViewport.y);
    
    int depth_lX = sampleForBorderInt(depthSampler, baseCoord + offset1X);
    int depth_dX = sampleForBorderInt(depthSampler, baseCoord + offset2X);
    
    int depth_lY = sampleForBorderInt(depthSampler, baseCoord + offset1Y);
    int depth_dY = sampleForBorderInt(depthSampler, baseCoord + offset2Y);
    
    float borderX = 1.0 - min(1.0, float(abs(depth_lX - depth_dX)) * getBorderParams().y);
    float borderY = 1.0 - min(1.0, float(abs(depth_lY - depth_dY)) * getBorderParams().y);

    return 1.0 - (borderX * borderY);
}

float borderEstimateIntThick(BORDER_SAMPLER depthSampler, vec2 baseCoord) {

    vec2 rViewport = reciprocalInputTextureDims();
    baseCoord = rViewport * baseCoord;      // (multiply out first to allow MADD operations below)

    // This is slightly tweaked version with an "X" pattern and "+" pattern,
    // with 8 taps. We have some control over the thickness with this version.
    // It's designed to work with bilinear disabled
    vec2 offset1X = vec2((-1.0) * rViewport.x, (-1.0) * rViewport.y);
    vec2 offset2X = vec2((+1.0) * rViewport.x, (+1.0) * rViewport.y);
    vec2 offset1Y = vec2((-1.0) * rViewport.x, (+1.0) * rViewport.y);
    vec2 offset2Y = vec2((+1.0) * rViewport.x, (-1.0) * rViewport.y);

    int depth1X = sampleForBorderInt(depthSampler, baseCoord + offset1X);
    int depth2X = sampleForBorderInt(depthSampler, baseCoord + offset2X);

    int depth1Y = sampleForBorderInt(depthSampler, baseCoord + offset1Y);
    int depth2Y = sampleForBorderInt(depthSampler, baseCoord + offset2Y);

    vec2 offsetS0 = vec2((-2.0) * rViewport.x, 0.0);
    vec2 offsetS1 = vec2(0.0, (-2.0) * rViewport.y);
    vec2 offsetS2 = vec2((+2.0) * rViewport.x, 0.0);
    vec2 offsetS3 = vec2(0.0, (+2.0) * rViewport.y);

    int depthS0 = sampleForBorderInt(depthSampler, baseCoord + offsetS0);
    int depthS1 = sampleForBorderInt(depthSampler, baseCoord + offsetS1);
    int depthS2 = sampleForBorderInt(depthSampler, baseCoord + offsetS2);
    int depthS3 = sampleForBorderInt(depthSampler, baseCoord + offsetS3);

#if 1
    float borderX = 1.0 - min(1.0, float(abs(depth1X - depth2X)) * getBorderParams().y);
    float borderY = 1.0 - min(1.0, float(abs(depth1Y - depth2Y)) * getBorderParams().y);
    float borderS0 = 1.0 - min(1.0, float(abs(depthS0 - depthS2)) * getBorderParams().y / float(getThickBorderWeight()));
    float borderS1 = 1.0 - min(1.0, float(abs(depthS1 - depthS3)) * getBorderParams().y / float(getThickBorderWeight()));

    return 1.0 - (borderX * borderY * borderS0 * borderS1);
#else
    int bx = (depth2Y + depth2X + depthS2/getThickBorderWeight()) - (depth1X + depth1Y + depthS0/getThickBorderWeight());
    int by = (depth1Y + depth2X + depthS3/getThickBorderWeight()) - (depth1X + depth2Y + depthS1/getThickBorderWeight());

    float borderX = 1.0 - min(1.0, float(abs(bx)) * getBorderParams().y);
    float borderY = 1.0 - min(1.0, float(abs(by)) * getBorderParams().y);

    return 1.0 - (borderX * borderY);
#endif
}

#endif
