
#if !defined(CONTACT_SOFT_SHADOW_SHADER)
#define CONTACT_SOFT_SHADOW_SHADER

float visibilityFunc(vec3 shadowCoord)
{
    return ShadowVisibilitySingleTap(shadowCoord.xy, shadowCoord.z);
}

float samplePCF3(vec3 shadowCoord)
{
    float result = 0.0;
    for (int x = -1; x <= 1; x++) {
        for (int y = -1; y <= 1; y++) {
            result += visibilityFunc(shadowCoord + vec3(x, y, 0.0) / float(GetShadowMapSize().x));
        }
    }
    return result / 9.0;
}

float samplePCF5(vec3 shadowCoord)
{
    float result = 0.0;
    for (int x = -2; x <= 2; x++) {
        for (int y = -2; y <= 2; y++) {
            result += visibilityFunc(shadowCoord + vec3(x, y, 0.0) / float(GetShadowMapSize().x));
        }
    }
    return result / 25.0;
}

float samplePCF7(vec3 shadowCoord)
{
    float result = 0.0;
    for (int x = -3; x <= 3; x++) {
        for (int y = -3; y <= 3; y++) {
            result += visibilityFunc(shadowCoord + vec3(x, y, 0.0) / float(GetShadowMapSize().x));
        }
    }
    return result / 49.0;
}

float samplePCF9(vec3 shadowCoord)
{
    float result = 0.0;
    for (int x = -4; x <= 4; x++) {
        for (int y = -4; y <= 4; y++) {
            result += visibilityFunc(shadowCoord + vec3(x, y, 0.0) / float(GetShadowMapSize().x));
        }
    }
    return result / 81.0;
}


float samplePCFQuantized(vec3 shadowCoord, float softness)
{
    if (softness <= 0.3)
        return samplePCF3(shadowCoord);
    else if (softness <= 0.7)
        return samplePCF5(shadowCoord);
    else if (softness <= 0.95)
        return samplePCF7(shadowCoord);
    else
        return samplePCF9(shadowCoord);
}

float evaluateContactHardeningShadow(vec3 shadowCoord, out float penumbraRatio)
{
    const int kernelSize = 7;
    const int blockerSearchRange = (kernelSize - 1) / 2;
    const float maxBlockers = float(kernelSize + 1) * float(kernelSize + 1);
    float shadowMapStep = 1.0 / float(GetShadowMapSize().x);
    float numBlockers = 0.0;
    float avgBlockerDepth = 0.0;

    for (int row = -blockerSearchRange; row <= blockerSearchRange; row += 2) {
        vec2 samplingCoord = shadowCoord.xy + vec2(-blockerSearchRange, row) * shadowMapStep;
        for (int col = -blockerSearchRange; col <= blockerSearchRange; col += 2) {
            
            // Nvidia and AMD's original implementations of CHS uses hardware gather4 command (textureGather()
            // in desktop OpenGL), and the overall logic is like this:
            //
            // vec4 depth4 = vec4(sampleLevelZero(u_shadowMap, tc + vec2(0, 0)).r,
            //                    sampleLevelZero(u_shadowMap, tc + vec2(0, shadowMapStep)).r,
            //                    sampleLevelZero(u_shadowMap, tc + vec2(shadowMapStep, 0)).r,
            //                    sampleLevelZero(u_shadowMap, tc + vec2(shadowMapStep, shadowMapStep)).r);
            // vec4 blocked4 = vec4(greaterThan(shadowCoord.zzzz, depth4)); // component wise comparison
            // numBlockers += dot(blocked4, vec4(1.0));
            // avgBlockerDepth += dot(depth4, blocked4);
            //
            // Since gather4 is not supported on mobile, we're approximating it with a single depth comparison.

            float depth = sampleLevelZero(u_shadowMap, samplingCoord).r;
            if (shadowCoord.z > depth) {
                // this sample "represents" 4 texels on the shadow map
                numBlockers += 4.0;
                avgBlockerDepth += 4.0 * depth;
            }

            samplingCoord.x += 2.0 * shadowMapStep;
        }
    }

    if (numBlockers == 0.0) // fully lit
        return 1.0;
    else if (numBlockers == maxBlockers) // fully shadowed
        return 0.0;

    avgBlockerDepth /= numBlockers;

    // compute ratio using formulas from PCSS article
    // http://developer.download.nvidia.com/shaderlibrary/docs/shadow_PCSS.pdf
    penumbraRatio = pow(clamp((shadowCoord.z - avgBlockerDepth) * GetShadowLightSize() / avgBlockerDepth, 0.0, 1.0), 1.2);

    return samplePCFQuantized(shadowCoord, penumbraRatio);
}

#endif // CONTACT_SOFT_SHADOW_SHADER
