#define SUPPORT_GLSL300
#include "prefix_tow.sh"

#if defined(GL_ES)
    precision highp float;
#endif

ATTRIBUTE vec4 a_cc3Position;
ATTRIBUTE vec4 a_cc3Normal;
ATTRIBUTE vec4 a_cc3Color;
ATTRIBUTE vec2 a_cc3TexCoord0;
ATTRIBUTE vec3 a_cc3Tangent;
#ifdef LIGHTMAP_ENABLED
ATTRIBUTE vec2 a_cc3TexCoord1;
#endif

// uniforms
uniform vec4 u_cc3MaterialDiffuseColor;

uniform float waveMaxHeight;
uniform float waveLength;
uniform float waveFrequency;

#define NUM_NORMAL_SOURCES 1
#define NUM_EXPLOSION_SOURCES 3
#define MAX_WIND_SOURCE_DISTANCE 225.0
#define RIPPLE_FREQUENCY 7.0
#define RIPPLE_AMPLITUDE 25.0

uniform int numActiveWindSources;
uniform int numActiveWindExplosionSources;
uniform vec4 windSources[NUM_NORMAL_SOURCES];
uniform float windIntensities[NUM_NORMAL_SOURCES];
uniform float windMaxDistance[NUM_NORMAL_SOURCES];
uniform vec4 windExplosionSources[NUM_EXPLOSION_SOURCES];
uniform float windExplosionIntensities[NUM_EXPLOSION_SOURCES];
uniform float windExplosionPercentComplete[NUM_EXPLOSION_SOURCES];
uniform float windExplosionMaxDistance[NUM_EXPLOSION_SOURCES];

// varyings
VARYING_OUT vec2 v_texCoord;
VARYING_OUT vec3 v_position_world;
VARYING_OUT vec4 v_material_diffuse_color;

#ifdef LIGHTMAP_ENABLED
VARYING_OUT vec2 v_lightMapCoord;
#endif

#if defined(FRESNEL_ENABLED)
VARYING_OUT vec3 v_N;
#endif

#include "shadow.vsh"
#include "system_uniforms.sh"

vec4 quaternionFromAxisAngle(vec3 axis, float angleRadians) {
    float halfAngle = angleRadians / 2.0;
    float sinHalfAngle = sin(halfAngle);
    float cosHalfAngle = cos(halfAngle);
    return vec4(axis * sinHalfAngle, cosHalfAngle);
}

vec3 rotate_vector(vec4 quat, vec3 vec)
{
    return (vec + 2.0 * cross(cross(vec, quat.xyz) + quat.w * vec, quat.xyz));
}

float windStrengthUsingDistance(vec4 windSource, float maxDistance) {
    return clamp((maxDistance - length(windSource)) / maxDistance, 0.0, 0.7);
}

float windStrength(vec4 windSource) {
    return windStrengthUsingDistance(windSource, MAX_WIND_SOURCE_DISTANCE);
}

float blinkNoise(float time, float period) {
    float noiseTime = mod(time, period) * 2.0 * 3.14159 / period;
    float noise = sin(noiseTime) * (0.5 * cos(noiseTime * 0.6366 + 56.7272) + 0.5);
    return mix(0.0, a_cc3Color.r, 0.5 * noise + 0.5);
}

void main() {
    v_texCoord = a_cc3TexCoord0;
    vec4 global_position = u_cc3MatrixModel * a_cc3Position;
    vec4 center_position = u_cc3MatrixModel * vec4(a_cc3Tangent, 1.0);

    vec4 position_model = a_cc3Position;
    float startingLength = length(global_position);

    //if you call normalize on a zero vector, glsl is a sad puppy
    vec4 windDirection = vec4(0.0, 0.0, 0.0, 0.0);
    vec4 windDirectionFromSource;
    float ripple = 0.0;
    float sourceInRangeFloat = 0.0;

    /*
    //Wind Debug - please do not delete
    ripple = 1.0;
    windDirectionFromSource = (vec4(0.0, 0.0, 0, 1) - center_position);
    windDirectionFromSource.y = 0.0;
    float distanceIntensity = windStrength(windDirectionFromSource);
    windDirection = windDirection + (2.0 *
                                     normalize(windDirectionFromSource) *
                                     distanceIntensity);
    sourceInRangeFloat = float((length(windDirectionFromSource) <= MAX_WIND_SOURCE_DISTANCE));// && numActiveWindSources > 0);
    */

    // Dynamic Wind Sources
    if (numActiveWindSources > 0) {
        windDirectionFromSource = windSources[0] - center_position;
        sourceInRangeFloat = float(length(windDirectionFromSource.xz) <= windMaxDistance[0]);
        windDirection = windDirection + (sourceInRangeFloat *
                                         windIntensities[0] *
                                         normalize(windDirectionFromSource));
    }

#define RIPPLE_SIZE 0.4

    // Static Decaying Wind Sources
    for (int i = 0; i < numActiveWindExplosionSources; i++) {
        float maxDistance = windExplosionMaxDistance[i];
        windDirectionFromSource = windExplosionSources[i] - global_position;
        float distanceIntensity = windStrengthUsingDistance(windDirectionFromSource, maxDistance);
        float distanceFromCenter = length(windDirectionFromSource.xz);
        bool shouldShowRipple = (distanceFromCenter >= maxDistance * windExplosionPercentComplete[i]) && (distanceFromCenter <= maxDistance * (windExplosionPercentComplete[i] + RIPPLE_SIZE));
        float shouldUseRippleFloat = float(shouldShowRipple == true);
        windDirection = windDirection + (normalize(windDirectionFromSource) *
                                         distanceIntensity *
                                         shouldUseRippleFloat);
        sourceInRangeFloat = max(sourceInRangeFloat, shouldUseRippleFloat);
        float wavePercent = distanceFromCenter / maxDistance;
        float percentDone = (wavePercent - windExplosionPercentComplete[i]) / RIPPLE_SIZE;
        ripple += windExplosionIntensities[i] * shouldUseRippleFloat * min(sin(percentDone * 2.0 * 3.14), 0.0);
    }

    // Generate wind and time noise
    float time = mod(getSceneTime(), 3.14159 * (2.0 * waveFrequency + 4.0 + 1.0));
    float movementBaseNumber = 2.0 * getSceneTime() + 10.0 * 3.14159;
    float movementRatio = 0.25 * (2.0 + cos(movementBaseNumber) + cos(movementBaseNumber / 2.0));
    float windSpeed = sourceInRangeFloat * 2.0 * min(length(windDirection), 2.0);
    float noiseWeight = mix(0.0, waveMaxHeight, a_cc3Color.b);
    float movementWeight = mix(0.0, waveMaxHeight, a_cc3Color.r + a_cc3Color.b);
    float dist = length(position_model);
    float globalLength = length(global_position);
    float phase = mod(((waveLength * dist) + (2.0 * waveFrequency + windSpeed + movementRatio / 2.0) * time), 3.14159 * 2.0);

    vec3 norm;
    vec4 position_world;
    vec4 positional_noise;
    vec4 quat;

    norm = normalize(center_position.xyz);
#if defined(FRESNEL_ENABLED)
    v_N = norm;
#endif

    float x = 0.7 * (movementWeight + noiseWeight) * (windDirection.x * norm.x + sin(phase));
    float y = 0.7 * (movementWeight + noiseWeight) * (windDirection.y * norm.y + sin(phase));
    float z = 0.7 * (movementWeight + noiseWeight) * (windDirection.z * norm.z + sin(phase));

    quat.xyzw = normalize(vec4(x, y, z, 1.0));
    positional_noise.xyz = rotate_vector((quat), (normalize(a_cc3Normal.xyz)));
    positional_noise.w = 0.0;
    windDirection.y = 0.0;
    windDirection.w = 0.0;

    // Add wind movement and noise movement
    if (sourceInRangeFloat > 0.5) {
        global_position = global_position - 0.8 * blinkNoise(time, 10.0) * movementRatio * (windSpeed * movementWeight * windDirection / 2.0);
        global_position = global_position - 0.2 * (ripple * windSpeed * movementWeight * windDirection / 2.0);
        float stepLength = length(center_position);
        float movePhase = mod(((waveLength * stepLength) - (2.0 * waveFrequency) * time), 3.14159 * 2.0);
        global_position = global_position + movementWeight * movementRatio * 0.3 * vec4(windDirection.x * sin(movePhase), 0.0, windDirection.z * sin(movePhase), 0.0);
    }
    global_position = global_position - 2.0 * (u_cc3MatrixModel * (noiseWeight * float(a_cc3Color.b > 0.0) * normalize(positional_noise)));
    global_position = startingLength * normalize(vec4(global_position.xyz, 1.0));
    v_position_world = global_position.xyz / global_position.w;
    gl_Position = getWorldToProj() * global_position;

    v_material_diffuse_color = u_cc3MaterialDiffuseColor;

#ifdef LIGHTMAP_ENABLED
    v_lightMapCoord = a_cc3TexCoord1;
#endif

    shadow_receive(global_position);
}
