#if defined(GL_ES)
    precision highp float;
#endif

attribute vec4 a_cc3Position;
attribute vec2 a_cc3TexCoord0;
attribute vec4 a_cc3Color;
attribute vec3 a_cc3Normal;

// uniforms
uniform mat4 u_cc3MatrixModel;
uniform mat4 u_cc3MatrixView;
uniform mat4 u_cc3MatrixProj;

// varyings
varying vec2 v_texCoord;
varying vec3 v_position_view;

#define M_PI 3.1415926535897932384626433832795

#ifdef GL_ES
mat3 transpose(mat3 matrix) {
    vec3 A = matrix[0];
    vec3 B = matrix[1];
    vec3 C = matrix[2];
    return mat3(vec3(A.x, B.x, C.x),
                vec3(A.y, B.y, C.y),
                vec3(A.z, B.z, C.z));
}
#endif

mat4 rotationInverse(mat4 matrix) {
    mat3 rotationScale = mat3(matrix);
    float scaleX = length(matrix[0]);
    float scaleY = length(matrix[1]);
    float scaleZ = length(matrix[2]);
    mat3 scaleInverse = mat3(vec3(1.0/scaleX, 0, 0),
                             vec3(0, 1.0/scaleY, 0),
                             vec3(0, 0, 1.0/scaleZ));
    return mat4(transpose(rotationScale * scaleInverse));
}

vec4 quaternionFromAxisAngle(vec3 axis, float angle) {
    float halfAngle = angle / 2.0;
    float sinHalfAngle = sin(halfAngle);
    float cosHalfAngle = cos(halfAngle);
    return vec4(axis * sinHalfAngle, cosHalfAngle);
}

vec4 quaternionMultiply(vec4 a, vec4 b) {
    return vec4(a.w * b.x + a.x * b.w + a.y * b.z - a.z * b.y,
                a.w * b.y + a.y * b.w + a.z * b.x - a.x * b.z,
                a.w * b.z + a.z * b.w + a.x * b.y - a.y * b.x,
                a.w * b.w - a.x * b.x - a.y * b.y - a.z * b.z);
}

mat4 mat4FromQuaternion(vec4 q) {
    vec4 col1 = vec4(1.0 - 2.0 * q.y * q.y - 2.0 * q.z - 2.0 * q.z,
                     2.0 * q.x * q.y + 2.0 * q.w * q.z,
                     2.0 * q.x * q.z - 2.0 * q.w * q.y,
                     0.0);
    vec4 col2 = vec4(2.0 * q.x * q.y - 2.0 * q.w * q.z,
                     1.0 - 2.0 * q.x * q.x - 2.0 * q.z * q.z,
                     2.0 * q.y * q.z + 2.0 * q.w * q.x,
                     0.0);
    vec4 col3 = vec4(2.0 * q.x * q.z + 2.0 * q.w * q.y,
                     2.0 * q.y * q.z - 2.0 * q.w * q.x,
                     1.0 - 2.0 * q.x * q.x - 2.0 * q.y * q.y,
                     0.0);
    vec4 col4 = vec4(0.0, 0.0, 0.0, 1.0);
    return mat4(col1, col2, col3, col4);
}

mat4 billboarded_orientation_matrix(mat4 modelMatrix, mat4 viewMatrix, vec3 normalInModel, vec3 billboard_axis) {
    mat4 rotationalInverse = rotationInverse(viewMatrix * modelMatrix);

    vec3 cameraInCamera = vec3(0.0, 0.0, -1.0);
    vec3 cameraInModel = mat3(rotationalInverse) * cameraInCamera;

    vec3 xInModel = cross(billboard_axis, normalInModel);
    vec3 xNormInModel = normalize(xInModel);
    float thetaRadians = asin(min(1.0, length(xInModel) / (length(billboard_axis) * length(normalInModel))));

    vec3 yInModel = cross(billboard_axis, cameraInModel);
    vec3 aInModel = cross(yInModel, xInModel);
    vec3 aNormInModel = normalize(aInModel);
    float dotA = dot(yInModel, xInModel);
    float omegaArcSinRadians = asin(min(1.0, length(aInModel) / (length(xInModel) * length(yInModel))));
    float omegaRadians = float(dotA < 0.0) * omegaArcSinRadians + float(dotA > 0.0) * (M_PI - omegaArcSinRadians);

    vec4 thetaQuaternion = quaternionFromAxisAngle(xNormInModel, M_PI / 2.0 - thetaRadians);
    vec4 omegaQuaternion = quaternionFromAxisAngle(aNormInModel, omegaRadians);
    vec4 omegaThetaQuaternion = quaternionMultiply(omegaQuaternion, thetaQuaternion);
    mat4 omegaThetaRotation = mat4FromQuaternion(omegaThetaQuaternion);

    return omegaThetaRotation;
}

void main() {
    mat4 billboardMatrix = billboarded_orientation_matrix(u_cc3MatrixModel, u_cc3MatrixView, a_cc3Normal, vec3(1.0, 0.0, 0.0));
    vec4 position_view = u_cc3MatrixView * u_cc3MatrixModel * billboardMatrix * a_cc3Position;
    v_texCoord = a_cc3TexCoord0;
    v_position_view = position_view.xyz;
    gl_Position = u_cc3MatrixProj * position_view;
}
