#define SUPPORT_GLSL300
#include "prefix.sh"
#include "rc_shadow.fsh"

precision highp int;
precision highp float;
precision highp sampler2D;

uniform struct {
    vec3 _cameraForwardDir; // always that of active camera, even in shadow pass
    int _filteringMode; // 0: Force no filtering, 1: Enabled on CrabbyFiltered
    int _aoMode;        // 0: None, 1: Height only, 2: Height and geometric
} SharedUniforms;

uniform struct {
    int _technique; // 0: SingleFrame, 1: TexturePacker, 2: Crabby, 3: CrabbyFiltered
    int _numFrames; // only needed by Crabby texture technique
    float _fps;
    float _scale;
    float _yOffset;
    float _aspectRatio;
    float _alphaThreshold;
    int _animationID;
    int _playOnce;
    int _facingCamera;
} MaterialUniforms;

//
// color texture used by SingleFrame and TexturePacker technique
//
uniform sampler2D s_cc3Texture2Ds[1];

//
// Crabby compression technique
//
uniform sampler2D uAtlasTexture;
uniform sampler2D uFramesTexture;

VARYING_IN vec2 v_frameOffset;
VARYING_IN vec2 v_frameScale;
VARYING_IN vec2 v_sourceImageSize;
VARYING_IN vec2 v_atlasTextureSize;

const float cBlockSize = 4.0;
const float cMaxBlocksWidth = 256.0 / 4.0;
const float cPaddedBlockSize = cBlockSize + 2.0;
const float cMaxPaddedBlocksWidth = floor(256.0 / cPaddedBlockSize);

//
// General varyings
//
VARYING_IN vec2 v_texCoord[1];
VARYING_IN float v_random[3];
VARYING_IN float v_occlusionFactor;


vec3 rgb2hsv(vec3 c)
{
    vec4 K = vec4(0.0, -1.0 / 3.0, 2.0 / 3.0, -1.0);
    vec4 p = mix(vec4(c.bg, K.wz), vec4(c.gb, K.xy), step(c.b, c.g));
    vec4 q = mix(vec4(p.xyw, c.r), vec4(c.r, p.yzx), step(p.x, c.r));

    float d = q.x - min(q.w, q.y);
    float e = 1.0e-10;
    return vec3(abs(q.z + (q.w - q.y) / (6.0 * d + e)), d / (q.x + e), q.x);
}

vec3 hsv2rgb(vec3 c)
{
    vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
    vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
    return c.z * mix(K.xxx, clamp(p - K.xxx, 0.0, 1.0), c.y);
}

vec4 decodeCrabbyTexture(vec2 texCoord)
{
    // transform texure coordinate into pixel space
    texCoord.y = 1.0 - texCoord.y;
    vec2 pixel = texCoord * v_sourceImageSize;

    // fetch block index from "frames" texture (GL_RG8 format) for this animation frame
    vec2 controlUV = texCoord * v_frameScale + v_frameOffset;
    vec2 blockIndexSample = textureLod(uFramesTexture, controlUV, 0.0).rg;
    ivec2 blockIndexBytes = ivec2(blockIndexSample * 255.0);
    float blockIndex = float(blockIndexBytes.x) + float(blockIndexBytes.y) * 256.0;

    // given block index, determine the pixel coordinate in the block atlas texture
    vec2 pixelCoord;
    if (MaterialUniforms._technique == 2) {
        vec2 blockCoord = vec2(mod(blockIndex, cMaxBlocksWidth), floor(blockIndex / cMaxBlocksWidth));
        vec2 blockPixel = floor(mod(pixel, cBlockSize));
        pixelCoord = blockCoord * cBlockSize + blockPixel;
    } else if (MaterialUniforms._technique == 3) {
        vec2 blockCoord = vec2(mod(blockIndex, cMaxPaddedBlocksWidth), floor(blockIndex / cMaxPaddedBlocksWidth));
        vec2 blockPixel = mod(pixel, cBlockSize) + vec2(1.0);
        pixelCoord = blockCoord * cPaddedBlockSize + blockPixel;
    }

    if (MaterialUniforms._technique == 3 && SharedUniforms._filteringMode == 1) {
        vec2 centerOffset = 1.0 / v_atlasTextureSize;
        pixelCoord += centerOffset + centerOffset * 0.5;
        return textureLod(uAtlasTexture, pixelCoord / v_atlasTextureSize, 0.0);
    } else { // crabby filtering debug override - no filtering
        return texelFetch(uAtlasTexture, ivec2(pixelCoord), 0);
    }
}

void main()
{
    vec4 color = vec4(0.0);
    if (MaterialUniforms._technique == 0 || MaterialUniforms._technique == 1) {
        color = textureLod(s_cc3Texture2Ds[0], v_texCoord[0].xy, 0.0);
    } else if (MaterialUniforms._technique == 2 || MaterialUniforms._technique == 3) {
        color = decodeCrabbyTexture(v_texCoord[0].xy);
    }

    if (color.a < MaterialUniforms._alphaThreshold)
        discard;

#ifdef USE_HSV_VARIATION
    vec3 hsv = rgb2hsv(color.rgb);
    hsv.r += (v_random[0] - 0.5) * 0.05;
    hsv.g += (v_random[1] - 0.5) * 0.2;
    hsv.b += (v_random[2] - 0.5) * 0.2;
    color.rgb = hsv2rgb(hsv);
#endif

    float height = (v_texCoord[0].y - (0.5 - MaterialUniforms._yOffset)) / (0.5 + MaterialUniforms._yOffset);
    float heightAo = 1.0 - clamp(0.2 + pow(height, 0.8), 0.0, 1.0);
    if (MaterialUniforms._technique == 0 || MaterialUniforms._technique == 1) {
        heightAo = 0.0;
    }
    float ao = heightAo * (0.25 + v_occlusionFactor * 0.75);
    if (SharedUniforms._aoMode == 0) {
        ao = 0.0;
    } else if (SharedUniforms._aoMode == 1) {
        ao = heightAo * 0.5;
    }

    color.rgb *= 1.0 - ao;
    gl_FragColor = color;

    shadow_receive();
}
