#define SUPPORT_GLSL300
#include "prefix.sh"
#undef VS_IN_S_POSITION
#undef VS_IN_S_TEXCOORD
#undef VS_OUT_S_TEXCOORD
#define VS_IN_S_POSITION 1
#define VS_IN_S_TEXCOORD 1
#define VS_OUT_S_TEXCOORD 1
#include "framework/framework.sh"
#include "framework/transforms.sh"
#include "rc_shadow.vsh"
#include "poisson.h"

precision highp sampler2D;

uniform struct {
    vec3 _cameraForwardDir; // always that of active camera, even in shadow pass
    int _filteringMode; // 0: Force no filtering, 1: Enabled on CrabbyFiltered
    int _aoMode;        // 0: None, 1: Height only, 2: Height and geometric
} SharedUniforms;

uniform struct {
    int _technique; // 0: SingleFrame, 1: TexturePacker, 2: Crabby, 3: CrabbyFiltered
    int _numFrames; // only needed by Crabby texture technique
    float _fps;
    float _scale;
    float _yOffset;
    float _aspectRatio;
    float _alphaThreshold;
    int _animationID;
    int _playOnce;
    int _facingCamera;
} MaterialUniforms;

uniform struct {
    mat4 _localToWorld;
    int _numInstances;
    int _randomSeed;
    int _distributionType;
    float _distributionRadius;
    float _animationSwitchTime;
} InstanceGroupUniforms;

//
// Crabby compression technique
//
uniform sampler2D uMetadataTexture;
VARYING_OUT vec2 v_frameOffset;
VARYING_OUT vec2 v_frameScale;
VARYING_OUT vec2 v_sourceImageSize;
VARYING_OUT vec2 v_atlasTextureSize;

//
// TexturePacker offsets (also used by Crabby techniques)
//
uniform sampler2D uSpriteOffsetTexture;

uniform sampler2D uBlueNoiseTexture;

ATTRIBUTE int a_instanceID;
ATTRIBUTE vec3 a_centerPosition;

VARYING_OUT float v_random[3];
VARYING_OUT float v_occlusionFactor;

#define M_PI 3.1415926
#define BLUE_NOISE_SIZE 64
#define BLUE_NOISE_TEXELS (BLUE_NOISE_SIZE * BLUE_NOISE_SIZE)


float rand(vec4 seed)
{
    float dot_product = dot(seed, vec4(12.9898,78.233, 45.164, 94.673));
    return fract(sin(dot_product) * 43758.5453);
}

uint unpackU16(vec2 sampledVal)
{
    return uint(sampledVal.r * 255.0) | (uint(sampledVal.g * 255.0) << 8);
}

uint unpackU8(float sampledVal)
{
    return uint(sampledVal * 255.0);
}

// calculate instance center position based on given distribution type
vec3 randomizePosition(vec3 centerPosition)
{
    vec2 offset;
    float occlusionFactor = 0.0;

    if (InstanceGroupUniforms._numInstances == 1) {
        offset = vec2(0.0);
    } else if (InstanceGroupUniforms._distributionType == 0 || InstanceGroupUniforms._distributionType == 1) {  // blue noise based
        int randOffset = int(rand(vec4(InstanceGroupUniforms._randomSeed) * 123.456) * float(BLUE_NOISE_TEXELS));
        int index = (a_instanceID + randOffset) % BLUE_NOISE_TEXELS;
        ivec2 noiseTexel = ivec2(index & 0x3F, (index / BLUE_NOISE_SIZE) & 0x3F);
        vec4 noise = texelFetch(uBlueNoiseTexture, noiseTexel, 0).rgba;

        if (InstanceGroupUniforms._distributionType == 0) { // disc
            float scale = sqrt(float(InstanceGroupUniforms._numInstances)) * 0.5;
            vec2 polarCoord = vec2(noise.r * 2.0 * M_PI, pow(noise.g, 0.5));
            vec2 discPos = vec2(cos(polarCoord.x), sin(polarCoord.x)) * polarCoord.y;
            offset = discPos * scale;
            occlusionFactor = pow(1.0 - polarCoord.y, 0.8);
        } else if (InstanceGroupUniforms._distributionType == 1) { // square
            float scale = sqrt(float(InstanceGroupUniforms._numInstances));
            vec2 squarePos = noise.rg - vec2(0.5);
            offset = squarePos * scale;
            occlusionFactor = pow(1.0 - max(abs(squarePos.x * 2.0), abs(squarePos.y * 2.0)), 0.5);
        }
    } else if (InstanceGroupUniforms._distributionType == 2) { // experimental
        int randOffset = int(rand(vec4(InstanceGroupUniforms._randomSeed) * 123.456) * 128.0);
        int index = (a_instanceID + randOffset) % 128;
        float scale = sqrt(float(InstanceGroupUniforms._numInstances)) * 0.5;
        offset = Poisson(index) * scale;
    }

    // pass geometric occlusion factor
    v_occlusionFactor = occlusionFactor;

    offset *= InstanceGroupUniforms._distributionRadius;
    return centerPosition + mat3(InstanceGroupUniforms._localToWorld) * vec3(offset.x, 0.0, offset.y);
}

vec3 getVertexPositionTexturePacker(vec3 inPosition, vec2 inTexCoord, out vec2 outTexCoord)
{
    vec4 meta1 = texelFetch(uSpriteOffsetTexture, ivec2(0, 5), 0);
    vec4 meta2 = texelFetch(uSpriteOffsetTexture, ivec2(1, 5), 0);

    uint numFrames = unpackU16(meta1.rg);
    uvec2 spritesheetSize = uvec2(unpackU16(meta2.rg), unpackU16(meta2.ba));

    // determine current animation frame
    uint currFrame;
    float fps = MaterialUniforms._fps;
    if (MaterialUniforms._playOnce == 1) {
        float timeOffset = v_random[0] * 0.5;
        float animationTime = getSceneTime() - InstanceGroupUniforms._animationSwitchTime - timeOffset;
        currFrame = clamp(uint(animationTime * fps), 0u, numFrames);
    } else {
        float startFrame = v_random[1] * float(numFrames);
        currFrame = uint(getSceneTime() * fps + startFrame) % numFrames;
    }

    vec4 c1 = texelFetch(uSpriteOffsetTexture, ivec2(currFrame, 4), 0);
    vec4 c2 = texelFetch(uSpriteOffsetTexture, ivec2(currFrame, 3), 0);
    vec4 c3 = texelFetch(uSpriteOffsetTexture, ivec2(currFrame, 2), 0);
    vec4 c4 = texelFetch(uSpriteOffsetTexture, ivec2(currFrame, 1), 0);
    vec4 c5 = texelFetch(uSpriteOffsetTexture, ivec2(currFrame, 0), 0);

    uvec4 frameInSpritesheet = uvec4(unpackU16(c1.rg), unpackU16(c1.ba), unpackU16(c2.rg), unpackU16(c2.ba));
    uvec4 frameInSource = uvec4(unpackU16(c3.rg), unpackU16(c3.ba), unpackU16(c4.rg), unpackU16(c4.ba));
    uvec2 sourceSize = uvec2(unpackU16(c5.rg), unpackU16(c5.ba));

    // flip vertically to convert from image space to UV space coordinate
    frameInSpritesheet.y = spritesheetSize.y - frameInSpritesheet.y - frameInSpritesheet.w;
    frameInSource.y = sourceSize.y - frameInSource.y - frameInSource.w;

    // determine vertex position and texcoord
#ifdef USE_FULL_BILLBOARD
    vec2 texCoord = inTexCoord * vec2(sourceSize) - vec2(frameInSource.xy) + vec2(frameInSpritesheet.xy);
    outTexCoord = texCoord / vec2(spritesheetSize);
    // v_clipRect = vec4(frameInSource.xy, frameInSource.xy + frameInSource.zw) / vec4(sourceSize.xyxy);
    // setTexCoord(vsout, 1, texCoord);
    return inPosition;
#else
    vec2 texCoord = inTexCoord * vec2(frameInSource.zw) + vec2(frameInSpritesheet.xy);
    outTexCoord = texCoord / vec2(spritesheetSize);

    vec2 texCoordInSource = (inTexCoord * vec2(frameInSource.zw) + vec2(frameInSource.xy)) / vec2(sourceSize);
    vec2 offsetTexCoord = texCoordInSource - vec2(0.5);
    vec2 aspect = vec2(MaterialUniforms._aspectRatio, 1.0);
    return vec3(offsetTexCoord * aspect, 0);
#endif
}

// perform billboard trimming to reduce number of pixels to shade
vec3 getTrimmedVertexPositionCrabby(uint currFrame, vec2 inTexCoord, out vec2 outTexCoord)
{
    vec4 c3 = texelFetch(uSpriteOffsetTexture, ivec2(currFrame, 2), 0);
    vec4 c4 = texelFetch(uSpriteOffsetTexture, ivec2(currFrame, 1), 0);
    vec4 c5 = texelFetch(uSpriteOffsetTexture, ivec2(currFrame, 0), 0);

    uvec4 frameInSource = uvec4(unpackU16(c3.rg), unpackU16(c3.ba), unpackU16(c4.rg), unpackU16(c4.ba));
    uvec2 sourceSize = uvec2(unpackU16(c5.rg), unpackU16(c5.ba));

    // flip vertically to convert from image space to UV space coordinate
    frameInSource.y = sourceSize.y - frameInSource.y - frameInSource.w;

    outTexCoord = (inTexCoord * vec2(frameInSource.zw) + vec2(frameInSource.xy)) / vec2(sourceSize);

    vec2 offsetTexCoord = outTexCoord - vec2(0.5);
    vec2 aspect = vec2(MaterialUniforms._aspectRatio, 1.0);
    return vec3(offsetTexCoord * aspect, 0);
}

vec3 getVertexPositionCrabby(vec3 inPosition, vec2 inTexCoord, out vec2 outTexCoord)
{
    // determine current animation frame
    uint currFrame;
    uint numFrames = uint(MaterialUniforms._numFrames);
    float fps = MaterialUniforms._fps;
    if (MaterialUniforms._playOnce == 1) { // clamp the animation to the end
        float timeOffset = v_random[0] * 0.5;
        float animationTime = getSceneTime() - InstanceGroupUniforms._animationSwitchTime - timeOffset;
        currFrame = clamp(uint(animationTime * fps), 0u, numFrames);
    } else {
        float startFrame = v_random[1] * float(numFrames);
        currFrame = uint(getSceneTime() * fps + startFrame) % numFrames;
    }

    // prepare offset information of this animation frame for crabby decoding
    vec4 pixel1 = texelFetch(uMetadataTexture, ivec2(currFrame, 3), 0);
    vec4 pixel2 = texelFetch(uMetadataTexture, ivec2(currFrame, 2), 0);
    vec4 pixel3 = texelFetch(uMetadataTexture, ivec2(currFrame, 1), 0);
    vec4 pixel4 = texelFetch(uMetadataTexture, ivec2(currFrame, 0), 0);

    float offsetX     = float(unpackU16(pixel1.rg)) / 65535.0;
    float offsetY     = float(unpackU16(pixel1.ba)) / 65535.0;
    float rcpWidth    = float(unpackU16(pixel2.rg)) / 65535.0;
    float rcpHeight   = float(unpackU16(pixel2.ba)) / 65535.0;
    float frameWidth  = float(unpackU16(pixel3.rg));
    float frameHeight = float(unpackU16(pixel3.ba));
    float atlasWidth  = float(unpackU16(pixel4.rg));
    float atlasHeight = float(unpackU16(pixel4.ba));

    v_frameOffset = vec2(offsetX, offsetY);
    v_frameScale = vec2(rcpWidth, rcpHeight);
    v_sourceImageSize = vec2(frameWidth, frameHeight);
    v_atlasTextureSize = vec2(atlasWidth, atlasHeight);

    // determine vertex position and texcoord
#ifdef USE_FULL_BILLBOARD
    outTexCoord = inTexCoord;
    return inPosition;
#else
    return getTrimmedVertexPositionCrabby(currFrame, inTexCoord, outTexCoord);
#endif
}

void vsMain(out VSOUT vsout, VSIN vsin)
{
    // prepare some per-instance random numbers for custom randomization
    v_random[0] = rand(vec4(a_instanceID, a_centerPosition.xyz));
    v_random[1] = rand(vec4(a_centerPosition.y, a_instanceID, a_centerPosition.xz));
    v_random[2] = rand(vec4(a_centerPosition.zyx, a_instanceID));

    // evaluate vertex position and texture coordinate based on texture technique used
    vec2 aspect = vec2(MaterialUniforms._aspectRatio, 1.0);
    vec3 inPosition = getPosition(vsin) * vec3(aspect, 0.0);
    vec2 inTexCoord = getTexCoord(vsin, 0);

    vec2 texCoord;
    vec3 vertexPosition;
    if (MaterialUniforms._technique == 0) {
        vertexPosition = inPosition;
        texCoord = inTexCoord;
    } else if (MaterialUniforms._technique == 1) {
        vertexPosition = getVertexPositionTexturePacker(inPosition, inTexCoord, texCoord);
    } else if (MaterialUniforms._technique == 2 || MaterialUniforms._technique == 3) {
        vertexPosition = getVertexPositionCrabby(inPosition, inTexCoord, texCoord);
    }
    setTexCoord(vsout, 0, texCoord);

    // apply random distribution
    vec3 instanceWorldPosition = randomizePosition(a_centerPosition);

    float rotationAngle = 0.0;

    // align billboard to camera
    if (MaterialUniforms._facingCamera == 1) {
        vec3 projectedCameraDir = normalize(SharedUniforms._cameraForwardDir * vec3(1, 0, 1));
        float billboardAngle = -atan(projectedCameraDir.z, projectedCameraDir.x) - M_PI * 0.5;
        rotationAngle += billboardAngle;
    } else {
        rotationAngle += asin(InstanceGroupUniforms._localToWorld[2][0]);
    }

    // apply random rotation
    if (InstanceGroupUniforms._numInstances > 1)
        rotationAngle += (v_random[2] - 0.5) * 0.1 * 2.0 * M_PI;

    float cosTheta = cos(rotationAngle);
    float sinTheta = sin(rotationAngle);
    mat3 rotation = mat3(cosTheta, 0.0, -sinTheta, 0.0, 1.0, 0.0, sinTheta, 0.0, cosTheta);

    // construct final vertex position
    vec3 yOffset = vec3(0.0, MaterialUniforms._yOffset, 0.0);
    vec3 worldPosition = instanceWorldPosition + (rotation * vertexPosition + yOffset) * MaterialUniforms._scale;
    vsout._position = transformPoint(getWorldToProj(), worldPosition);

    // generate shadow coords
    shadow_receive(vec4(worldPosition, 1.0));
}

#include "framework/glsles/framework_glsles.vert"
