
#include "library/texture_utility.sh"

#if __VERSION__ >= 300

//===================================================================================================================//
// UNIFORMS //
//===================================================================================================================//
uniform samplerCube DiffuseIBL;
uniform samplerCube SpecularIBL;
uniform sampler2D   GlossLUT;
uniform sampler2D   DiffusePanoIBL;
uniform struct {
    vec4 _diffuse[9];
} EnvironmentLighting;

//===================================================================================================================//
// XLE LIGHTING CORE //
//===================================================================================================================//
struct LightScreenDest
{
    uvec2 pixelCoords;
    uint sampleIndex;
};

const uint GlobalSamplingPassCount = 16u;
const uint GlobalSamplingPassIndex = 0u;
vec3 ReadSkyReflectionTexture(vec3 coord, float mip) { return vec3(0.0); }

#include "XLE/XLEShaderBridge.h"
#include "XLE/Lighting/ImageBased.h"
#include "XLE/Lighting/SphericalHarmonics.h"

struct PBRParameters
{
    float3  _ambientColor;
    float   _dielectricF0;
    float   _roughness;
    float   _metal;
    float   _bakedAmbientOcclusion;
};

PBRParameters LoadPBRParameters(vec4 propertiesTextureTap)
{
    PBRParameters result;
    result._ambientColor = float3(0.0f, 0.0f, 0.0f);
    result._dielectricF0 = 18.0f / 255.0f;
    #if 0
            // (note; this is attempt to extract our properties from the old "property" map parameters)
        float mat_mask = propertiesTextureTap.r;                // In Substance -- "Specular" converted map
        float mat_gloss = propertiesTextureTap.g;               // In Substance -- "Unity 4 Gloss" converted map
        float mat_metallic = propertiesTextureTap.b;            // In Substance -- "F0" converted map
        float mat_cloth = propertiesTextureTap.a;               // In Substance -- "User0"

        // float roughness = Sq(1.0f - mat_mask);
        result._roughness = Sq(1.0f - mat_gloss);
        result._metal = mat_metallic;
        result._bakedAmbientOcclusion = 1.0f;
    #else
        result._roughness = propertiesTextureTap.r;             // In Substance -- "Roughness"
        result._metal = propertiesTextureTap.g;                 // In Substance -- "Metallic"
        result._bakedAmbientOcclusion = propertiesTextureTap.b; // In Substance -- "Ambient occlusion"
    #endif
    return result;
}


vec3 CalculateMantisLighting(vec4 propertiesTextureTap, vec3 diffuseAlbedoSRGB, vec3 N, vec3 V)
{
// note -- our coordinate system is +Y up
//          On Kuro model, +X is to his left, +Z is his forward

    LightScreenDest lsd;
    lsd.pixelCoords = uint2(0,0);
    lsd.sampleIndex = 0u;

    PBRParameters params = LoadPBRParameters(propertiesTextureTap);

    float3 directDiffuseAlbedo = SRGBToLinear(diffuseAlbedoSRGB);
    // float3 diffuseAlbedo = DiffuseColorBlend(lerp(directDiffuseAlbedo, float3(0.5, 0.5, 0.5), params._metal));
    float3 diffuseAlbedo = DiffuseColorBlend(directDiffuseAlbedo);
    
    float3 diffuseIBL = ResolveSH_Opt4(EnvironmentLighting._diffuse, AdjSkyCubeMapCoords(N));

    float ambientOcclusion = params._bakedAmbientOcclusion;
    float3 finalDiffuse = (ambientOcclusion*(1.0f - params._metal))*(params._ambientColor + diffuseIBL)*diffuseAlbedo;

    // note that we're using the linear space version of diffuseAlbedo for the metal F0 response
    // depending on how this texture is authored, we may want either this or the original SRGB version
    float3 F0 = lerp(vec3(params._dielectricF0, params._dielectricF0, params._dielectricF0), directDiffuseAlbedo, params._metal);
    SpecularParameters specParams = SpecularParameters_RoughF0(params._roughness, F0);
    vec3 specularIBL = SampleSpecularIBL(N, V, specParams, lsd);
    float specularOcclusion = TriAceSpecularOcclusion(dot(N, V), ambientOcclusion);

    // DavidJ -- note -- we pumped up the specular for WM3D; but we don't want this to get into Iceland.
    const float specularScale = 1.0f; // 3.14159f;

    float3 result = finalDiffuse + (specularScale * specularOcclusion) * specularIBL;

    // back to SRGB space for the calling shader
    return LinearToSRGB(result);
}

static const int MantisLightingDebug_Begin             = 10;
static const int MantisLightingDebug_End               = MantisLightingDebug_Begin + 11;
static const int MantisLightingDebug_Roughness         = MantisLightingDebug_Begin + 0;
static const int MantisLightingDebug_Metal             = MantisLightingDebug_Begin + 1;
static const int MantisLightingDebug_F0                = MantisLightingDebug_Begin + 2;
static const int MantisLightingDebug_AmbientOcclusion  = MantisLightingDebug_Begin + 3;
static const int MantisLightingDebug_SpecularOcclusion = MantisLightingDebug_Begin + 4;
static const int MantisLightingDebug_DiffuseIBL        = MantisLightingDebug_Begin + 5;
static const int MantisLightingDebug_SpecularIBL       = MantisLightingDebug_Begin + 6;
static const int MantisLightingDebug_GlossLUT          = MantisLightingDebug_Begin + 7;
static const int MantisLightingDebug_DiffuseAlbedo     = MantisLightingDebug_Begin + 8;
static const int MantisLightingDebug_FinalDiffuse      = MantisLightingDebug_Begin + 9;
static const int MantisLightingDebug_FinalLighting     = MantisLightingDebug_Begin + 10;

vec3 CalculateMantisLightingDebug(int debugId, vec4 propertiesTextureTap, vec3 diffuseAlbedoSRGB, vec3 N, vec3 V)
{
    LightScreenDest lsd;
    lsd.pixelCoords = uint2(0,0);
    lsd.sampleIndex = 0u;

    float3 directDiffuseAlbedo = SRGBToLinear(diffuseAlbedoSRGB);
    float3 diffuseAlbedo = DiffuseColorBlend(directDiffuseAlbedo);
    PBRParameters params = LoadPBRParameters(propertiesTextureTap);
    float ambientOcclusion = params._bakedAmbientOcclusion;
    
    float3 diffuseIBL = ResolveSH_Opt4(EnvironmentLighting._diffuse, AdjSkyCubeMapCoords(N));
    
    float3 finalDiffuse = (ambientOcclusion*(1.0f - params._metal))*(params._ambientColor + diffuseIBL)*diffuseAlbedo;

    float3 F0 = lerp(vec3(params._dielectricF0, params._dielectricF0, params._dielectricF0), directDiffuseAlbedo, params._metal);
    SpecularParameters specParams = SpecularParameters_RoughF0(params._roughness, F0);
    vec3 specularIBL = SampleSpecularIBL(N, V, specParams, lsd);
    float specularOcclusion = TriAceSpecularOcclusion(dot(N, V), ambientOcclusion);
    float3 result = finalDiffuse + specularOcclusion * specularIBL;

    if (debugId == MantisLightingDebug_Roughness)               { return LinearToSRGB(vec3(params._roughness, params._roughness, params._roughness)); }
    else if (debugId == MantisLightingDebug_Metal)              { return LinearToSRGB(vec3(params._metal, params._metal, params._metal)); }
    else if (debugId == MantisLightingDebug_F0)                 { return LinearToSRGB(F0); }
    else if (debugId == MantisLightingDebug_AmbientOcclusion)   { return LinearToSRGB(vec3(ambientOcclusion, ambientOcclusion, ambientOcclusion)); }
    else if (debugId == MantisLightingDebug_SpecularOcclusion)  { return LinearToSRGB(vec3(specularOcclusion, specularOcclusion, specularOcclusion)); }
    else if (debugId == MantisLightingDebug_DiffuseIBL)         { return LinearToSRGB(diffuseIBL); }
    else if (debugId == MantisLightingDebug_SpecularIBL)        { return LinearToSRGB(specularIBL); }
    else if (debugId == MantisLightingDebug_FinalDiffuse)       { return LinearToSRGB(finalDiffuse); }
    else if (debugId == MantisLightingDebug_FinalLighting)      { return LinearToSRGB(result); }
    else if (debugId == MantisLightingDebug_GlossLUT) {
        float2 glossLut = SplitSumIBL_IntegrateBRDF(params._roughness, dot(N, V), 0u);
        return LinearToSRGB(vec3(glossLut.xy, 0.0f));
    } else if (debugId == MantisLightingDebug_DiffuseAlbedo) {
        return LinearToSRGB((1.0f - params._metal) * diffuseAlbedo);
    }

    return vec3(0.0f, 0.0f, 0.0f);
}

#else

#undef MANTIS_PBR

#endif

