//NOTE: this is a placeholder file until I deploy the new shader file organization - Simon

#if __VERSION__ >= 300

//===================================================================================================================//
// UNIFORMS //
//===================================================================================================================//
uniform samplerCube DiffuseIBL;
uniform samplerCube SpecularIBL;
uniform sampler2D   GlossLUT;
uniform sampler2D   DiffusePanoIBL;
uniform struct
{
    vec4 _diffuse[9];
} EnvironmentLighting;

//===================================================================================================================//
// XLE LIGHTING CORE //
//===================================================================================================================//
struct LightScreenDest
{
    uvec2 pixelCoords;
    uint sampleIndex;
};

const uint GlobalSamplingPassCount = 16u;
const uint GlobalSamplingPassIndex = 0u;
vec3 ReadSkyReflectionTexture(vec3 coord, float mip) { return vec3(0.0); }

#include "XLE/XLEShaderBridge.h"
#include "XLE/Lighting/ImageBased.h"
#include "XLE/Lighting/SphericalHarmonics.h"
#include "library/texture_utility.sh"

struct PBRParameters
{
    float3  _ambientColor;
    float   _dielectricF0;
    float   _roughness;
    float   _metal;
    float   _bakedAmbientOcclusion;
    float   _user0Mask;//most commonly used for unlit mask "emissive" - but not really emissive in our current implementation
};

PBRParameters LoadPBRParameters(vec4 propertiesTextureTap)
{
    PBRParameters result;
    result._ambientColor = float3(0.0f, 0.0f, 0.0f);
    result._dielectricF0 = 18.0f / 255.0f;

    result._roughness = propertiesTextureTap.r;             // In Substance -- "Roughness"
    result._metal = propertiesTextureTap.g;                 // In Substance -- "Metallic"
    result._bakedAmbientOcclusion = propertiesTextureTap.b; // In Substance -- "Ambient occlusion"
    result._user0Mask = propertiesTextureTap.a;             // Self-Illum Mask

    return result;
}

float3 GetPBRLighting(vec3 N)
{
    return ResolveSH_Opt4(EnvironmentLighting._diffuse, AdjSkyCubeMapCoords(N));
}

vec3 ApplyPBRLighting(float3 diffuseIBL, vec4 propertiesTextureTap, vec3 diffuseAlbedoSRGB, vec3 position, vec3 N, vec3 V, float glossBoost)
{
    // note -- our coordinate system is +Y up
    //          On Kuro model, +X is to his left, +Z is his forward

    LightScreenDest lsd;
    lsd.pixelCoords = uint2(0,0);
    lsd.sampleIndex = 0u;

    float3 diffuseAlbedo = SRGBToLinear(diffuseAlbedoSRGB);

    PBRParameters params = LoadPBRParameters(propertiesTextureTap);
    float ambientOcclusion = params._bakedAmbientOcclusion;
    float3 finalDiffuse = (ambientOcclusion*(1.0f - params._metal))*(params._ambientColor + diffuseIBL)*diffuseAlbedo;

    // note that we are using the linear space version of diffuseAlbedo for the metal F0 response
    // depending on how this texture is authored, we may want either this or the original SRGB version

    float3 F0 = lerp(vec3(params._dielectricF0, params._dielectricF0, params._dielectricF0), diffuseAlbedo, params._metal);
    SpecularParameters specParams = SpecularParameters_RoughF0(params._roughness, F0);
    vec3 specularIBL = SampleSpecularIBL(N, V, specParams, lsd);
    specularIBL *= (1.0 + glossBoost * (1.0 - params._roughness));

    float specularOcclusion = TriAceSpecularOcclusion(dot(N, V), ambientOcclusion);

    float3 result = finalDiffuse + specularOcclusion * specularIBL;

    // back to SRGB space for the calling shader
    return LinearToSRGB(result);
}

#else

// Just a fallback PBR implementation for OSX w/ desktop OpenGL 1.x
// Note that sampling from IBL cubemap mip-maps is not supported; reflection will always look like roughness = 0

#include "library/texture_utility.sh"
#define SampleLevel_Default(sampler, coords, mipmap) textureCube(sampler, coords)
#define saturate(x) clamp(x, 0.0, 1.0)
#define lerp(x, a, b) mix(x, a, b)

uniform samplerCube SpecularIBL;
uniform sampler2D   GlossLUT;
uniform struct
{
    vec4 _diffuse[9];
} EnvironmentLighting;

struct PBRParameters
{
    vec3    _ambientColor;
    float   _dielectricF0;
    float   _roughness;
    float   _metal;
    float   _bakedAmbientOcclusion;
    float   _user0Mask;
};

struct SpecularParameters
{
    float roughness;
    vec3  F0;
    vec3  transmission;
    bool  mirrorSurface;
};

PBRParameters LoadPBRParameters(vec4 propertiesTextureTap)
{
    PBRParameters result;
    result._ambientColor = vec3(0.0f, 0.0f, 0.0f);
    result._dielectricF0 = 18.0f / 255.0f;

    result._roughness = propertiesTextureTap.r;             // In Substance -- "Roughness"
    result._metal = propertiesTextureTap.g;                 // In Substance -- "Metallic"
    result._bakedAmbientOcclusion = propertiesTextureTap.b; // In Substance -- "Ambient occlusion"
    result._user0Mask = propertiesTextureTap.a;             // Self-Illum Mask

    return result;
}

float TriAceSpecularOcclusion(float NdotV, float ao)
{
    float q = (NdotV + ao);
    return saturate(q * q - 1.f + ao);
}

float RoughnessToMipmap(float roughness)
{
    const float SpecularIBLMipMapCount = 9.f;
    return saturate(roughness - 0.05f) * SpecularIBLMipMapCount;
}

vec3 ResolveSH_Opt4(vec4 premulCoefficients[9], vec3 dir)
{
    vec3 result = premulCoefficients[0].rgb;

    result += premulCoefficients[1].rgb * dir.y;
    result += premulCoefficients[2].rgb * dir.z;
    result += premulCoefficients[3].rgb * dir.x;

    vec3 dirSq = dir * dir;
    result += premulCoefficients[4].rgb * (dir.x * dir.y);
    result += premulCoefficients[5].rgb * (dir.z * dir.y);
    result += premulCoefficients[6].rgb * dirSq.z;
    result += premulCoefficients[7].rgb * (dir.x * dir.z);
    result += premulCoefficients[8].rgb * (dirSq.x - dirSq.y);
    return result;
}

vec3 SampleSpecularIBL(vec3 normal, vec3 viewDirection, SpecularParameters specParam)
{
    float NdotV = dot(normal, viewDirection);
    vec3 R = 2.f * dot(viewDirection, normal) * normal - viewDirection;
    vec3 prefilteredColor = SampleLevel_Default(SpecularIBL, R, RoughnessToMipmap(specParam.roughness)).rgb;
    vec2 envBRDF = texture2D(GlossLUT, vec2(clamp(NdotV, 0.0f, 1.0f), clamp(specParam.roughness, 0.0f, 1.0f))).xy;

    return prefilteredColor * (specParam.F0 * envBRDF.x + envBRDF.y);
}

// public
vec3 GetPBRLighting(vec3 N)
{
    return ResolveSH_Opt4(EnvironmentLighting._diffuse, N); 
}

// public
vec3 ApplyPBRLighting(vec3 diffuseIBL, vec4 propertiesTextureTap, vec3 diffuseAlbedoSRGB, vec3 position, vec3 N, vec3 V, float glossBoost)
{
    vec3 diffuseAlbedo = SRGBToLinear(diffuseAlbedoSRGB);

    PBRParameters params = LoadPBRParameters(propertiesTextureTap);
    float ambientOcclusion = params._bakedAmbientOcclusion;
    vec3 finalDiffuse = (ambientOcclusion*(1.0f - params._metal))*(params._ambientColor + diffuseIBL)*diffuseAlbedo;

    SpecularParameters specParams;
    specParams.roughness = params._roughness;
    specParams.F0 = lerp(vec3(params._dielectricF0, params._dielectricF0, params._dielectricF0), diffuseAlbedo, params._metal);

    vec3 specularIBL = SampleSpecularIBL(N, V, specParams);
    specularIBL *= (1.0 + glossBoost * (1.0 - params._roughness));

    float specularOcclusion = TriAceSpecularOcclusion(dot(N, V), ambientOcclusion);

    vec3 result = finalDiffuse + specularOcclusion * specularIBL;
    return LinearToSRGB(result);
}

#endif
