
#define SUPPORT_GLSL300
#include "prefix.sh"
#include "texture_utility.sh"
#include "projection_utility.sh"

uniform sampler2D s_cc3Texture2Ds[4];
VARYING_IN vec2 v_texCoord;

#if __VERSION__ >= 300
    #define o_fragColor destColor
#else
    #define o_fragColor gl_FragColor
#endif

void main() {

    //
    // We're going to perform a bilateral upsample of the low res buffers onto the high res buffers
    // This is basically a bilinear filter of the low res texture. However, we use the difference
    // between the high res depth and the low res depth to bias the weights of the filter. Our goal
    // is to make sure the high resolution edges remain sharp and uneffected, even if low res
    // geometry has been rendered in the same location.
    //
    // There are other methods to consider, also... Bungie uses a method they called variance depth
    // comparisons -- which is designed for particles, so that the depth values in the buffer are
    // considered parameters into a optical thickness equation, rather than concrete depths. That
    // would work well for particles and things emulating spheres.
    //
    
    // Following the bilateral upsampling method presented by AMD at GDC, and described here:
    // http://amd-dev.wpengine.netdna-cdn.com/wordpress/media/2012/10/ShopfMixedResolutionRendering.pdf
    
    // Note, assuming OpenGL ES pixel coordinate system (which is similar to DX10+)
    // fragment origin is the lower level of the fragment. Fragment centers are at (0.5, 0.5)
    //
    
    vec2 adjustedCoord = gl_FragCoord.xy * 0.5;

    // The blending algorithm here is a little complex. This is a special case implementation of bilinear
    // filtering that is designed for the upsampling each single pixels in a to 2x2 blocks.
    // We can use gl_FragCoord to calculate the source coordinate. We will offset this by half a pixel in
    // X and Y later on -- this selects the correct block of 2x2 input pixels to sample on.
    // We can use the fractional part of the coordinates to pick the correct bilinear weights. We
    // select the weights the suit the half pixel offset that comes afterwards.
    
    // Note that when drawing the first pixel, gl_FragCoord is (0.5, 0.5) -- (based on
    // experimentation on Air 2 device). In other words, gl_FragCoords is the center of the output
    // pixel (not the lower left corner).
    
    // When selecting the bilinear weights, x and y should be either .25 or .75 (so a comparison
    // against 0.5 should be safe at lower precision)
    
    vec4 vBilinearWeights[4];
    vBilinearWeights[0] = vec4(9.0/16.0, 3.0/16.0, 3.0/16.0, 1.0/16.0);
    vBilinearWeights[1] = vec4(3.0/16.0, 9.0/16.0, 1.0/16.0, 3.0/16.0);
    vBilinearWeights[2] = vec4(3.0/16.0, 1.0/16.0, 9.0/16.0, 3.0/16.0);
    vBilinearWeights[3] = vec4(1.0/16.0, 3.0/16.0, 3.0/16.0, 9.0/16.0);
    vec4 tapWeights;
    if (fract(adjustedCoord.x) < 0.5) {
        if (fract(adjustedCoord.y) < 0.5) {
            tapWeights = vBilinearWeights[3];
        } else {
            tapWeights = vBilinearWeights[1];
        }
    } else {
        if (fract(adjustedCoord.y) < 0.5) {
            tapWeights = vBilinearWeights[2];
        } else {
            tapWeights = vBilinearWeights[0];
        }
    }
    
    adjustedCoord -= vec2(0.5, 0.5); // (or just use adjustedCoord -= vec2(0.25, 0.25); to try to sample exact pixel centers)
    
    vec2 lowResReciprocalSize = 2.0 * reciprocalViewportSize();
    
    // note that we're assuming hardware bilinear is off here! If hardware bilinear is enabled, the filtering will be incorrect.
    vec4 c0 = sampleLevelZero(s_cc3Texture2Ds[0], (adjustedCoord + vec2(0.0, 0.0)) * lowResReciprocalSize).rgba;
    vec4 c1 = sampleLevelZero(s_cc3Texture2Ds[0], (adjustedCoord + vec2(1.0, 0.0)) * lowResReciprocalSize).rgba;
    vec4 c2 = sampleLevelZero(s_cc3Texture2Ds[0], (adjustedCoord + vec2(0.0, 1.0)) * lowResReciprocalSize).rgba;
    vec4 c3 = sampleLevelZero(s_cc3Texture2Ds[0], (adjustedCoord + vec2(1.0, 1.0)) * lowResReciprocalSize).rgba;
    
    float totalW = dot(vec4(c0.a, c1.a, c2.a, c3.a), tapWeights);
    float totalA = c0.a + c1.a + c2.a + c3.a;
    if (totalW == 0.0) discard;
    
    o_fragColor.rgb =
          (c0.rgb * tapWeights.x * c0.a / totalW)
        + (c1.rgb * tapWeights.y * c1.a / totalW)
        + (c2.rgb * tapWeights.z * c2.a / totalW)
        + (c3.rgb * tapWeights.w * c3.a / totalW)
        ;
    
    o_fragColor.a = dot(
        vec4(c0.a > 0.0001f ? 1.0f : 0.0f,
             c1.a > 0.0001f ? 1.0f : 0.0f,
             c2.a > 0.0001f ? 1.0f : 0.0f,
             c3.a > 0.0001f ? 1.0f : 0.0f), tapWeights);
}


