
#if defined(GL_ES)
    precision highp float;
#endif

//#define USE_TEXTURE

uniform sampler2D s_cc3Texture2Ds[2];
uniform float inBlack;
uniform float inGamma;
uniform float inWhite;
uniform vec3 hsl;
uniform float blendWeight;

varying vec2 v_texCoord;

#ifdef USE_TEXTURE

// texture lookup method

void main() {
    const float TEXTURE_SIZE = 16.0;
    vec4 texel = texture2D(s_cc3Texture2Ds[0], v_texCoord);
    // calculate look up coordinates based on pixel color
    float redComponent = texel.r / TEXTURE_SIZE;
    float greenComponent = 1.0 - texel.g;
    float blueComponent = float(int(texel.b * 255.0) / int(TEXTURE_SIZE)) * TEXTURE_SIZE / 255.0;
    vec2 lookUpCoordinates = vec2(redComponent + blueComponent, greenComponent);
    // return the mapped color
    vec3 lookUpColor = texture2D(s_cc3Texture2Ds[1], lookUpCoordinates).rgb;
    vec3 finalColor = mix(texel.rgb, lookUpColor, blendWeight);
    gl_FragColor = vec4(finalColor, texel.a);
}

#else

// photoshop parameters method

#define GammaCorrection(color, gamma) pow(color, vec3(1.0 / gamma))
#define LevelsControlInputRange(color, minInput, maxInput) min(max(color - vec3(minInput), vec3(0.0)) / (vec3(maxInput) - vec3(minInput)), vec3(1.0))
#define LevelsControlInput(color, minInput, gamma, maxInput) GammaCorrection(LevelsControlInputRange(color, minInput, maxInput), gamma)
#define LevelsControlOutputRange(color, minOutput, maxOutput) mix(vec3(minOutput), vec3(maxOutput), color)
#define LevelsControl(color, minInput, gamma, maxInput, minOutput, maxOutput) LevelsControlOutputRange(LevelsControlInput(color, minInput, gamma, maxInput), minOutput, maxOutput)

// HSL

vec3 RGBToHSL(vec3 color) {
	vec3 hsl; // init to 0 to avoid warnings ? (and reverse if + remove first part)

	float fmin = min(min(color.r, color.g), color.b);    //Min. value of RGB
	float fmax = max(max(color.r, color.g), color.b);    //Max. value of RGB
	float delta = fmax - fmin;             //Delta RGB value

	hsl.z = (fmax + fmin) / 2.0; // Luminance

	if (delta == 0.0) {		//This is a gray, no chroma...
		hsl.x = 0.0;	// Hue
		hsl.y = 0.0;	// Saturation
	}
	else {                                    //Chromatic data...
		if (hsl.z < 0.5)
			hsl.y = delta / (fmax + fmin); // Saturation
		else
			hsl.y = delta / (2.0 - fmax - fmin); // Saturation

		float deltaR = (((fmax - color.r) / 6.0) + (delta / 2.0)) / delta;
		float deltaG = (((fmax - color.g) / 6.0) + (delta / 2.0)) / delta;
		float deltaB = (((fmax - color.b) / 6.0) + (delta / 2.0)) / delta;

		if (color.r == fmax )
			hsl.x = deltaB - deltaG; // Hue
		else if (color.g == fmax)
			hsl.x = (1.0 / 3.0) + deltaR - deltaB; // Hue
		else if (color.b == fmax)
			hsl.x = (2.0 / 3.0) + deltaG - deltaR; // Hue

		if (hsl.x < 0.0)
			hsl.x += 1.0; // Hue
		else if (hsl.x > 1.0)
			hsl.x -= 1.0; // Hue
	}

	return hsl;
}

float HueToRGB(float f1, float f2, float hue) {
	if (hue < 0.0)
		hue += 1.0;
	else if (hue > 1.0)
		hue -= 1.0;
	float res;
	if ((6.0 * hue) < 1.0)
		res = f1 + (f2 - f1) * 6.0 * hue;
	else if ((2.0 * hue) < 1.0)
		res = f2;
	else if ((3.0 * hue) < 2.0)
		res = f1 + (f2 - f1) * ((2.0 / 3.0) - hue) * 6.0;
	else
		res = f1;
	return res;
}

vec3 HSLToRGB(vec3 hsl) {
	vec3 rgb;

	if (hsl.y == 0.0)
		rgb = vec3(hsl.z); // Luminance
	else {
		float f2;

		if (hsl.z < 0.5)
			f2 = hsl.z * (1.0 + hsl.y);
		else
			f2 = (hsl.z + hsl.y) - (hsl.y * hsl.z);

		float f1 = 2.0 * hsl.z - f2;

		rgb.r = HueToRGB(f1, f2, hsl.x + (1.0/3.0));
		rgb.g = HueToRGB(f1, f2, hsl.x);
		rgb.b= HueToRGB(f1, f2, hsl.x - (1.0/3.0));
	}
	
	return rgb;
}

// HSV

vec3 rgb2hsv(vec3 c) {
    vec4 K = vec4(0.0, -1.0 / 3.0, 2.0 / 3.0, -1.0);
    vec4 p = mix(vec4(c.bg, K.wz), vec4(c.gb, K.xy), step(c.b, c.g));
    vec4 q = mix(vec4(p.xyw, c.r), vec4(c.r, p.yzx), step(p.x, c.r));

    float d = q.x - min(q.w, q.y);
    float e = 1.0e-10;
    return vec3(abs(q.z + (q.w - q.y) / (6.0 * d + e)), d / (q.x + e), q.x);
}

vec3 hsv2rgb(vec3 c) {
    vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
    vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
    return c.z * mix(K.xxx, clamp(p - K.xxx, 0.0, 1.0), c.y);
}

void main() {
    vec4 texel = texture2D(s_cc3Texture2Ds[0], v_texCoord);
//    vec3 converted = RGBToHSL(texel.rgb); converted += hsl; vec3 final = HSLToRGB(converted);
    vec3 hsv = rgb2hsv(texel.rgb);
    hsv = clamp(hsv + hsl, 0.0, 1.0);
    vec3 rgb = hsv2rgb(hsv);
    vec3 shiftedTexel = LevelsControlInput(rgb, inBlack, inGamma, inWhite);
    gl_FragColor = vec4(shiftedTexel, 1);
}

#endif