
#if !defined(PROGRAMMABLE_BLENDING_SHADER)
#define PROGRAMMABLE_BLENDING_SHADER

/**
 * Provides functionality to apply programmable blending to current fragment.
 * If programmable blending is not supported, use a checkerboard pattern.
 */

//#define PROGRAMMABLE_BLENDING
//#define DIFFERENCE_BLENDING
//#define SCREEN_BLENDING
//#define GENERIC_PROGRAMMABLE_BLENDING


#if defined(GENERIC_PROGRAMMABLE_BLENDING) && defined(GL_EXT_shader_framebuffer_fetch)
vec3 getProgrammableBlendingValues()        { return ParticleFixedMaterial_programmableBlendingValues; }
#endif

void apply_programmable_blending(vec4 sourceColor) {
#if defined(PROGRAMMABLE_BLENDING)
    #if defined(GL_EXT_shader_framebuffer_fetch)
        #if __VERSION__ < 300
            lowp vec4 destColor = gl_LastFragData[0];
        #endif

        #if defined(DIFFERENCE_BLENDING)
            gl_FragColor = abs(destColor - sourceColor);
        #endif

        #if defined(SCREEN_BLENDING)
            gl_FragColor.rgb = sourceColor.rgb * sourceColor.a + destColor.rgb - sourceColor.rgb * sourceColor.a * destColor.rgb;
            gl_FragColor.a = 1.0;
        #endif

        #if defined(GENERIC_PROGRAMMABLE_BLENDING)
            /* Generic programmable blending with mixing, multiplying, and adding */
            vec3 programmableBlendingValues = getProgrammableBlendingValues();
            float blend_mix = programmableBlendingValues.x;
            float blend_multiply = programmableBlendingValues.y;
            float blend_additive = programmableBlendingValues.z;

            /* alpha blending with blend_mix */
            float source_alpha_with_blend_mix = sourceColor.a * blend_mix;
            lowp vec3 frame = sourceColor.rgb * source_alpha_with_blend_mix + destColor.rgb * (1.0 - source_alpha_with_blend_mix);
            /* multiplicative blending */
            float source_alpha_with_blend_multiply = sourceColor.a * blend_multiply;
            frame += frame.rgb * sourceColor.rgb * source_alpha_with_blend_multiply;
            /* additive blending */
            frame += sourceColor.rgb * sourceColor.a * blend_additive;
            gl_FragColor.rgb = frame;
            gl_FragColor.a = 1.0;
        #endif
    #else //  defined(GL_EXT_shader_framebuffer_fetch)
        /* For programmable blending on unsupported hardware, create a checkerboard pattern */
        float square_size = 10.0;
        vec4 checkerboard_color_a = vec4(0.62, 0.62, 0.62, 1.0);
        vec4 checkerboard_color_b = vec4(0.86, 0.86, 0.86, 1.0);
        float xCheck = float(int(mod(gl_FragCoord.x / square_size, 2.0)));
        float yCheck = float(int(mod(gl_FragCoord.y / square_size, 2.0)));
        vec4 checkerboard_color = mix(checkerboard_color_a, checkerboard_color_b, mod(xCheck + yCheck, 2.0));
        gl_FragColor = mix(gl_FragColor, checkerboard_color, 0.7);
    #endif
#else // defined(PROGRAMMABLE_BLENDING)
    gl_FragColor = sourceColor;
#endif
}

#endif // PROGRAMMABLE_BLENDING_SHADER
