

#if !defined(PROJECTION_UTILITY_SHADER)
#define PROJECTION_UTILITY_SHADER

#include "../framework/system_uniforms.sh"

////////////////////////////////////////////////////////////////////////////////////////////////////
    //      P R O J E C T I O N   U T I L I T I E S      //
////////////////////////////////////////////////////////////////////////////////////////////////////

/**
 * Given "depthSample" value from a depth buffer, calculate the eqivalent value world space units.
 *
 * Depth values writen into the depth buffer are not linear in world space. They are biased to allow
 * for more precision near the camera.
 *
 * Often, however, we want to know what this corresponds to in linear world space units (ie, which
 * might be meters or centimeters, or whatever). This function converts a value that has been read
 * from a depth buffer into linear world space units.
 */
#if defined(METAL_INSTEAD_OF_OPENGL)
float depthSampleToWorldSpaceDepth(constant CocosUniforms& CocosUniforms, float depthSample)
#else
float depthSampleToWorldSpaceDepth(float depthSample)
#endif
{
    float zn = 2.0 * depthSample - 1.0;         // Normalized device coordinates
    return getCameraFrustumDepth().w / (zn + getCameraFrustumDepth().z);

    // Equivalent to:
    // float zFar = getCameraFrustumDepth().x;
    // float zNear = getCameraFrustumDepth().y;
    // return zNear * zFar / (zFar - depthSample * (zFar - zNear));
}

#if defined(METAL_INSTEAD_OF_OPENGL)
#define reciprocalViewportSize()    fn_reciprocalViewportSize(CocosUniforms)
vec2 fn_reciprocalViewportSize(constant CocosUniforms& CocosUniforms)
#else
vec2 reciprocalViewportSize()
#endif
{ 
    return 1.0 / getViewport().zw; 
}

/**
 * Given "depthSample" value from a depth buffer, calculate the view space position of current fragment.
 *
 * Reference: https://stackoverflow.com/questions/32227283/getting-world-position-from-depth-buffer-value
 */
#if defined(METAL_INSTEAD_OF_OPENGL)
vec4 viewSpacePositionFromDepthSample(constant CocosUniforms& CocosUniforms, float depthSample, matrix_float4x4 invProjMatrix)
#else
vec4 viewSpacePositionFromDepthSample(float depthSample, mat4 invProjMatrix)
#endif
{
    vec2 uv = (gl_FragCoord.xy - getViewport().xy) / getViewport().zw;
    vec3 ndcPos = 2.0 * vec3(uv, depthSample) - 1.0;
    vec4 clipPos = vec4(ndcPos, 1.0);
    vec4 viewPos = invProjMatrix * clipPos;
    viewPos /= viewPos.w;
    
    return viewPos;
}

#endif // PROJECTION_UTILITY_SHADER
