
#if !defined(SHADOW_FRAG_SHADER)
#define SHADOW_FRAG_SHADER

// In sync with PGShadowFilteringMode in PGShadowManager.h
#define SHADOW_FILTERING_NONE       0
#define SHADOW_FILTERING_PCF4       1
#define SHADOW_FILTERING_STOCHASTIC 2
#define SHADOW_FILTERING_CHS        3

// Enable comparison mode by default
#if !defined(SHADOW_USE_COMPARISON_SAMPLER)
    #if __VERSION__ >= 300
        #define SHADOW_USE_COMPARISON_SAMPLER 1
    #else
        #define SHADOW_USE_COMPARISON_SAMPLER 0
    #endif
#endif

// Set default mode to PCF4 if unspecified
#if !defined(SHADOW_FILTERING_MODE)
    #define SHADOW_FILTERING_MODE SHADOW_FILTERING_PCF4
#endif

#if defined(SHADOWS_ENABLED) && !defined(SHADOWS_DISABLED) && !defined(SHADOW_GEN)
    #define VSOUT_SHADOW_COORD

    #include "shadow_uniforms.sh"
    #include "texture_utility.sh"

    VARYING_IN vec4 v_shadow;

    float ShadowVisibilitySingleTap(vec2 sampleUV, float comparisonDepth)
    {
        #if (SHADOW_USE_COMPARISON_SAMPLER != 0) && (__VERSION__ >= 300)
            return sampleLevelZero(u_shadowComparisonTexture, vec3(sampleUV, comparisonDepth));
        #else
            float sampledValue = sampleLevelZero(u_shadowMap, sampleUV).r;
            return float(comparisonDepth <= sampledValue);
        #endif
    }

    float ShadowVisibility4Tap(vec2 sampleUV, float comparisonDepth)
    {
        const int sampleCount = 4;
        vec2 offsets[sampleCount];
        offsets[0] = vec2(-0.94201624, -0.39906216);
        offsets[1] = vec2(0.94558609, -0.76890725);
        offsets[2] = vec2(-0.094184101, -0.92938870);
        offsets[3] = vec2(0.34495938, 0.29387760);

        float accumulatedSample = 0.0;
        for (int i = 0; i < sampleCount; ++i) {
            vec2 offset = offsets[i] / GetShadowMapSize().x;
            accumulatedSample += ShadowVisibilitySingleTap(sampleUV + offset, comparisonDepth);
        }

        return accumulatedSample / float(sampleCount);
    }

    float ShadowVisibilityStochastic(vec2 sampleUV, float comparisonDepth)
    {
        #if __VERSION__ >= 300
            const int sampleCount = 3;
            vec2 offsetTable[sampleCount];
            offsetTable[0] = vec2(0.0, .33);
            offsetTable[1] = vec2(.4, -.15);
            offsetTable[2] = vec2(-.45, -.6);

            ivec2 t = ivec2(gl_FragCoord.xy) & 0x3;
            vec2 sampleRotate = texelFetch(u_shadowRotateTable, t, 0).rg;

            float accumulatedSample = 0.0;
            for (int c = 0; c < sampleCount; ++c) {
                vec2 offset = vec2(offsetTable[c].x * sampleRotate.x - offsetTable[c].y * sampleRotate.y,
                                   offsetTable[c].x * sampleRotate.y + offsetTable[c].y * sampleRotate.x) * 0.5;
                accumulatedSample += ShadowVisibilitySingleTap(sampleUV + offset, comparisonDepth);
            }

            return accumulatedSample / float(sampleCount);
        #else
            return 0.0;
        #endif
    }

    #if SHADOW_FILTERING_MODE == SHADOW_FILTERING_CHS
        #include "contact_soft_shadow.fsh"
        float ShadowVisibilityContactHardening(vec2 sampleUV, float comparisonDepth)
        {
            float penumbraRatio;
            return evaluateContactHardeningShadow(vec3(sampleUV, comparisonDepth), penumbraRatio);
        }
    #endif

    float UberShadowVisibility(vec2 sampleUV, float comparisonDepth)
    {
        #if SHADOW_FILTERING_MODE == SHADOW_FILTERING_NONE
            return ShadowVisibilitySingleTap(sampleUV, comparisonDepth);
        #elif SHADOW_FILTERING_MODE == SHADOW_FILTERING_STOCHASTIC
            return ShadowVisibilityStochastic(sampleUV, comparisonDepth);
        #elif SHADOW_FILTERING_MODE == SHADOW_FILTERING_CHS
            return ShadowVisibilityContactHardening(sampleUV, comparisonDepth);
        #else // when unspecified, use SHADOW_FILTERING_PCF4
            return ShadowVisibility4Tap(sampleUV, comparisonDepth);
        #endif
    }
#endif

vec3 apply_shadow(vec3 color)
{
    #if defined(VSOUT_SHADOW_COORD)
        // Boundary check, prevent reading beyond the edge of the shadow texture. We consider everything outside of
        // shadow projection region to be "unshadowed"
        if (v_shadow.x < 0.0 || v_shadow.x > 1.0 || v_shadow.y < 0.0 || v_shadow.y > 1.0) {
            return color;
        }

        float shadowCoverage = 1.0 - UberShadowVisibility(v_shadow.xy, v_shadow.z);
        color *= mix(vec3(1.0), GetShadowColor(), shadowCoverage * GetShadowIntensity());
    #endif
    return color;
}

void shadow_receive()
{
    #if defined(VSOUT_SHADOW_COORD)
        gl_FragColor.rgb = apply_shadow(gl_FragColor.rgb);
    #endif
}

#endif // SHADOW_FRAG_SHADER
