#include "shadow.fsh"
#include "environmentconfig.sh"
#include "system_uniforms.sh"

//===================================================================================================================//
// UNIFORMS //
//===================================================================================================================//

uniform sampler2D s_cc3Texture2Ds[7];

uniform samplerCube u_reflectionMap;
uniform samplerCube nebulaMap;


#if defined(U_SPEC)
    uniform float SpecMult;
    uniform float SpecPower;
    uniform vec3 SpecColor;
#endif

#if defined(U_REFL_COLOR)
    uniform vec3 ReflectColor;
#endif

#if defined(U_ALPHA_MULTIPLY)
    uniform float AlphaMultiply;
#endif

#if defined(U_DEFAULT_LIGHT_DIR) || defined(U_DEFAULT_SPEC_LIGHT)
    uniform vec3 GlobalUniform_SpecularLightDir;
#endif

#if defined(U_REFL_VALUE)
    uniform float ReflectMultiply;
#endif

#if defined(U_SPEC_SETTINGS)
    uniform float SpecularPower;
    uniform float SpecularIntensity;
    uniform vec4 SpecularColor;
#endif

#if defined(U_FRESNEL_LIGHT_COLOR)
    uniform vec4 FresnelLightColor;
    uniform float FresnelLightMultiplier;
#endif

#if defined(U_ALPHAFADE)
    uniform float AlphaFade;
#endif

#if defined(U_LOADING_FADE)
    uniform float LoadingFade;
    uniform vec4 LoadingColor;
    uniform float PulseSpeed;
    uniform float PulseTime;
    uniform vec3 WorldPosition;
#endif

#if defined(U_FADE_COLOR)
    uniform vec4 FadeColor;
#endif

//===================================================================================================================//
// SAMPLER INDICIES //
//===================================================================================================================//

//diffuse texture default
#if !defined(DIFFUSE_MAP_0_INDEX)
    #define DIFFUSE_MAP_0_INDEX 0
#endif


//===================================================================================================================//
// VARYINGS //
//===================================================================================================================//

//instancing
#if defined(INSTANCING_ENABLED)
    #if defined(INTEGER_ATTRIBUTES)
        VARYING_IN int v_instanceID;
    #else
        VARYING_IN float v_instanceID; // "int" with small constant added to handle flooring
    #endif
    #define INSTANCE_ID (int(v_instanceID))
#else
    #define INSTANCE_ID 0
#endif

//positions
#if defined(VS_WORLD_POSITION)
    VARYING_IN vec3 v_position_world;
#endif
VARYING_IN vec3 v_position_local;

//texture coordinates
VARYING_IN vec2 v_texCoord0;
#if defined(VS_UV1)
    VARYING_IN vec2 v_texCoord1;
#endif

//material color
#if defined(VS_DIFFUSE_COLOR)
    VARYING_IN vec4 v_material_diffuse_color;
#endif

//normals
VARYING_IN vec3 v_N;
#if defined(VS_TANGENTS)
    VARYING_IN vec3 v_T;
    VARYING_IN vec3 v_B;
#endif

//vertex color
#if defined(VS_VERTEX_COLOR)
    VARYING_IN vec4 v_vertexColor;
#endif

//world lighting
#if defined(VS_WORLD_LIGHTING) && !defined(VS_FIXEDWORLD_LIGHTING)
    VARYING_IN vec3 v_world_light_dir;
    vec3 getWorldLightDirection() { return v_world_light_dir; }
#elif defined(VS_WORLD_LIGHTING) && defined(VS_FIXEDWORLD_LIGHTING)
    uniform vec3 WorldLightDir;
    vec3 getWorldLightDirection() { return WorldLightDir; }
#endif

//world normals
#if defined(VS_WORLD_NORMAL)
    VARYING_IN vec3 v_world_normal_dir;
#endif

//time
#if defined(VS_TIME)
    VARYING_IN float v_time;
#endif

//===================================================================================================================//
// MAIN SHADER FUNCTIONS //
//===================================================================================================================//


#if defined(INSTANCING_ENABLED) && defined(VS_DIFFUSE_COLOR) && !defined(SHADOW_GEN)
    vec4 getMaterialColor() { return v_material_diffuse_color; }
#else
    uniform vec4 u_cc3MaterialDiffuseColor;
    vec4 getMaterialColor() { return u_cc3MaterialDiffuseColor; }
#endif

//utility function globals
vec4 diffuseColor;
#if defined(USE_NORMAL_MAP)
    vec3 normalMap;
    vec3 vertNormal;
#endif
vec3 N;
#if defined(USE_REFLECT)
    vec3 V;
    float NdotV;
#endif

vec3 lighting;

#if defined(USE_MULTI0_MASK)
    vec3 multi0Mask;
#endif

#if defined(USE_SPEC)
    vec3 specColor;
#endif

#if defined(USE_FRESNEL)
    float fresnel;
#endif

#if defined(USE_MATERIAL_ALPHA)
    float AlphaFinal;
#endif

#if defined(USE_REFL_MAP)
    vec3 ReflColor;
    float SpecReflMask;
#endif

#if defined(USE_MATERIAL_FRESNEL)
    float FresnelMask;
#endif

#if defined(USE_DEFAULT_SPEC)
    float SpecValue;
#endif

#if defined(USE_SPHERE_COORD)
    vec2 sphereCoord0;
#endif

void setup_textures()
{
    diffuseColor = getMaterialColor() * texture2D(s_cc3Texture2Ds[DIFFUSE_MAP_0_INDEX], v_texCoord0);
    #if defined(S_NORMAL_MAP)
        normalMap = texture2D(s_cc3Texture2Ds[NORMAL_MAP_0_INDEX], v_texCoord0).rgb;
    #endif
    #if defined(S_MULTI0_MASK)
        multi0Mask = texture2D(s_cc3Texture2Ds[MATERIAL_MAP_0_INDEX], v_texCoord0).rgb;
    #endif
    #if defined(USE_MATERIAL_ALPHA)
        AlphaFinal = texture2D(s_cc3Texture2Ds[MATERIAL_MAP_0_INDEX], v_texCoord0).r;
        AlphaFinal *= AlphaMultiply;
    #endif
    #if defined(USE_MATERIAL_SPEC_REFL)
        SpecReflMask = texture2D(s_cc3Texture2Ds[MATERIAL_MAP_0_INDEX], v_texCoord0).g;
    #endif
}

void setup_normals()
{
    N = normalize(v_N);
    #if defined(USE_LEGACY_WORLD_LIGHTING)
        vertNormal = normalize(N * 2.0);
        normalMap *= vec3(2.0);
    #endif
    #if defined(USE_NORMAL_MAP)
        vec3 normal_t = normalMap * 2.0 - 1.0;
        N = normalize(mat3(v_T, v_B, v_N) * normal_t);
    #endif
}

void setup_sphere_coords()
{
    #if defined(USE_SPHERE_COORD)
        vec3 n = normalize(mat3(getWorldToCamera()) * reflect(N, V));
        sphereCoord0 = vec2(0.5, 0.5) + 0.5 * vec2(n.x, n.y);
    #endif
    #if defined(USE_REFL_MAP)
        ReflColor = texture2D(s_cc3Texture2Ds[REFL_MAP_0_INDEX], sphereCoord0).rgb;
    #endif
    #if defined(USE_MATERIAL_FRESNEL)
        FresnelMask = texture2D(s_cc3Texture2Ds[MATERIAL_MAP_0_INDEX], sphereCoord0).b;
    #endif
}

void setup_reflect()
{
    #if defined(USE_REFLECT)
        #if defined(USE_WORLD_LIGHTING)
            V = normalize(getWorldLightDirection());
        #elif defined(USE_LEGACY_WORLD_LIGHTING)
            V = normalize(getCameraPositionWorld() - v_position_world);
        #else
            V = normalize(getCameraPositionWorld() - v_position_world);
        #endif
        NdotV = clamp(dot(N, V), 0.0, 1.0);
    #endif
}

void setup_fresnel_texture()
{
    #if defined(USE_FRESNEL)   
        #if defined(USE_LEGACY_WORLD_LIGHTING)
            fresnel = 1.0 - min(max(0.0, dot(V, normalize(vertNormal))), 0.99);
        #else
            fresnel = 1.0 - min(max(0.0, dot(V, normalize(v_N))), 0.99);
        #endif
        #if defined(USE_MULTI0_MASK)
            fresnel = texture2D(s_cc3Texture2Ds[MATERIAL_MAP_0_INDEX], vec2(fresnel, 0.1)).g;
            fresnel *= multi0Mask.r;
        #endif
    #endif
}

void setup_world_lighting()
{
    lighting = vec3(1.0);
    #if defined(USE_WORLD_LIGHTING)
        float light_value = clamp(dot(N, normalize(getWorldLightDirection())) * 0.5 + 0.5, 0.0, 1.0);
        lighting = mix(EnvironmentConfig.WorldAmbientColor, EnvironmentConfig.WorldDiffuseColor, light_value);
    #elif defined(USE_LEGACY_WORLD_LIGHTING)
        float light_value = clamp(dot(N, normalize(getWorldLightDirection())) * 0.65 + 0.15, 0.0, 1.0);
        lighting = mix(EnvironmentConfig.WorldAmbientColor, EnvironmentConfig.WorldDiffuseColor, light_value);
        lighting = vec3(1.2);
    #elif defined(USE_DEFAULT_LIGHTING)
        float light_value = clamp(dot(N, normalize(GlobalUniform_SpecularLightDir)) * 0.5 + 0.5, 0.0, 1.0);
        lighting = mix(EnvironmentConfig.WorldAmbientColor, EnvironmentConfig.WorldDiffuseColor, light_value);
    #endif
}

void apply_lighting()
{
    gl_FragColor = diffuseColor;
    #if !defined(USE_LEGACY_WORLD_LIGHTING)
        gl_FragColor.rgb *= lighting;
    #endif
    #if defined(USE_SPEC)
        #if defined(USE_LEGACY_WORLD_LIGHTING)
            vec3 r = reflect(-normalize(getWorldLightDirection()), vertNormal);
        #else
            vec3 r = reflect(-normalize(getWorldLightDirection()), N);
        #endif
        float spec_dot_value = dot(V, r);
        float spec_value = pow(max(0.0, spec_dot_value), SpecPower);

        specColor = spec_value * SpecColor;
        #if defined(USE_MULTI0_MASK)
            specColor *= vec3(multi0Mask.r);
        #endif
        specColor *= vec3(SpecMult);
        gl_FragColor.rgb += specColor;
    #endif

    #if defined(USE_DEFAULT_ENV_MAP)
        vec3 R = reflect(-V, N);
        vec3 reflection = textureCube(nebulaMap, R).rgb;
        gl_FragColor.rgb += reflection * ReflectColor.rgb;
    #endif

    #if defined(USE_FRESNEL)
        gl_FragColor.rgb += vec3(fresnel);
    #endif
    #if defined(USE_LEGACY_WORLD_LIGHTING)
        gl_FragColor.rgb *= lighting;
    #endif
}

void apply_reflection()
{
    #if defined(USE_REFL_MAP)
        gl_FragColor.rgb *= mix(vec3(1.0), gl_FragColor.rgb * ReflColor * ReflectMultiply * AlphaFinal, SpecReflMask * ReflectMultiply);
    #endif
}

void apply_specular()
{
    #if defined(USE_DEFAULT_SPEC)
        float spec_power = SpecularPower;

        float spec_intensity = SpecularIntensity;

        vec3 spec_color = SpecularColor.rgb;

        vec3 spec_reflect = reflect(normalize(GlobalUniform_SpecularLightDir), N);
        float NdotL = clamp(dot(-spec_reflect, V), 0.0, 1.0);

        float light_specular = pow(NdotL, spec_power);

        float spec_mask = SpecReflMask;

        SpecValue = light_specular * spec_intensity * spec_mask;

        gl_FragColor.rgb *= vec3(1.0 + (SpecValue * spec_color));
    #endif
}

void apply_fresnel_light()
{
    #if defined(USE_MATERIAL_FRESNEL)
        gl_FragColor.rgb += vec3(FresnelMask) * FresnelLightColor.rgb * FresnelLightMultiplier;
    #endif
}

void apply_fog()
{
    #if defined(USE_FOG)
        vec3 Vx = (getCameraPositionWorld() - v_position_world);
        float Vx_length = length(Vx);
        float fog_far = EnvironmentConfig.FogNearFarMax.x + EnvironmentConfig.FogNearFarMax.y;
        float fog_blend = clamp((Vx_length - EnvironmentConfig.FogNearFarMax.x) / fog_far, 0.0, EnvironmentConfig.FogNearFarMax.z);
        gl_FragColor.rgb = mix(gl_FragColor.rgb, EnvironmentConfig.FogColor, fog_blend);
    #endif
}

void apply_fades()
{
    #if defined(USE_LOADING_FADE)

        float time = PulseTime;
        #if defined(USE_TIME_VALUE)
            time = v_time;
        #endif
        vec3 refN = N;
        #if defined(USE_NORMAL_MAP)
            refN = vertNormal;
        #endif
        vec3 refV = V;
        refV = normalize(vec3(0.0, 1.0, 0.0));
        float loading_shader_uv = dot(refN, refV);
        loading_shader_uv *= 0.45;
        loading_shader_uv += time * PulseSpeed;

        vec3 dir = normalize(v_position_world - WorldPosition);
        float u = atan(dir.z, dir.x) * 0.5 / M_PI + 0.25;
        float v = dir.y * 0.35 + 0.5;
        vec2 sphere_uv = vec2(u * -1.0, v) * vec2(2.0, 0.75);

        float fresnel_wave = texture2D(s_cc3Texture2Ds[MATERIAL_MAP_0_INDEX], vec2(-loading_shader_uv, 0.1)).g;

        float grid = texture2D(s_cc3Texture2Ds[AUX_MAP_0_INDEX], sphere_uv).r;

        float appliedLoadingFade = mix(dot(vec3(0.299, 0.587, 0.114), gl_FragColor.rgb) * LoadingFade, 1.0, LoadingFade);

        vec3 R = reflect(-V, N);
        vec3 reflection = textureCube(nebulaMap, R).rgb;


        vec3 adjustedWave = vec3(grid * fresnel_wave);
        vec3 loadingColorAdjusted = LoadingColor.rgb + vec3(1.0 - LoadingFade) * reflection + adjustedWave;

        gl_FragColor.rgb = mix(gl_FragColor.rgb, loadingColorAdjusted, fresnel_wave * appliedLoadingFade * grid);
        

        gl_FragColor.rgb *= mix(vec3(1.0), vec3(fresnel_wave) + reflection, appliedLoadingFade);


        float alphaAdjusted = clamp(fresnel_wave * grid + 0.5, 0.0, 1.0);
        alphaAdjusted = 1.0 - fresnel_wave * fresnel_wave * fresnel_wave;

        gl_FragColor.a *= mix(1.0, alphaAdjusted, appliedLoadingFade);
    
    #endif

    #if defined(USE_FADE_COLOR)
        float gray = dot(vec3(0.299, 0.587, 0.114), mix(gl_FragColor.rgb, vec3(0.75), FadeColor.a));

        gl_FragColor.rgb = mix(gl_FragColor.rgb, FadeColor.rgb * gray, FadeColor.a);


    #endif
}

void apply_alpha()
{
    #if defined(USE_MATERIAL_ALPHA)
        #if defined(USE_MATERIAL_FRESNEL)
            float alpha_fresnel = clamp(dot(V, N), 0.0, 1.0);
            alpha_fresnel = clamp(1.0 - pow(alpha_fresnel, 4.0), 0.25, 1.0);
            gl_FragColor.a *= mix(AlphaFinal, FresnelLightColor.a, max(FresnelMask * FresnelLightMultiplier, SpecValue * alpha_fresnel));
            gl_FragColor.a = clamp(gl_FragColor.a + alpha_fresnel, 0.0, 1.0);
        #else
            gl_FragColor.a *= AlphaFinal;
        #endif
        
    #endif

    #if defined(USE_ALPHAFADE)
        gl_FragColor.a *= AlphaFade;
    #endif
}

void apply_shadows()
{
    world_shadow_receive();
}

