#include "shadow.vsh"
#include "environmentconfig.sh"
#include "system_uniforms.sh"

//===================================================================================================================//
// ATTRIBUTES //
//===================================================================================================================//

//positions
ATTRIBUTE vec4 a_cc3Position;

//normals
ATTRIBUTE vec3 a_cc3Normal;
#if defined(A_TANGENTS)
    ATTRIBUTE vec3 a_cc3Tangent;
    ATTRIBUTE vec3 a_cc3Bitangent;
#endif

//vertex color
#if defined(A_VERTEX_COLOR)
    ATTRIBUTE vec4 a_cc3Color;
#endif

//texture coordinates
ATTRIBUTE vec2 a_cc3TexCoord0;
#if defined(A_UV1)
    ATTRIBUTE vec2 a_cc3TexCoord1;
#endif

//instancing
#if defined(UNIFORM_BASED_INSTANCING)
    #if defined(INTEGER_ATTRIBUTES)
        ATTRIBUTE int a_cc3InstanceId;
    #else
        ATTRIBUTE float a_cc3InstanceId;
    #endif
#endif

//skinning
#if defined(A_SKINNING)
    ATTRIBUTE vec4 a_cc3BoneWeights;
    #if defined(INTEGER_ATTRIBUTES)
        ATTRIBUTE ivec4 a_cc3BoneIndices;
    #else
        ATTRIBUTE vec4 a_cc3BoneIndices;
    #endif
#endif

//===================================================================================================================//
// CONSTANTS //
//===================================================================================================================//

//instancing
#if defined(INSTANCING_ENABLED)
    #if defined(C_SKINNING)
        #define MAX_INSTANCES 10
    #else
        #define MAX_INSTANCES 64
    #endif
#else
    #define MAX_INSTANCES 1
#endif

//skinning
#if defined(C_SKINNING)
    #define MAX_BONE_MATRICES 56
#endif

//===================================================================================================================//
// UNIFORMS //
//===================================================================================================================//

//skinning
#if defined(U_SKINNING)
    uniform int u_cc3VertexBoneCount;

    // DavidJ -- NOTE -- the elements of "u_cc3BoneMatricesModel4x3" are in transposed order
    //      relative to u_cc3BoneMatricesModel. This is required to ensure that the matrix
    //      can fit in 3 full vec4 rows.
    uniform vec4 u_cc3BoneMatricesModel4x3[3*MAX_BONE_MATRICES];

#endif


//instancing
#ifdef INSTANCING_ENABLED
    // unsupported
#else
    uniform vec4 u_cc3MaterialDiffuseColor;
    vec4 currentMaterialDiffuseColor() { return u_cc3MaterialDiffuseColor; }
    mat4 currentModelMatrix() { return getModelToWorld(); }
    int currentBoneMatrixOffset() { return 0; }
#endif

//time
uniform float u_cc3ObjectTime[MAX_INSTANCES];

//===================================================================================================================//
// VARYINGS //
//===================================================================================================================//

//instancing
#if defined(INSTANCING_ENABLED)
    #if defined(INTEGER_ATTRIBUTES)
        VARYING_OUT int v_instanceID
    #else
        VARYING_OUT float v_instanceID; // "int" with small constant added to handle flooring
    #endif
#endif

//positions
#if defined(VS_WORLD_POSITION)
    VARYING_OUT vec3 v_position_world;
#endif
VARYING_OUT vec3 v_position_local;

//texture coordinates
VARYING_OUT vec2 v_texCoord0;
#if defined(VS_UV1) 
    VARYING_OUT vec2 v_texCoord1;
#endif

//material color
#if defined(VS_DIFFUSE_COLOR)
    VARYING_OUT vec4 v_material_diffuse_color;
#endif

//normals
VARYING_OUT vec3 v_N;
#if defined(VS_TANGENTS)
    VARYING_OUT vec3 v_T;
    VARYING_OUT vec3 v_B;
#endif

//vertex color
#if defined(VS_VERTEX_COLOR)
    VARYING_OUT vec4 v_vertexColor;
#endif

//world lighting
#if defined(VS_WORLD_LIGHTING)
    VARYING_OUT vec3 v_world_light_dir;
#endif

//world normals
#if defined(VS_WORLD_NORMAL)
    VARYING_OUT vec3 v_world_normal_dir;
#endif

//time
#if defined(VS_TIME)
    VARYING_OUT float v_time;
#endif

//===================================================================================================================//
// UTILITY FUNCTIONS //
//===================================================================================================================//

mat4 modelMatrix;
vec4 position_world;

//skinning
#if defined(USE_SKINNING)
    vec3 finalPosition_local;
    vec3 finalNormal_local;
    vec3 finalTangent_local;
    vec3 finalBitangent_local;
#endif

//skinning
#if defined(USE_SKINNING)
    #if defined(VS_TANGENTS)
        vec3 GetAttributeTangent()      { return a_cc3Tangent; }
        vec3 GetAttributeBitangent()    { return a_cc3Bitangent; }
    #else
        vec3 GetAttributeTangent()      { return vec3(1.0, 0.0, 0.0); }
        vec3 GetAttributeBitangent()    { return vec3(0.0, 0.0, 1.0); }
    #endif
    #if !defined(SHADOW_GEN)
        vec3 GetAttributeNormal()       { return a_cc3Normal; }
    #else
        vec3 GetAttributeNormal()       { return vec3(0.0, 1.0, 0.0); }
    #endif
#endif

//skinning
vec3 skinPositionSingleWeight(vec3 position, int boneIndex, float weight) 
{
    #if defined(USE_SKINNING)
        vec4 boneMatrixA = u_cc3BoneMatricesModel4x3[boneIndex*3+0];
        vec4 boneMatrixB = u_cc3BoneMatricesModel4x3[boneIndex*3+1];
        vec4 boneMatrixC = u_cc3BoneMatricesModel4x3[boneIndex*3+2];
        vec3 result;
        result.x = dot(boneMatrixA, vec4(position.xyz, 1.0));
        result.y = dot(boneMatrixB, vec4(position.xyz, 1.0));
        result.z = dot(boneMatrixC, vec4(position.xyz, 1.0));
        return result * weight;
    #else
        return vec3(0.0);
    #endif
}

//skinning
vec3 skinDirectionSingleWeight(vec3 direction, int boneIndex, float weight) 
{
    #if defined(USE_SKINNING)  
        vec4 boneMatrixA = u_cc3BoneMatricesModel4x3[boneIndex*3+0];
        vec4 boneMatrixB = u_cc3BoneMatricesModel4x3[boneIndex*3+1];
        vec4 boneMatrixC = u_cc3BoneMatricesModel4x3[boneIndex*3+2];
        vec3 result;
        result.x = dot(boneMatrixA.xyz, direction.xyz);
        result.y = dot(boneMatrixB.xyz, direction.xyz);
        result.z = dot(boneMatrixC.xyz, direction.xyz);
        return result * weight;
    #else
        return vec3(0.0);
    #endif
}

//skinning
void vertexToEyeSpace(vec4 position, vec3 normal, vec3 tangent, vec3 bitangent) 
{    
    #if defined(USE_SKINNING)  
        ivec4 boneIndices = ivec4(a_cc3BoneIndices);
        vec4 boneWeights = a_cc3BoneWeights;
        
        vec3 workingPosition = skinPositionSingleWeight(position.xyz, currentBoneMatrixOffset() + boneIndices.x, boneWeights.x);

        finalNormal_local = skinDirectionSingleWeight(normal, currentBoneMatrixOffset() + boneIndices.x, boneWeights.x);
        finalTangent_local = skinDirectionSingleWeight(tangent, currentBoneMatrixOffset() + boneIndices.x, boneWeights.x);
        finalBitangent_local = skinDirectionSingleWeight(bitangent, currentBoneMatrixOffset() + boneIndices.x, boneWeights.x);

        for (int i = 1; i < 4; ++i)
        {
            if (i < u_cc3VertexBoneCount)
            {
                // "rotate" the vector components
                boneIndices = boneIndices.yzwx;
                boneWeights = boneWeights.yzwx;

                workingPosition += skinPositionSingleWeight(position.xyz, currentBoneMatrixOffset() + boneIndices.x, boneWeights.x);
            
                finalNormal_local += skinDirectionSingleWeight(normal, currentBoneMatrixOffset() + boneIndices.x, boneWeights.x);
                finalTangent_local += skinDirectionSingleWeight(tangent, currentBoneMatrixOffset() + boneIndices.x, boneWeights.x);
                finalBitangent_local += skinDirectionSingleWeight(bitangent, currentBoneMatrixOffset() + boneIndices.x, boneWeights.x);
            }
        }
        
        finalPosition_local = workingPosition.xyz;
    #endif
}

//===================================================================================================================//
// MAIN SHADER FUNCTIONS //
//===================================================================================================================//


void setup_positions() 
{
    vec4 position = a_cc3Position;

    modelMatrix = currentModelMatrix();

    #if defined(USE_SKINNING)
        vertexToEyeSpace(position, GetAttributeNormal(), GetAttributeTangent(), GetAttributeBitangent());
    #endif

    #if defined(USE_SKINNING)
        v_position_local = finalPosition_local.xyz;
        position_world = modelMatrix * vec4(finalPosition_local.xyz, 1.0);
    #elif !defined(SHADOW_GEN)
        v_position_local = position.xyz;
        position_world = modelMatrix * position;
    #else
        v_position_local = vec3(0.0);
        position_world = modelMatrix * position;
    #endif

    #if !defined(SHADOW_GEN)
        gl_Position = getWorldToProj() * vec4(position_world.xyz, 1.0);
    #else
        gl_Position = getModelToProj() * vec4(position.xyz, 1.0);
    #endif

    #if defined(VS_WORLD_POSITION)
        v_position_world = position_world.xyz;
    #endif   

}

void setup_texcoords()
{

    #if defined(USE_TIME_VALUE)
        v_time = getSceneTime();
    #endif
    v_texCoord0 = a_cc3TexCoord0;
    #if defined(USE_UV1)
        v_texCoord1 = a_cc3TexCoord1;
    #endif
}

void setup_material_diffuse_color()
{
    #if defined(INSTANCING_ENABLED)
        v_material_diffuse_color = currentMaterialDiffuseColor();
    #endif
}

void setup_vertex_color()
{
    #if defined(USE_VERTEX_COLOR)
        v_vertexColor = a_cc3Color;
    #endif
}

void setup_world_lighting()
{
    #if defined(VS_WORLD_LIGHTING)
        v_world_light_dir = EnvironmentConfig.WorldLightDirection * mat3(getWorldToCamera());
    #endif
}

void setup_normals()
{
    #if defined(USE_SKINNING)
        v_N = normalize(mat3(modelMatrix) * finalNormal_local);
        #if defined(USE_NORMAL_MAP)
            v_T = normalize(mat3(modelMatrix) * finalTangent_local);
            v_B = normalize(mat3(modelMatrix) * finalBitangent_local);
        #endif
    #else
        v_N = normalize(mat3(modelMatrix) * a_cc3Normal);
        #if defined(USE_NORMAL_MAP)
            v_T = normalize(mat3(modelMatrix) * a_cc3Tangent);
            v_B = normalize(mat3(modelMatrix) * a_cc3Bitangent);
            vec3 vv = normalize(v_position_world - getCameraPositionWorld());
            v_world_normal_dir = mat3(getWorldToCamera()) * reflect(v_N, vv);
        #endif
    #endif
}

void setup_instancing()
{
    // A small constant must be added because interpolation could cause the value to fall slightly below the original
    // int, thus causing the typecast to access the lower index.
    #if defined(INSTANCING_ENABLED)
        #if defined(INTEGER_ATTRIBUTES)
            v_instanceID = INSTANCE_ID;
        #else
            v_instanceID = float(INSTANCE_ID) + 0.1;
        #endif
    #endif
}

void setup_shadows()
{
    #if !defined(SHADOW_GEN)
        shadow_receive(position_world);
    #endif
}
