
#if !defined(SKINNING_SH)
#define SKINNING_SH

#include "framework/framework.sh"
#include "framework/instancing.sh"
#include "framework/transforms.sh"

#if defined(METAL_INSTEAD_OF_OPENGL)
/* Metal TODO -- skinning */
#else

#if !defined(MAX_BONE_MATRICES)
    #define MAX_BONE_MATRICES 1
#endif

////////////////////////////////////////////////////////////////////////////////////////////////////

uniform int u_cc3VertexBoneCount;

// DavidJ -- NOTE -- the elements of "u_cc3BoneMatricesModel4x3" are in transposed order
//      relative to u_cc3BoneMatricesModel. This is required to ensure that the matrix
//      can fit in 3 full vec4 rows.
#if !defined(SKIN_BONE_TO_WORLD)
    uniform vec4 u_cc3BoneMatricesModel4x3[3*MAX_BONE_MATRICES];
    uniform mat4 u_cc3BoneMatricesModel[MAX_BONE_MATRICES];

    vec4 boneMatrixToModelCol(int idx) { return u_cc3BoneMatricesModel4x3[idx]; }
    vec4 boneMatrixToWorldCol(int idx) { return vec4(0,0,0,0); }
    mat4 boneMatrixToModel(int idx) { return u_cc3BoneMatricesModel[idx]; }
    mat4 boneMatrixToWorld(int idx) { return mat4(0); }
#else
// Thiago -- NOTE -- no support for 4x4 global bone matrices as of now
    uniform vec4 u_cc3BoneMatricesGlobal4x3[3*MAX_BONE_MATRICES];

    vec4 boneMatrixToModelCol(int idx) { return vec4(0,0,0,0); }
    vec4 boneMatrixToWorldCol(int idx) { return u_cc3BoneMatricesGlobal4x3[idx]; }
    mat4 boneMatrixToModel(int idx) { return mat4(0); }
#endif

vec3 skinPositionSingleWeight(vec3 position, int boneIndex, float weight, int targetSpace) {
    #if defined(SKIN_4x3)
        vec4 boneMatrixA = (targetSpace==0) ? boneMatrixToModelCol(boneIndex*3+0) : boneMatrixToWorldCol(boneIndex*3+0);
        vec4 boneMatrixB = (targetSpace==0) ? boneMatrixToModelCol(boneIndex*3+1) : boneMatrixToWorldCol(boneIndex*3+1);
        vec4 boneMatrixC = (targetSpace==0) ? boneMatrixToModelCol(boneIndex*3+2) : boneMatrixToWorldCol(boneIndex*3+2);
        vec3 result;
        result.x = dot(boneMatrixA, vec4(position.xyz, 1.0));
        result.y = dot(boneMatrixB, vec4(position.xyz, 1.0));
        result.z = dot(boneMatrixC, vec4(position.xyz, 1.0));
        return result * weight;
    #else
        mat4 boneMatrix = (targetSpace==0) ? boneMatrixToModel(boneIndex) : boneMatrixToWorld(boneIndex);
        return (boneMatrix * vec4(position.xyz, 1.0) * weight).xyz;
    #endif
}

vec3 skinDirectionSingleWeight(vec3 direction, int boneIndex, float weight, int targetSpace) {
    #if defined(SKIN_4x3)
        vec4 boneMatrixA = (targetSpace==0) ? boneMatrixToModelCol(boneIndex*3+0) : boneMatrixToWorldCol(boneIndex*3+0);
        vec4 boneMatrixB = (targetSpace==0) ? boneMatrixToModelCol(boneIndex*3+1) : boneMatrixToWorldCol(boneIndex*3+1);
        vec4 boneMatrixC = (targetSpace==0) ? boneMatrixToModelCol(boneIndex*3+2) : boneMatrixToWorldCol(boneIndex*3+2);
        vec3 result;
        result.x = dot(boneMatrixA.xyz, direction.xyz);
        result.y = dot(boneMatrixB.xyz, direction.xyz);
        result.z = dot(boneMatrixC.xyz, direction.xyz);
        return result * weight;
    #else
        mat4 boneMatrix = (targetSpace==0) ? boneMatrixToModel(boneIndex) : boneMatrixToWorld(boneIndex);
        return mat3(boneMatrix) * direction.xyz * weight;
    #endif
}


#endif // OpenGL

#if defined(METAL_INSTEAD_OF_OPENGL)
void getCoords_Internal(thread vec3& outPosition, thread RawTangentFrame& outTangentFrame, VSIN vsin, int targetSpace, constant CocosNodeBasedUniforms& CocosNodeBasedUniforms)
#else

void getCoords_Internal(out vec3 outPosition, out RawTangentFrame outTangentFrame, VSIN vsin, int targetSpace)
#endif
{

    #if VS_IN_S_BONE_WEIGHTS
    
        vec3 inputPosition      = vsin._position.xyz;
        RawTangentFrame inputTF = getTangentFrame(vsin);
    
        ivec4 boneIndices       = ivec4(vsin._boneIndices);
        vec4 boneWeights        = vsin._boneWeights;
        int instanceOffset      = getBonesPerInstance() * getInstanceID(vsin);
    
        vec3 workingPosition    = skinPositionSingleWeight(inputPosition.xyz,   boneIndices.x + instanceOffset, boneWeights.x, targetSpace);
        vec3 workingAxis0       = skinDirectionSingleWeight(inputTF._axesA,     boneIndices.x + instanceOffset, boneWeights.x, targetSpace);
        vec3 workingAxis1       = skinDirectionSingleWeight(inputTF._axesB,     boneIndices.x + instanceOffset, boneWeights.x, targetSpace);
    
        for (lowp int i = 1; i < 4; ++i) {
            if (i < u_cc3VertexBoneCount) {
                // "rotate" the vector components, to bring the next values into the 'x' element position
                boneIndices = boneIndices.yzwx;
                boneWeights = boneWeights.yzwx;
                
                workingPosition += skinPositionSingleWeight(inputPosition.xyz,  boneIndices.x + instanceOffset, boneWeights.x, targetSpace);
                workingAxis0    += skinDirectionSingleWeight(inputTF._axesA,    boneIndices.x + instanceOffset, boneWeights.x, targetSpace);
                workingAxis1    += skinDirectionSingleWeight(inputTF._axesB,    boneIndices.x + instanceOffset, boneWeights.x, targetSpace);
            }
        }

        // Note that if there is a scale on the bone-to-world transform (which could come from a scale on
        // the model-to-world), then the tangent frame is not guaranteed to have unit length vectors
        // from here. In worse cases (such as when there is a skew on the transform, or nonuniform
        // scale combined with a rotation), we can actually cause the tangent frame vectors to become
        // non-orthogonal.
        outPosition = workingPosition;
        outTangentFrame._axesA = workingAxis0;
        outTangentFrame._axesB = workingAxis1;
        outTangentFrame._handiness = inputTF._handiness;
        outTangentFrame._type = inputTF._type;
    
    #elif VS_IN_S_POSITION

        if (targetSpace == 0) {
            outPosition = vsin._position.xyz;
            outTangentFrame = getTangentFrame(vsin);
        } else {
            outPosition = transformPoint(getModelToWorld(), vsin._position.xyz).xyz;
            outTangentFrame = transformTangentFrame(getModelToWorldNoScale(), getTangentFrame(vsin));
        }

    #endif
}

#if defined(METAL_INSTEAD_OF_OPENGL)
#else
void getModelSpace(out vec3 outPosition, out RawTangentFrame outTangentFrame, VSIN vsin) {
    getCoords_Internal(outPosition, outTangentFrame, vsin, 0);
}

void getWorldSpace(out vec3 outPosition, out RawTangentFrame outTangentFrame, VSIN vsin) {
    getCoords_Internal(outPosition, outTangentFrame, vsin, 1);
}
#endif

#endif

